/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * arrange.c
 *
 * Builds and manipulates the arrangement
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/

#include <stdio.h>
#include <math.h>
#include "arrange.h"
#include "arrange_geom.h"
#include "arrange_macros.h"
#include "arrange_type.h"
#include "dheap.h"
#include "lists.h"
#include "polygon.h"
#include "standards.h"

#include <values.h>

/*---- initial block size for dheaps ----*/
#define ARRANGE_DHEAP_INIT_SIZE 64
/*#define ARRANGE_DHEAP_INIT_SIZE 2048*/

#define D_NEW(type,dh) ((type*) Dheap_Alloc((dh),sizeof(type)))

#define VGEOM(v)   (the_A->vgeom = push(the_A->vgeom, (v)))
#define EGEOM(e)   (the_A->egeom = push(the_A->egeom, (e)))

/*---- three types of arrangements ----*/
#define IS_SINGLE(a)    ((a)->mirror == NULL)
#define IS_DOUBLE(a)    (((a)->mirror) && ((a)->mirror != (a)))
#define IS_DOUBLESYM(a) ((a)->mirror == (a))


#ifdef ARRANGE_DEBUG
/*----toggle this in the debugger to turn on/off arrange_verbose ----*/
Boolean arrange_verbose=FALSE;
Boolean arrange_mike=FALSE;

/*************************************************************/
/***
  At least for now, it will be nice to have all the objects threaded
  in a simple list so that it's possible to get to them all without
  traversing arrangement
  
  WARNING -- Copying entire structures will OVERWRITE the link.
  ***/
#endif


TrapVert* Make_TrapVert(the_A)
  Arrange	*the_A;
{
    TrapVert	*v;
    
    v = D_NEW(TrapVert,the_A->Dheap_verts);
    v->id = ++the_A->num_verts;
    v->mirror = NULL;
    v->direction = NULL;
    v->polys = NULL;
    v->tlTrap = v->trTrap = v->blTrap = v->brTrap = NULL;
    v->topTV = v->botTV = NULL;
    
#ifdef ARRANGE_DEBUG
    v->link=the_A->vert_link;  the_A->vert_link=v;
#endif
    
    return(v);
}


TrapThreadVert* Make_TrapThreadVert(the_A)
  Arrange	*the_A;
{
    TrapThreadVert	*tv;
    
    tv = D_NEW(TrapThreadVert,the_A->Dheap_tverts);
    tv->id = ++the_A->num_tverts;
    tv->source = NULL;
    tv->lNeigh = tv->rNeigh = NULL;
    
#ifdef ARRANGE_DEBUG
    tv->link=the_A->tvert_link;  the_A->tvert_link=tv;
#endif
    
    return(tv);
}


TrapEdge* Make_TrapEdge(the_A)
  Arrange	*the_A;
{
    TrapEdge	*e;
    
    e = D_NEW(TrapEdge,the_A->Dheap_edges);
    e->id = ++the_A->num_edges;
    e->mirror = NULL;
    e->normal = NULL;
    e->polys_on = e->polys_above = e->polys_below = NULL;
    e->lThrVert = e->rThrVert = NULL;
    e->tlTrap = e->trTrap = e->blTrap = e->brTrap = NULL;
    
#ifdef ARRANGE_DEBUG
    e->link=the_A->edge_link;  the_A->edge_link=e;
#endif

    return(e);
}

Trap* Make_Trap(the_A)
  Arrange	*the_A;
{
    Trap	*t;
    
    t = D_NEW(Trap,the_A->Dheap_traps);
    t->id = ++the_A->num_traps;
    t->lvert = t->rvert = NULL;
    t->lvtype = t->rvtype = com;
    t->te = t->be = NULL;
    t->hist = NULL;
    
#ifdef ARRANGE_DEBUG
    t->link=the_A->trap_link;  the_A->trap_link=t;
#endif
    
    return(t);
}


TrapHist* Make_TrapHist(the_A)
  Arrange	*the_A;
{
    TrapHist	*h;
    
    h = D_NEW(TrapHist,the_A->Dheap_hist);
    h->id = ++the_A->num_hists;
    h->nodetype = leaf;
    h->data.v = NULL;
    h->trap = NULL;
    h->ans[0] = h->ans[1] = NULL;
    
#ifdef ARRANGE_DEBUG
    h->link=the_A->hist_link;  the_A->hist_link=h;
#endif
    
    return(h);
}


PolyTree* Make_PTree(the_A)
  Arrange	*the_A;
{
    PolyTree	*pt;
    
    pt = D_NEW(PolyTree,the_A->Dheap_ptree);
    pt->source = NULL;
    pt->next = NULL;
    
    return(pt);
}


PolySource* Make_PSource(the_A)
  Arrange	*the_A;
{
    PolySource	*ps;
    
    ps = D_NEW(PolySource,the_A->Dheap_ptree);
    ps->poly = NULL;
    ps->flipped = FALSE;
    
    return(ps);
}





#ifdef ARRANGE_DEBUG

Link_Init_Single(the_A)
  Arrange	*the_A;
{
    the_A->vert_link=NULL;
    the_A->tvert_link=NULL;
    the_A->edge_link=NULL;
    the_A->trap_link=NULL;
    the_A->hist_link=NULL;
}

#endif




/*************************************************************
  Function:  Arrange_vert_nexttrap(the_A, the_vert, the_trap, ccw,
  result_trap, result_edge)
  Input:
      the_A    --- an arrangement
      the_vert --- TrapVert to walk around
      the_trap --- current Trap in walk
      ccw      --- walk counterclockwise?  (or else clockwise)
  Output:
      result_trap --- next trap walking (counter)clockwise
      result_edge --- next edge walking (counter)clockwise
  Effects:    <none>
  
  Description:  This routine simulates a (counter)clockwise circular
  list around a TrapVert, even though no such list exists explicitly.
  
  Given a trap, this will return the next trap seen when walking
  around the vert in a (counter)clockwise order.  It will also
  return the edge that is crossed to get there if it exists.
  (Otherwise, NULL will be returned for that parameter)
  
  If the walk hits the horizon, then result_edge will be set to
  the edge that was hit, and result_trap will be set to NULL.
  **************************************************************/
void Arrange_vert_nexttrap(Arrange	*the_A,
			   TrapVert	*the_vert,
			   Trap		*the_trap,
			   Boolean	ccw,
			   Trap		**result_trap,
			   TrapEdge	**result_edge)
{
    
    Boolean	left_of_vert;
    Boolean	going_up;
    
    if (T_lvert(the_trap) == the_vert) {
	left_of_vert = FALSE;
    } else if (T_rvert(the_trap) == the_vert) {
	left_of_vert = TRUE;
    } else {
	/*--- Error...trap is not incident to vert ---*/
	printf("Error in Arrange_vert_nexttrap...\n");
	printf("  the_vert is not incident on the_trap.\n");
	exit(0);
    }
    
    going_up = (left_of_vert != ccw);
    
    
    if (going_up) {
	if (the_vert == (left_of_vert ? E_rvert(T_te(the_trap)) :
			 E_lvert(T_te(the_trap)))) {
	    /*--- crossing an edge ---*/
	    (*result_edge) = T_te(the_trap);
	    (*result_trap) = (left_of_vert ? E_trTrap(*result_edge) :
			      E_tlTrap(*result_edge));
	} else {
	    /*--- crossing a thread ---*/
	    (*result_edge) = NULL;
	    (*result_trap) = (left_of_vert ? T_trn(the_trap) :
			      T_tln(the_trap));
	}
    } else {
	if (the_vert == (left_of_vert ? E_rvert(T_be(the_trap)) :
			 E_lvert(T_be(the_trap)))) {
	    /*--- crossing an edge ---*/
	    (*result_edge) = T_be(the_trap);
	    (*result_trap) = (left_of_vert ? E_brTrap(*result_edge) :
			      E_blTrap(*result_edge));
	} else {
	    /*--- crossing a thread ---*/
	    (*result_edge) = NULL;
	    (*result_trap) = (left_of_vert ? T_brn(the_trap) :
			      T_bln(the_trap));
	}
    }
}



/*************************************************************
  Function:  Arrange_ThreadVert_Direction_And_Edge(the_A,tvert,
  fake_edge, endpoint)
  Input:
      the_A --- an arrangement
      tvert --- a TrapThreadVert
  Output:
      fake_edge --- the upward edge simulating a thread
      endpoint  --- the vector direction of tvert
  Effects:    <none>
  
  Description:  This routine calculates the upward normal to the
  "fake" edge that would like immediately overlapping the thread
  through tvert, and it then calculates the normalized direction
  of tvert by intersecting this edge with the true edge which
  the thread hits.
  
  **************************************************************/
void Arrange_ThreadVert_Direction_And_Edge(the_A, tvert, fake_edge, endpoint)
  Arrange*		the_A;
  TrapThreadVert	*tvert;
  Vector		*fake_edge;
  Vector		*endpoint;
{
    
    double	scale;
    VertGeom*	temp_vert;
    EdgeGeom*	temp_edge;
    Boolean	freed = FALSE;
    
    temp_edge = EdgeThroughVerts(V_direction(the_A->tpole),
				 V_direction(TV_source(tvert)),
				 NULL);
    
    temp_vert = EdgeEdgeIntersect(temp_edge,E_normal(TV_edge(tvert)),
				  NULL,NULL,NULL,NULL);

    if (OnHorizonVert(temp_vert)) {
	FreeVertGeom(temp_vert);
	freed = TRUE;
	temp_vert = (TV_from_above(tvert) ?
		     V_direction(E_lvert(TV_edge(tvert))) :
		     V_direction(E_rvert(TV_edge(tvert))));
    }
    
    (*endpoint) = InqVertVector(temp_vert);
    if (!freed) FreeVertGeom(temp_vert);

    (*fake_edge) = InqEdgeVector(temp_edge);
    FreeEdgeGeom(temp_edge);
}



/*************************************************************
  Function:  Arrange_ThreadVert_Direction(the_A,tvert,answer)
  Input:	
      the_A --- an arrangement
      tvert --- a TrapThreadVert
  Output:
      answer --- the vector direction of tvert
  Effects:    <none>
  
  Description:  This routine calculates the normalized
  directional vector for a threadvert, and places the
  result in 'answer' upon return.
  
  **************************************************************/
void Arrange_ThreadVert_Direction(the_A, tvert, answer)
  Arrange		*the_A;
  TrapThreadVert	*tvert;
  Vector		*answer;
{
    Vector	fake_edge;
    
    Arrange_ThreadVert_Direction_And_Edge(the_A,tvert,&fake_edge,answer);
}


/*************************************************************
  Function:  Arrange_Init_Single()
  Input:     <none>
  Output:    pointer to a newly initialized arrangment
  Effects:   <none>
  
  Description:  This routine creates a single arrangement
  on a hemisphere.
  
  The current initialization creates two vertices as the
  top and bottom poles, and two edges connecting them as
  the left and right horizon.  The entire hemisphere is
  a single trap bounded from above and below by the edge,
  and having the two poles as the right and left vertices.
  
  **************************************************************/
Arrange*	Arrange_Init_Single()
{
    Arrange	*the_A;
    Trap	*n_trap;
    TrapVert	*tpole;			/* the "north" pole */
    TrapVert	*bpole;			/* the "south" pole */
    TrapEdge	*lhorizon;		/* the left horizon */
    TrapEdge	*rhorizon;		/* the right horizon */
    TrapHist	*t_hist;
    
    the_A = ALLOCATE(Arrange);
    
    the_A->Dheap_verts = Dheap_New(ARRANGE_DHEAP_INIT_SIZE);
    the_A->Dheap_tverts = Dheap_New(ARRANGE_DHEAP_INIT_SIZE);
    the_A->Dheap_edges = Dheap_New(ARRANGE_DHEAP_INIT_SIZE);
    the_A->Dheap_traps = Dheap_New(ARRANGE_DHEAP_INIT_SIZE);
    the_A->Dheap_hist = Dheap_New(ARRANGE_DHEAP_INIT_SIZE);
    the_A->Dheap_ptree = Dheap_New(ARRANGE_DHEAP_INIT_SIZE);
    the_A->polys=NULL;
    the_A->mirror = NULL;
    the_A->vgeom = the_A->egeom = NULL;
    
    the_A->num_verts = the_A->num_tverts =
	the_A->num_edges = the_A->num_traps =
	    the_A->num_hists = 0;
    
#ifdef ARRANGE_DEBUG
    Link_Init_Single(the_A);
#endif
    
    tpole = the_A->tpole = Make_TrapVert(the_A);
    bpole = the_A->bpole = Make_TrapVert(the_A);
    rhorizon = Make_TrapEdge(the_A);
    lhorizon = Make_TrapEdge(the_A);
    n_trap = Make_Trap(the_A);
    
    n_trap->rvert = tpole;
    n_trap->lvert = bpole;
    n_trap->lvtype = n_trap->rvtype = com;
    n_trap->be = rhorizon;
    n_trap->te = lhorizon;
    t_hist = n_trap->hist = the_A->hist_root = Make_TrapHist(the_A);
    t_hist->trap = n_trap;
    
    tpole->direction = GetPoleGeom(FALSE);
    VGEOM(tpole->direction);
    tpole->tlTrap = tpole->blTrap = n_trap;
    
    bpole->direction = GetPoleGeom(TRUE);
    VGEOM(bpole->direction);
    bpole->brTrap = bpole->trTrap = n_trap;
    
    rhorizon->normal = GetHorizonGeom(FALSE);
    EGEOM(rhorizon->normal);
    rhorizon->tlTrap = rhorizon->trTrap = n_trap;
    
    lhorizon->normal = GetHorizonGeom(TRUE);
    EGEOM(lhorizon->normal);
    lhorizon->brTrap = lhorizon->blTrap = n_trap;
    
    return(the_A);
}


/*************************************************************
  Function:  Arrange_Init_Symmetric()
  Input:     <none>
  Output:    pointer to a newly initialized arrangment
  Effects:   <none>
  
  Description:  This routine creates a single arrangement,
  and then 'glues' it to itself so that the "back side"
  behaves in an anti-symmetric manner to the front.
  
  The current initialization creates two vertices as the
  top and bottom poles, and two edges connecting them as
  the left and right horizon.  The entire hemisphere is
  a single trap bounded from above and below by the edge,
  and having the two poles as the right and left vertices.
  
  **************************************************************/
Arrange*	Arrange_Init_Symmetric()
{
    Arrange	*the_A;
    Trap	*trap;
    
    the_A = Arrange_Init_Single();
    
    the_A->mirror = the_A;	/* make mirror a self pointer */
    
    the_A->tpole->mirror = the_A->bpole;
    the_A->bpole->mirror = the_A->tpole;
    
    trap = V_tlTrap(the_A->tpole);
    
    T_te(trap)->mirror = T_be(trap);
    T_be(trap)->mirror = T_te(trap);
    
    return(the_A);
}


/*************************************************************
  Function:  Arrange_Init_Double()
  Input:     <none>
  Output:    pointer to a newly initialized arrangment
  Effects:   <none>
  
  Description:  This routine creates two separate arrangments
  and then 'glues' them together using the mirror pointers
  so that they behave topologically as a sphere.
  
  The current initialization creates two vertices as the
  top and bottom poles, and two edges connecting them as
  the left and right horizon.  The entire hemisphere is
  a single trap bounded from above and below by the edge,
  and having the two poles as the right and left vertices.
  
  The hemispheres are glued so that the top poles are
  glued to each other, as are the bottom poles, and each
  horizon is glued to the 'opposite' horizon of the mirror
  hemisphere.
  
  **************************************************************/
Arrange*	Arrange_Init_Double()
{
    Arrange	*the_A;
    
    the_A = Arrange_Init_Single();
    the_A->mirror = Arrange_Init_Single();
    the_A->mirror->mirror = the_A;
    
    /* set up the rest of the pole/horizon mirrors correctly */
    the_A->tpole->mirror = the_A->mirror->tpole;
    the_A->mirror->tpole->mirror = the_A->tpole;
    the_A->bpole->mirror = the_A->mirror->bpole;
    the_A->mirror->bpole->mirror = the_A->bpole;
    
    T_be(V_tlTrap(the_A->tpole))->mirror =
	T_te(V_tlTrap(the_A->mirror->tpole));
    E_mirror(T_be(V_tlTrap(the_A->tpole)))->mirror =
	T_be(V_tlTrap(the_A->tpole));
    T_te(V_tlTrap(the_A->tpole))->mirror =
	T_be(V_tlTrap(the_A->mirror->tpole));
    E_mirror(T_te(V_tlTrap(the_A->tpole)))->mirror =
	T_te(V_tlTrap(the_A->tpole));
    
    return(the_A);
}



/*************************************************************
  Function: Arrange_Free_Single(the_A)
  Input:  arrangement
  Output: <none>
  Effects: frees structure and decendants
  
  Description:  Frees up all the dynamic memory, and then
  frees the Arrange structure itself.  This does NOT do
  anything to the mirror arrangement.
  
  **************************************************************/
Arrange_Free_Single(the_A)
  Arrange*	the_A;
{
    List	elt;

    /*-------------------------------------------------------
      the creation of separate Geom structures had
      complicated the simplicity of the dheap scheme.
      -------------------------------------------------------*/
    for (elt = the_A->vgeom; elt;  elt=NXT(elt))
	FreeVertGeom(GET(elt,VertGeom));
    delete_list(the_A->vgeom);
    for (elt = the_A->egeom; elt;  elt=NXT(elt))
	FreeEdgeGeom(GET(elt,EdgeGeom));
    delete_list(the_A->egeom);

    Dheap_Freeall(the_A->Dheap_verts);
    Dheap_Freeall(the_A->Dheap_tverts);
    Dheap_Freeall(the_A->Dheap_edges);
    Dheap_Freeall(the_A->Dheap_traps);
    Dheap_Freeall(the_A->Dheap_hist);
    Dheap_Freeall(the_A->Dheap_ptree);
    FREE(the_A);
}


/*************************************************************
  Function: Arrange_Free(the_A)
  Input:  arrangement
  Output: <none>
  Effects: frees structure and decendants (inlcuding mirror)
  
  Description:  Frees up all the dynamic memory, and then
  frees the Arrange structure itself.  If mirror exists,
  it is also freeed.
  
  **************************************************************/
void Arrange_Free(the_A)
  Arrange*	the_A;
{
    if (the_A->mirror && (the_A->mirror != the_A))
	Arrange_Free_Single(the_A->mirror);
    
    Arrange_Free_Single(the_A);
}


/*************************************************************
  Function: Arrange_PointLocation(the_A,p)
  Input: arrangement & VertGeom p
  Output: ArrangeCell with respect to query P
  Effects: <none>
  
  Description:  This uses the History to locate a trap
  containing the query point, and then afterwards checks
  the boundary verts/edges explicitly.
  
  Uses a precision error in checking boundaries.
  
  **************************************************************/
ArrangeCell Arrange_PointLocation(the_A,p)
  Arrange	*the_A;
  struct VertGeom*	p;
{
    TrapHist		*t_hist;
    int			test;
    ArrangeCell		result;
    TrapVert		*v1,*v2;
    TrapEdge		*e1,*e2;
    
    t_hist = the_A->hist_root;
    
    while(t_hist->nodetype != leaf) {
	switch (t_hist->nodetype) {
	  case threadtest:
	    test = Thread_Test(p,t_hist->data.v);
	    break;
	    
	  case edgetest:
	    test = Edge_Test(p,t_hist->data.e);
	    break;
	}
	
	t_hist = t_hist->ans[test];
    }
    
    /*---- save reference trap ----*/
    result.ref_trap = t_hist->trap;
    
    
    /*---- Now, check the boundaries ----*/
    v1 = T_lvert(t_hist->trap);
    v2 = T_rvert(t_hist->trap);

    /*---- check a horizon edge first ---*/
    if (ON_HORIZON_EDGE(T_te(t_hist->trap))) {
	e1 = T_te(t_hist->trap);
	e2 = T_be(t_hist->trap);
    } else {
	e1 = T_be(t_hist->trap);
	e2 = T_te(t_hist->trap);
    }

    if (VertVertIncedent(p, V_direction(v1))) {
	result.t = vert;
	result.d.vert = v1;
    } else if (VertVertIncedent(p, V_direction(v2))) {
	result.t = vert;
	result.d.vert = v2;
    } else if (VertEdgeIncedent(p, E_normal(e1))) {
	result.t = edge;
	result.d.edge = e1;
    } else if (VertEdgeIncedent(p, E_normal(e2))) {
	result.t = edge;
	result.d.edge = e2;
    } else {
	result.t = trap;
	result.d.trap = t_hist->trap;
    }

    return(result);
}



/*************************************************************
  Function: Find_LeftNeighThreadVert(the_A,edge,vgeom,right_of)
  Input:
      the_A -- a valid arrangement
      edge -- an edge
      vgeom -- VertGeom for source vertex
      right_of -- a ThreadVert on edge, used as hint (see below)
  Output:
      ThreadVert on edge lying immediately left of where
      a thread through 'vgeom' would hit edge.
  Effects: <none>
  
  Description:
  
  Find_LeftNeighThreadVert -- given a vertgeom and an edge, this function
  returns the ThreadVert on the edge which is the nearest from the left
  to the thread defined by the vertgeom. This function is used to find
  where to insert a New Vert or ThreadVert.
  
  'right_of' if given, is a shortcut to a ThreadVert on the edge that
  comes with a guarantee that the place were looking for is to the
  right_of this ThreadVert.  This way, instead of starting the search
  from the far left, we can start from here and be assured that we
  haven't missed the correct location.
  
  The return value is a pointer to the lNeigh of the vert in the list,
  or is NULL if the vert is between the edge's lvert and lThrVert, or
  if the edge has no threadverts.
  **************************************************************/
TrapThreadVert* Find_LeftNeighThreadVert(the_A,edge,vgeom,right_of)
  Arrange		*the_A;
  TrapEdge		*edge;
  VertGeom*		vgeom;
  TrapThreadVert	*right_of;	/* extra help -- may be NULL */
{
    TrapThreadVert	*lneigh;
    
    lneigh = ((right_of) ?
	      /*---- if given, start search at hint ----*/
	      right_of :
	      
	      /*---- else no hint...so check lThrVert... ----*/
	      (((E_lThrVert(edge)) &&
		(Thread_Test(vgeom,
			     V_direction(TV_source(E_lThrVert(edge)))) ==
		 ANS_RIGHT)) ?
	       
	       /*---- start search at lThrVert ----*/
	       E_lThrVert(edge) :
	       
	       /*---- stop search ----*/
	       NULL));
    
    
    if (lneigh) {	/* not done yet */
	while (TV_rNeigh(lneigh) &&
	       (Thread_Test(vgeom,
			    V_direction(TV_source(TV_rNeigh(lneigh)))) ==
		ANS_RIGHT)) {
	    /*---- find correct spot ----*/
	    lneigh = TV_rNeigh(lneigh);
	}
    }
    
    return(lneigh);
}




/*************************************************************
  Function:  Add_ThreadVert(the_A,edge,vert,right_of)
  Input:
      the_A -- a valid arrangement
      edge -- an edge
      vert  -- source vertex
      right_of -- a ThreadVert on edge, used as hint (see below)
  Output:
      new ThreadVert on 'edge' based on 'vec'
  Effects:  only inserts new ThreadVert in edge's list
  
  Description:
  
  Add_ThreadVert --  creates a new TrapThreadVert based on vert,
  inserts it into the proper place in edge's list.
  
  'right_of' if given, is a shortcut to a ThreadVert on the edge that
  comes with a guarantee that the place were looking for is to the
  right_of this ThreadVert.  This way, instead of starting the search
  from the far left, we can start from here and be assured that we
  haven't missed the correct location.
  
  **************************************************************/
TrapThreadVert* Add_ThreadVert(the_A, edge, vert, right_of)
  Arrange		*the_A;
  TrapEdge		*edge;
  TrapVert		*vert;
  TrapThreadVert	*right_of;	/* extra help -- may be NULL */
{
    TrapThreadVert	*new_Tvert;
    
    new_Tvert = Make_TrapThreadVert(the_A);
    new_Tvert->source = vert;
    
    new_Tvert->lNeigh =
	Find_LeftNeighThreadVert(the_A,edge,V_direction(vert),right_of);
    
    if (TV_lNeigh(new_Tvert)) {
	new_Tvert->rNeigh = TV_rNeigh(TV_lNeigh(new_Tvert));
	TV_lNeigh(new_Tvert)->rNeigh = new_Tvert;
    } else {
	new_Tvert->rNeigh = E_lThrVert(edge);
	edge->lThrVert = new_Tvert;
    }
    
    if (TV_rNeigh(new_Tvert)) {
	TV_rNeigh(new_Tvert)->lNeigh = new_Tvert;
    } else {
	edge->rThrVert = new_Tvert;
    }
    
    return(new_Tvert);
}




/*************************************************************
  Function:  Remove_ThreadVert(the_A,tv)
  Input:
      the_A -- a valid arrangement
      tv -- the TrapThreadVert to delete
  Output: <none>
  Effects:  completely removes 'tv' from its edge, and updates
  edge, but does NOT destroy 'tv' itself, and does
  NOT effect any of the neighboring traps.
  
  **************************************************************/
Remove_ThreadVert(the_A,tv)
  Arrange		*the_A;
  TrapThreadVert	*tv;
{
    
    if (TV_lNeigh(tv)) {
	TV_lNeigh(tv)->rNeigh = TV_rNeigh(tv);
    } else {
	TV_edge(tv)->lThrVert = TV_rNeigh(tv);
    }
    
    if (TV_rNeigh(tv)) {
	TV_rNeigh(tv)->lNeigh = TV_lNeigh(tv);
    } else {
	TV_edge(tv)->rThrVert = TV_lNeigh(tv);
    }
    
}




/*************************************************************
  Function: Extend_Thread(the_A,new_vert,orig_trap,vert_type,r_edge)
  Input:
      the_A -- a valid arrangement
      new_vert -- a newly created vertex (possibly on an edge)
      orig_trap -- trap incident to new vertex
      vert_type -- top/mid/bot  (never com!)
      r_edge -- if top/bot, this gives "new" part of edge incident
                  to new_vert
  Output: <none>
  Effects: splits orig_trap in two by extending thread from new_vert
  
  Description:
  
  Extend_Thread -- given a newly created vertex (possibly on an
  edge), and a specific trap, this function will extend the thread
  through that vertex all the way across the trap.  In doing this,
  it will also entirely update all involved data structures to
  account for the split of the trap into two traps.
  
  The only assumption made is that any trap that is bounded by a
  a split edge on some side references the "original" part of the split
  edge as the bounding edge. In the case where the vertex lies on a
  split edge, then the parameter 'r_edge' is a reference to the new
  edge that was created.
  
  Also, note that for a vertex on an edge, this function must
  be called twice, once to extend into the trap above, and once
  for the trap below.
  **************************************************************/
Extend_Thread(the_A,new_vert,orig_trap,vert_type,r_edge)
  Arrange		*the_A;
  TrapVert		*new_vert;
  Trap			*orig_trap;
  enum TrapVertType	vert_type;	/* top/mid/bot (never com) */
  TrapEdge		*r_edge;	/* NULL if vert_type=='mid' */
{
    TrapHist		*left_hist;
    Trap		*new_trap;
    
    
    new_trap = Make_Trap(the_A);
    
    /* We will reuse orig_trap as the left trap of the two new
       traps, and we will use new_trap as the resulting right trap.
       Must be VERY careful to change all references to trap to
       point to new_trap if referring to resulting right trap. */
    
    if (vert_type!=top) {
	new_vert->tlTrap = orig_trap;
	new_vert->trTrap = new_trap;
    }
    
    if (vert_type!=bot) {
	new_vert->blTrap = orig_trap;
	new_vert->brTrap = new_trap;
    }
    
    /*---- insert new ThreadVert's ----*/
    if (vert_type!=top) {
	new_vert->topTV =
	    Add_ThreadVert(the_A, T_te(orig_trap), new_vert,
			   T_tlv(orig_trap));
    }
    
    if (vert_type!=bot) {
	new_vert->botTV =
	    Add_ThreadVert(the_A, T_be(orig_trap), new_vert,
			   T_blv(orig_trap));
    }
    
    /*---- Fix up trap and new trap to reflect changes ----*/
    new_trap->lvert = new_vert;
    new_trap->lvtype = vert_type;
    new_trap->rvert = T_rvert(orig_trap);
    new_trap->rvtype = T_rvtype(orig_trap);
    orig_trap->rvert = new_vert;
    orig_trap->rvtype = vert_type;
    
    
    /*---- Fix up top/bot edge for new/old traps ----*/
    new_trap->te = (vert_type==top) ? r_edge : T_te(orig_trap);
    new_trap->be = (vert_type==bot) ? r_edge : T_be(orig_trap);
    
    
    /*---- Fix up edge's tl/tr/bl/br traps when necessary ----*/
    switch (vert_type) {
      case top:
	T_te(new_trap)->brTrap = E_brTrap(T_te(orig_trap));
	T_te(orig_trap)->brTrap = orig_trap;
	T_te(new_trap)->blTrap = new_trap;
	break;

      case bot:
	T_be(new_trap)->trTrap = E_trTrap(T_be(orig_trap));
	T_be(orig_trap)->trTrap = orig_trap;
	T_be(new_trap)->tlTrap = new_trap;
	break;
    }

    
    /*---- Update ThreadVert/Edge references to traps ----*/
    if (V_ON_TOP(T_rvtype(new_trap))) {
	T_te(new_trap)->brTrap = new_trap;
    } else {
	T_rvert(new_trap)->tlTrap = new_trap;
    }

    if (V_ON_BOT(T_rvtype(new_trap))) {
	T_be(new_trap)->trTrap = new_trap;
    } else {
	T_rvert(new_trap)->blTrap = new_trap;
    }


    /*---- Update right Vert's possible references to trap ----*/
    if (V_tlTrap(T_rvert(new_trap)) == orig_trap)
	T_rvert(new_trap)->tlTrap = new_trap;
    if (V_blTrap(T_rvert(new_trap)) == orig_trap)
	T_rvert(new_trap)->blTrap = new_trap;


    /*---- Update History nodes ----*/
    left_hist = Make_TrapHist(the_A);
    new_trap->hist = Make_TrapHist(the_A);
    
    /*** don't forget -- can't copy whole structure unless careful
      not to ruin the 'link' -- to be safe, I copy fields ***/
    left_hist->trap = orig_trap;
    T_hist(new_trap)->trap = new_trap;
    
    T_hist(orig_trap)->nodetype = threadtest;
    T_hist(orig_trap)->data.v = V_direction(new_vert);
    T_hist(orig_trap)->trap = NULL;
    T_hist(orig_trap)->ans[ANS_RIGHT] = T_hist(new_trap);
    T_hist(orig_trap)->ans[ANS_LEFT] = left_hist;
    T_hist(orig_trap) = left_hist;
}


/*************************************************************
  Function: Split_Edge_DoIt(the_A,new_vec,old_edge,hint)
  Input:
      the_A -- a valid arrangement
      new_vec -- a VertGeom lying on old_edge
      old_edge -- the edge to be split
      hint -- ThreadVert guaranteed to be left of new_vec
  Output:
      new TrapVert that was inserted in splitting edge
  Effects: splits edge into two pieces, updates entire Arrange
  structure to reflect split, and new threads.
  
  Description:
  
  Split_Edge_DoIt -- given a VertGeom which should lie on an edge,
  this function will create a new vertex, and then breaks the
  edge into two edges with the new vertex at the breakpoint.
  Following this, it immediately calls the function necessary
  to extend threads both upward and downward from the new vertex.
  
  By the end of this routine, all data structures will be updated to
  reflect the changes in the arrangement.
  
  If the old_edge is part of the horizon, this procedure does NOT
  split the mirror edge.
  
  The function returns the new part of the edge.
  (to the right of the split)
  **************************************************************/
TrapEdge* Split_Edge_DoIt(the_A,new_vec,old_edge,hint)
  Arrange		*the_A;
  VertGeom*		new_vec;
  TrapEdge		*old_edge;
  TrapThreadVert	*hint;
{
    TrapVert		*new_vert;
    TrapEdge		*new_edge;
    TrapThreadVert	*breakpoint,*walk;
    Trap		*trap_above, *trap_below;
    
    new_vert = Make_TrapVert(the_A);
    new_vert->direction = CopyVertGeom(new_vec);
    VGEOM(new_vert->direction);
    
    /*---- First, must find the place to insert the new vertex. ----*/
    breakpoint =
	Find_LeftNeighThreadVert(the_A,old_edge,V_direction(new_vert),hint);
    
    /*-- walk backwards through ThreadVert's to find trap above/below --*/
    if (breakpoint) {
	for (walk=TV_lNeigh(breakpoint);
	     (walk && (TV_from_above(walk) == TV_from_above(breakpoint)));
	     walk = TV_lNeigh(walk));
    }
    
    if (breakpoint) {
	trap_above = (TV_from_above(breakpoint) ? 
		      TV_rTrap(breakpoint) :
		      (walk ? TV_rTrap(walk) : E_tlTrap(old_edge)));
	
	trap_below = (TV_from_above(breakpoint) ?
		      (walk ? TV_rTrap(walk) : E_blTrap(old_edge)) :
		      TV_rTrap(breakpoint));
    } else {
	trap_above = E_tlTrap(old_edge);
	trap_below = E_blTrap(old_edge);
    }    
    
    
    /*---- Now we are at the correct spot to break the edge ----*/
    new_edge = Make_TrapEdge(the_A);
    new_edge->mirror = E_mirror(old_edge);
    new_edge->normal = CopyEdgeGeom(E_normal(old_edge));
    EGEOM(new_edge->normal);
    new_edge->polys_on = E_polys_on(old_edge);
    new_edge->polys_below = E_polys_below(old_edge);
    new_edge->polys_above = E_polys_above(old_edge);
    
    
    if (!breakpoint) {
	/*---- we're between left vert and left threadvert ----*/
	new_edge->lThrVert = E_lThrVert(old_edge);
	new_edge->rThrVert = E_rThrVert(old_edge);
	old_edge->lThrVert = old_edge->rThrVert = NULL;
    } else if (TV_rNeigh(breakpoint) == NULL) {
	/*---- we're between right threadvert and right vert ----*/
	new_edge->lThrVert = new_edge->rThrVert = NULL;
    } else {
	/*---- we're between two different threadverts ----*/
	new_edge->lThrVert = TV_rNeigh(breakpoint);
	new_edge->rThrVert = E_rThrVert(old_edge);
	old_edge->rThrVert = breakpoint;
	
	E_lThrVert(new_edge)->lNeigh = NULL;
	E_rThrVert(old_edge)->rNeigh = NULL;
    }
    
    /*---- Now, must walk the new_edge fixing up threadvert->edge, ----*/
    /*---- and fixed up Trap->te and Trap->be where applicable     ----*/
    for (walk=E_lThrVert(new_edge); walk; walk=TV_rNeigh(walk)) {
	if (TV_from_above(walk)) {
	    TV_rTrap(walk)->be = new_edge;
	} else {
	    TV_rTrap(walk)->te = new_edge;
	}
    }
    
    
    if (trap_above) {
	Extend_Thread(the_A,new_vert,trap_above,bot,new_edge);
    }
    
    if (trap_below) {
	Extend_Thread(the_A,new_vert,trap_below,top,new_edge);
    }
    
    return(new_edge);
}


/*************************************************************
  Function: Split_Edge(the_A,new_vec,old_edge,hint)
  Input:
      the_A -- a valid arrangement
      new_vec -- a VertGeom lying on old_edge
      old_edge -- the edge to be split
      hint -- ThreadVert guaranteed to be left of new_vec
  Output:
      new TrapVert that was inserted in splitting edge
  Effects: splits edge into two pieces, updates entire Arrange
  structure to reflect split, and new threads.
  
  Description:
  
  Split_Edge -- given a vector which should lie on an edge,
  this function will create a new vertex, and then breaks the
  edge into two edges with the new vertex at the breakpoint.
  Following this, it immediately calls the function necessary
  to extend threads both upward and downward from the new vertex.
  
  By the end of this routine, all data structures will be updated to
  reflect the changes in the arrangement.
  
  If the old_edge is part of the horizon, this procedure ALSO
  splits the mirror edge.
  
  The function returns the new part of the edge.
  (to the right of the split)
  **************************************************************/
TrapEdge* Split_Edge(the_A,new_vec,old_edge,hint)
  Arrange		*the_A;
  VertGeom*		new_vec;
  TrapEdge		*old_edge;
  TrapThreadVert	*hint;
{
    TrapEdge	*new_edge1;
    TrapEdge	*new_edge2;
    VertGeom*	reverse_vec;

    new_edge1 = Split_Edge_DoIt(the_A,new_vec,old_edge,hint);
    
    if (E_mirror(old_edge)) {
	/*---- special case...must split mirror edge too. ----*/
	reverse_vec = CopyVertGeom(new_vec);
	NegateVertGeom(reverse_vec);
	new_edge2 = Split_Edge_DoIt(the_A->mirror, reverse_vec,
				    E_mirror(old_edge),NULL);
	FreeVertGeom(reverse_vec);
	
	E_rvert(old_edge)->mirror = E_rvert(E_mirror(old_edge));
	E_rvert(E_mirror(old_edge))->mirror = E_rvert(old_edge);
	E_mirror(new_edge1)->mirror = new_edge1;
	E_mirror(new_edge2)->mirror = new_edge2;
    }
    
    return(new_edge1);
}



/*************************************************************
  Function:  PolyTree_AppendList(the_A,polys,old_tree)
  Input:
      the_A -- a valid arrangement
      polys -- a list of (PolySource)'s to add to PolyTree
      old_tree -- an existing PolyTree
  Output:
      a new PolyTree with new list appended to old_tree
  Effects: <none>
  
  Description:  This provides a simple tool which will take a
  List of (PolySource) and append the items to the beginning
  of a PolyTree.
  
  However, this routine does NOT take the time to check whether
  a PolySource is already present, rather it adds all members
  of the list blindly.
  
  **************************************************************/
PolyTree* PolyTree_AppendList(the_A,polys,old_tree)
  Arrange	*the_A;
  List		polys;
  PolyTree	*old_tree;
{
    List	elt;
    PolyTree	*new_polys;
    PolyTree	*temp;
    
    temp = old_tree;
    for (elt=polys; elt; elt=NXT(elt)) {
	new_polys = Make_PTree(the_A);
	new_polys->next = temp;
	new_polys->source = GET(elt,PolySource);
	temp = new_polys;
    }
    
    return(temp);
}




/*************************************************************
  Function: Arrange_AddVertex(the_A,direction,polys_on,location)
  Input:
      the_A -- a valid arrangement
      direction -- VertGeom representing new vertex
      polys_on -- polygons entered *only* when on new vertex
      location -- result of previous point location query.
  Output:
      new TrapVert based on direction
  Effects: updates entire Arrange structure to reflect change
  
  Description:
  
  Arrange_AddVertex --  This is the high level routine to add a
  (possibly) new vertex to the arrangement.  It does everything
  necessary depending on what type of cell (trap/edge/vert) the
  new vert lies on top of.

  This routine will also handle any details necessary because of
  any mirrors.
  
  **************************************************************/
TrapVert* Arrange_AddVertex(the_A,direction,polys_on,location)
  Arrange		*the_A;
  struct VertGeom*	direction;
  List			polys_on;	/* List of (PolySource)   */
  ArrangeCell		location;	/* result of point location */
{
    TrapVert		*t_vert;
    VertGeom*		snapped;
    
    switch (location.t) {
      case vert:
	t_vert = location.d.vert;
	break;
	
      case edge:
	/*---- force vertex to lie on edge ----*/
	snapped = CopyVertGeom(direction);
	SnapVertToEdge(snapped, E_normal(location.d.edge),
		       V_direction(E_lvert(location.d.edge)),
		       V_direction(E_rvert(location.d.edge)));
	
	Split_Edge(the_A, snapped, location.d.edge,
		   ((location.d.edge == T_te(location.ref_trap)) ?
		    T_tlv(location.ref_trap) : T_blv(location.ref_trap)));

	FreeVertGeom(snapped);
	
	t_vert = E_rvert(location.d.edge);
	break;
	
      case trap:
	t_vert = Make_TrapVert(the_A);
	t_vert->direction = CopyVertGeom(direction);
	VGEOM(t_vert->direction);
	Extend_Thread(the_A,t_vert,location.d.trap,mid,NULL);
	
    }
    
    t_vert->polys = PolyTree_AppendList(the_A,polys_on,V_polys(t_vert));
    
    return(t_vert);
}



/*************************************************************
  Function: Trap_LeaveVertex(the_A,v,norm,toward_left)
  Input:
      the_A -- a valid arrangement
      v -- a vertex in arrangement
      norm -- EdgeGeom representing an edge passing through v
      toward_left -- which way do we walk along edge from v?
  Output:
      the Trap that we hit if walking from v as directed
  Effects: <none>
  
  Description:
  
  Trap_LeaveVertex --  Given the geometry of an edge, and a vertex
  which it passes through, this procedure returns the trap that
  the edge would pass through to the right (or left) of the
  vertex.   This can be used when navigating, or adding a new edge.
  **************************************************************/
Trap* Trap_LeaveVertex(the_A, v, norm, toward_left)
  Arrange	*the_A;
  TrapVert	*v;
  EdgeGeom*	norm;
  Boolean	toward_left;
{
    Trap	*t;
    
    t=NULL;	/* default in case walking across horizon */
    
    if (toward_left) {
	if (V_tlTrap(v)) {
	    for (t=V_tlTrap(v);
		 ((E_rvert(T_be(t))==v) &&
		  (!EdgeEdge_Above(V_direction(v),norm,
				   E_normal(T_be(t)),FALSE)));
		 t = E_brTrap(T_be(t)));
	} else if (OnHorizonEdge(norm)) {
	    /*--- on the left horizon moving towards bottom pole ---*/
	    t = T_bln(V_brTrap(v));
	}
    } else {
	if (V_brTrap(v)) {
	    for (t=V_brTrap(v);
		 ((E_lvert(T_te(t))==v) &&
		  (!EdgeEdge_Above(V_direction(v),E_normal(T_te(t)),
				   norm,TRUE)));
		 t = E_tlTrap(T_te(t)));
	} else if (OnHorizonEdge(norm)) {
	    /*--- on the right horizon moving towards top pole ---*/
	    t = T_trn(V_tlTrap(v));
	}
    }
    
    return(t);
}


/*************************************************************
  Function:  AddEdge_Phase1(<stuff>)
  Input:
      the_A --- the arrangement
      loop__enter_vertex ---  starting vertex
      from_left ---  moving left->right or right->left
      true_goal ---  VertGeom of final goal of new edge
      goal_normal --- EdgeGeom describing ideal new edge
  Output:
      return_normal --- recalculated normal based on the fact
            that we're starting at loop__enter_vertex
      return_trap ---  trap which we walk through when leaving
            from enter_vertex
  Effects: <none>	    

  Description:
     During the tracing of a new edge, this routine is called
     when a vertex is passed to recalculate a more precise
     direction of travel incase we have gone off course a bit.
     Also, it determines which trap we walk through when leaving
     along this new path.
  **************************************************************/
AddEdge_Phase1(return_normal, return_trap,
	       the_A, loop__enter_vertex, from_left,
	       true_goal, goal_normal)
  
  /*--- return variables ---*/
  EdgeGeom*	*return_normal;
  Trap		**return_trap;
  
  /*--- input variables ---*/
  Arrange	*the_A;
  TrapVert	*loop__enter_vertex;
  Boolean	from_left;
  VertGeom*	true_goal;
  EdgeGeom*	goal_normal;
{
    
    TrapEdge	*loop__edge;
    EdgeGeom*	loop__new_normal;
    Trap	*trap_walk;
    
#ifdef ARRANGE_DEBUG
    if (arrange_verbose)
	printf("Starting phase one...\n");
#endif
    
    /*****************************************************
     ******             phase one             ************
     *****************************************************/
    /*  Calculate ideal normal to trace.   This may get changed
	later if we pass 'close' to a vertex that we stop at */

    if (loop__enter_vertex->mirror &&
	(VertVertIncedent(V_direction(loop__enter_vertex->mirror),
			  true_goal))) {
	/*--- antipodes ---*/
	loop__new_normal = CopyEdgeGeom(goal_normal);
    } else {
	loop__new_normal =
	    EdgeThroughVerts(V_direction(loop__enter_vertex), true_goal,
			     goal_normal);
    }
    
    trap_walk = Trap_LeaveVertex(the_A,loop__enter_vertex,
				 loop__new_normal, (!from_left));
    
    /*--- set return variables ---*/
    (*return_normal) = loop__new_normal;
    (*return_trap) = trap_walk;
}



/*************************************************************
  Function:  AddEdge_Phase2(<stuff>)
  Input:
      the_A  ---  the arrangement
      from_left --- moving left->right or right->left
      loop__enter_vertex --- vertex to start walking from
      trap_walk ---  trap around enter_vertex which we leave through
      loop__new_normal --- EdgeGeom of edge path to trace
      true_goal ---   VertGeom of final endpoint of edge
  Output:
      return_stack --- list of traps we pass through before hitting
          another vertex.  (sorted from left->right)
      return_opp ---   opp_vertex we reached at end of phase
  Effects:  May add goal_vertex or split edge to create new vertex
    if path crosses an edge before reaching goal
  
  Description:
    While inserting a new edge, we proceed by pahses, where each phase
    represents adding one "section" of the complete edge based on
    where it is broken up into pieces.   This routine starts at
    loop__enter_vertex and leaves through trap_walk along the path
    given towards the goal.  This will go from trap to trap, keeping
    a record of the list, until either an edge/vert is hit or the goal
    is reached.  If the goal is reached, then a new vertex is inserted
    if necessary.  If an edge is reached, then the edge is split by
    a new vertex at the intersection point.

  **************************************************************/
AddEdge_Phase2(return_stack, return_opp, the_A, from_left,
	       loop__enter_vertex, trap_walk, loop__new_normal,
	       true_goal)
  
  /*--- return variables ---*/
  List		*return_stack;
  TrapVert	**return_opp;

  /*--- input variables ---*/
  Arrange	*the_A;
  Boolean	from_left;
  TrapVert	*loop__enter_vertex;
  Trap		*trap_walk;
  EdgeGeom*	loop__new_normal;
  VertGeom*	true_goal;

{
    enum ExitCase {goal=1, oppvert, topneigh, botneigh,
		       topedge, botedge};
    
    List	loop__trap_stack;
    TrapVert	*loop__opp_vertex;
    
    VertGeom*		intersect_vec;
    VertGeom*		opp_vector;
    enum ExitCase	phase2_exit;
    Boolean		through_opp;
    Boolean		check_above;
    TrapEdge		*edge_of_interest;
    ArrangeCell		temp_loc;

    /*****************************************************
     ******             phase two             ************
     *****************************************************/
    
#ifdef ARRANGE_DEBUG
    if (arrange_verbose)
	printf("Starting phase two...\n");
#endif

    loop__trap_stack = NULL;
    loop__opp_vertex = NULL;
    while (!loop__opp_vertex) {		/* continue the walk */
	
#ifdef ARRANGE_DEBUG
	if (arrange_verbose)
	    printf(" starting phase two loop...\n");
#endif	


	opp_vector = (from_left ? V_direction(T_rvert(trap_walk)) :
		      V_direction(T_lvert(trap_walk)));
	
	
	/*** must figure out if our walk passes through
	  opp_vertex, or through thread, or through top/bot
	  edge, or stops inside the trap at the goal ***/
	
	/*----
	  NOTE:  the following relies heavily on the fact that
	  "equal" gets grouped with ANS_RIGHT by Thread_Test
	  ----*/
	
	
	/*---- see whether path goes "through" opp vertex ----*/
	through_opp =
	    VertEdgeIncedent(opp_vector, loop__new_normal) &&
		(!OnPoleVert(opp_vector) ||
		 (from_left != VertOnLeftHorizon(opp_vector)));
	
	/*---- see whether we go above/below opp vertex ----*/
	check_above = 
	    (Edge_Test(opp_vector, loop__new_normal) == ANS_BELOW);

	edge_of_interest =
	    (check_above ? T_te(trap_walk) : T_be(trap_walk));
	
	/*--- calculate possible intersection with edge ---*/
	if ((loop__enter_vertex ==
	     (from_left ?
	      T_lvert(trap_walk) :
	      T_rvert(trap_walk)) &&
	     (loop__enter_vertex ==
	      (from_left ?
	       E_lvert(edge_of_interest) :
	       E_rvert(edge_of_interest))))) {
	    /*--- special case...starting at vert ---*/
	    intersect_vec =
		CopyVertGeom(V_direction(loop__enter_vertex));
	} else {
	    intersect_vec =
		EdgeEdgeIntersect(loop__new_normal,
				  E_normal(edge_of_interest),
				  (from_left ?
				   V_direction(loop__enter_vertex) :
				   true_goal),
				  (from_left ?
				   true_goal :
				   V_direction(loop__enter_vertex)),
				  V_direction(E_lvert(edge_of_interest)),
				  V_direction(E_rvert(edge_of_interest))
				  );

	    /*---- be careful around horizon ----*/
	    if (OnHorizonVert(intersect_vec) &&
		(from_left == VertOnLeftHorizon(intersect_vec)))
		/*---- gave us wrong intersection point ----*/
		/*---- we want the one in front of us   ----*/
		NegateVertGeom(intersect_vec);
	}


	if (
	    /*--- goal is after thread ----*/
	    ((from_left ==
	      (Thread_Test(true_goal, opp_vector) == ANS_RIGHT)) ||
	     (OnHorizonEdge(loop__new_normal))) &&

	    /*---- && if don't intersect edge in trap: ----*/

	    /*--- i.e. (a) starting at edge endpoint ---*/
	    ((loop__enter_vertex == (from_left ?
				     E_lvert(edge_of_interest) :
				     E_rvert(edge_of_interest))) ||

	     /*--- or (b) edge intersection not inside trap ---*/
	     (Thread_Test(V_direction(T_lvert(trap_walk)),
			  intersect_vec) == ANS_RIGHT) ||
	     (Thread_Test(intersect_vec,
			  V_direction(T_rvert(trap_walk))) == ANS_RIGHT))
	    ) {

	    /*---- we hit thread first ----*/
	    phase2_exit =
		(through_opp ? oppvert :
		 (check_above ? topneigh : botneigh));

	} else if ((check_above ==
		    (Edge_Test(true_goal,
			       E_normal(edge_of_interest)) == ANS_BELOW)) &&
		   (!VertVertIncedent(true_goal,intersect_vec))) {
	    
	    /*---- we hit true_goal first ----*/
	    phase2_exit =
		((through_opp &&
		  VertVertIncedent(true_goal,opp_vector)) ?
		 oppvert : goal);
	} else {
	    /*---- we hit edge first ----*/
	    phase2_exit =
		(through_opp ? oppvert :
		 (check_above ? topedge : botedge));
	}
	
	/** now we know exit condition, and if we pass through
	  an edge, intersect_vec is the intersection point ***/

	switch (phase2_exit) {
	  case oppvert:
	    loop__opp_vertex = (from_left ?
				T_rvert(trap_walk) :
				T_lvert(trap_walk));
	    loop__trap_stack =
		push(loop__trap_stack,trap_walk);
	    break;
	    
	  case topneigh:
	  case botneigh:
	    loop__trap_stack = 
		push(loop__trap_stack,trap_walk);
	    trap_walk = (from_left ?
			 ((phase2_exit==topneigh) ?
			  T_trn(trap_walk) : T_brn(trap_walk)) :
			 ((phase2_exit==topneigh) ?
			  T_tln(trap_walk) : T_bln(trap_walk)));
	    break;
	    
	  case goal:
	    temp_loc.t = trap;
	    temp_loc.d.trap = temp_loc.ref_trap = trap_walk;
	    loop__opp_vertex =
		Arrange_AddVertex(the_A,true_goal,NULL,temp_loc);
	    trap_walk = (from_left ?
			 V_tlTrap(loop__opp_vertex) :
			 V_brTrap(loop__opp_vertex));
	    loop__trap_stack =
		push(loop__trap_stack,trap_walk);
	    break;
	    
	  case topedge:
	  case botedge:
	    if (phase2_exit == topedge) {
		Split_Edge(the_A,intersect_vec,T_te(trap_walk),
			   T_tlv(trap_walk));
		loop__opp_vertex = E_rvert(T_te(trap_walk));
	    } else {
		Split_Edge(the_A,intersect_vec,T_be(trap_walk),
			   T_blv(trap_walk));
		loop__opp_vertex = E_rvert(T_be(trap_walk));
	    }
	    
	    if (!from_left)
		trap_walk = T_trn(trap_walk);
	    
	    loop__trap_stack =
		push(loop__trap_stack,trap_walk);
	    break;
	}

	FreeVertGeom(intersect_vec);
    }		/* end of phase two loop */
    
    if (from_left)
	loop__trap_stack = reverse(loop__trap_stack,NULL);
    
    /*---- Now loop__trap_stack is left-to-right ----*/
    
    
    /*--- set return variables ---*/
    (*return_stack) = loop__trap_stack;
    (*return_opp) = loop__opp_vertex;
}




/*************************************************************
  Function:  AddEdge_Phase3()
  Input:
      the_A --- the arrangement
      from_left --- moving left->right or right->left
      goal_normal ---  original normal for new edge
      loop__enter_vertex --- first endpoint of new edge
      loop__opp_vertex ---  second endpoint of new edge
      loop__trap_stack ---  list of traps new edge passes through
  Output:
      return_edge --- point to newly created edge
  Effects:
      Adds a new edge into the arrangment and updates the structures.

  Description:
     This routine retraces the path of a new edge between two
  verts already part of the arrangement.  The new edge breaks up
  some traps into two, and this routine adjusts the arrangement to
  completely add this new edge into the arrangment.
  
  **************************************************************/
AddEdge_Phase3(return_edge, the_A, from_left, goal_normal,
	       loop__enter_vertex, loop__opp_vertex, loop__trap_stack)
  
  /*--- return variables ---*/
  TrapEdge	**return_edge;
  
  /*--- input variables ---*/
  Arrange	*the_A;
  Boolean	from_left;
  EdgeGeom*	goal_normal;
  TrapVert	*loop__enter_vertex;
  TrapVert	*loop__opp_vertex;
  List		loop__trap_stack;
  
{
    TrapEdge	*loop__edge;

    List		elt;
    Trap		*trap_walk;
    EdgeGeom*		new_normal;
    Trap		*top_trap;
    Trap		*bot_trap;
    Trap		*partial_trap;
    Boolean		close_top;
    Boolean		close_bot;
    TrapHist		*temp_hist;

    
    /*****************************************************
     ******              phase three          ************
     *****************************************************/
    
#ifdef ARRANGE_DEBUG
    if (arrange_verbose)
	printf("Starting phase three...\n");
#endif

    /* Now that we know both endpoints, recalculate a more
       precise normal to use for splitting up traps */

    if (loop__enter_vertex->mirror == loop__opp_vertex) {
	/*--- antipodes ---*/
	new_normal = CopyEdgeGeom(goal_normal);
    } else {
	new_normal =
	    EdgeThroughVerts(V_direction(loop__enter_vertex),
			     V_direction(loop__opp_vertex),
			     goal_normal);
    }


    /*--- check for degeneracy ---*/
    loop__edge = NULL;
    trap_walk = GET(loop__trap_stack,Trap);

    if ((loop__enter_vertex == 
	 (from_left ? E_lvert(T_te(trap_walk)) :
	  E_rvert(T_te(trap_walk)))) &&
	(loop__opp_vertex == 
	 (from_left ? E_rvert(T_te(trap_walk)) :
	  E_lvert(T_te(trap_walk)))) &&
	(EdgeEdgeIncedent(new_normal,E_normal(T_te(trap_walk))))) {
	loop__edge = T_te(trap_walk);
    } else if ((loop__enter_vertex == 
		(from_left ? E_lvert(T_be(trap_walk)) :
		 E_rvert(T_be(trap_walk)))) &&
	       (loop__opp_vertex == 
		(from_left ? E_rvert(T_be(trap_walk)) :
		 E_lvert(T_be(trap_walk)))) &&
	       (EdgeEdgeIncedent(new_normal,E_normal(T_be(trap_walk))))) {
	loop__edge = T_be(trap_walk);
    }

    if (!loop__edge) {
	
	/*******************************************************************/
	/* delete this part...eventually */
	/*---- be careful around horizon ----*/
	if (OnHorizonEdge(new_normal)) {
	    printf("ARRANGE --- How did this happen???\n");
	}
	/*******************************************************************/
	
	loop__edge = Make_TrapEdge(the_A);
	loop__edge->normal = CopyEdgeGeom(new_normal);
	EGEOM(loop__edge->normal);
	
	/*--- set up the split of the first trap ---*/
	partial_trap = Make_Trap(the_A);
	partial_trap->te = T_te(trap_walk);
	partial_trap->be = loop__edge;
	loop__edge->tlTrap = partial_trap;
	loop__edge->blTrap = trap_walk;
	partial_trap->lvert = (from_left ?
			       loop__enter_vertex : loop__opp_vertex);
	switch (T_lvtype(trap_walk)) {
	  case com:
	  case top:
	    T_te(partial_trap)->blTrap = partial_trap;
	    partial_trap->lvtype = com;
	    break;
	    
	  case bot:
	  case mid:
	    partial_trap->lvtype = bot;
	    T_lvert(partial_trap)->trTrap = partial_trap;
	    trap_walk->lvtype = ((T_lvtype(trap_walk) == bot) ? com: top);
	    break;
	}
	partial_trap->hist = Make_TrapHist(the_A);
	T_hist(partial_trap)->trap = partial_trap;
        
	top_trap = partial_trap;
	bot_trap = NULL;
	
	/*----  Begin walking the edge ----*/
	for (elt=loop__trap_stack;  elt; elt=NXT(elt)) {
	    
#ifdef ARRANGE_DEBUG
	    if (arrange_verbose)
		printf("  starting phase three loop...\n");
#endif
	    
	    trap_walk = GET(elt,Trap);
	    if (!bot_trap) {
		bot_trap = trap_walk;
		bot_trap->te = loop__edge;
	    } else {
		top_trap = trap_walk;
		top_trap->be = loop__edge;
	    }
	    
	    
	    
	    /* must fix up left thread verts and left vert */
	    /* Note: as it happens, the left vert->brTrap and
	       vert->trTrap are already correctly set */
	    
	    /*---- fix up history to reflect breaking trap ----*/
	    temp_hist = T_hist(trap_walk);	/* save old node */
	    trap_walk->hist = Make_TrapHist(the_A);
	    T_hist(trap_walk)->trap = trap_walk;
	    
	    /*-- now fix up old history node to be an edge test --*/
	    temp_hist->nodetype = edgetest;
	    temp_hist->data.e = E_normal(loop__edge);
	    temp_hist->trap = NULL;
	    temp_hist->ans[ANS_ABOVE] = T_hist(top_trap);
	    temp_hist->ans[ANS_BELOW] = T_hist(bot_trap);
	    
	    /*---- which trap(s) to close? ----*/
	    if (T_rvert(trap_walk) != (from_left ? loop__opp_vertex :
				       loop__enter_vertex)) {
		/* normal case...one of the two traps closes,
		   the other becomes a partial trap.  */
		
		switch(T_rvtype(trap_walk)) {
		    /*--- case com: MUST have been end of line. ---*/
		    
		  case top:
		    close_top = TRUE;
		    close_bot = FALSE;
		    break;
		    
		  case bot:
		    close_top = FALSE;
		    close_bot = TRUE;
		    break;
		    
		  case mid:
		    close_top =
			((T_brn(trap_walk) == T_trn(trap_walk)) ?
			 /*---- must redo check ----*/
			 (Edge_Test(V_direction(T_rvert(trap_walk)),
				    new_normal) == ANS_ABOVE) :
			 (GET(NXT(elt),Trap) == T_brn(trap_walk)));
		    close_bot = (!close_top);
		    break;
		}

		
		/*---- fix up right thread verts, if necessary ----*/
		/*----   and edge references, if necessary ----*/
		if (close_top) {
		    if (T_brv(trap_walk)) {
			Remove_ThreadVert(the_A,T_brv(trap_walk));
			T_brv(trap_walk)->rNeigh = NULL;
			T_brv(trap_walk)->lNeigh = E_rThrVert(loop__edge);
			if (E_rThrVert(loop__edge))
			    E_rThrVert(loop__edge)->rNeigh =
				T_brv(trap_walk);
			loop__edge->rThrVert = T_brv(trap_walk);
			if (!E_lThrVert(loop__edge))
			    loop__edge->lThrVert = T_brv(trap_walk);
		    }
		}

		if (close_bot) {
		    if (T_trv(trap_walk)) {
			Remove_ThreadVert(the_A,T_trv(trap_walk));
			T_trv(trap_walk)->rNeigh = NULL;
			T_trv(trap_walk)->lNeigh = E_rThrVert(loop__edge);
			if (E_rThrVert(loop__edge))
			    E_rThrVert(loop__edge)->rNeigh =
				T_trv(trap_walk);
			loop__edge->rThrVert = T_trv(trap_walk);
			if (!E_lThrVert(loop__edge))
			    loop__edge->lThrVert = T_trv(trap_walk);
		    }
		}

		
		/*-- close up right side if partial trap closing --*/
		if (((close_bot) && (partial_trap == bot_trap)) ||
		    ((close_top) && (partial_trap == top_trap))) {
		    partial_trap->rvert = T_rvert(trap_walk);
		    partial_trap->rvtype = T_rvtype(trap_walk);
		    
		    if ((partial_trap == top_trap) &&
			(T_rvtype(partial_trap) == top))
			T_te(partial_trap)->brTrap = partial_trap;
		    if ((partial_trap == bot_trap) &&
			(T_rvtype(partial_trap) == bot))
			T_be(partial_trap)->trTrap = partial_trap;
		    
		    /*---- fix up right vert's references ----*/
		    if (V_tlTrap(T_rvert(partial_trap)) == trap_walk)
			T_rvert(partial_trap)->tlTrap = partial_trap;
		    if (V_blTrap(T_rvert(partial_trap)) == trap_walk)
			T_rvert(partial_trap)->blTrap = partial_trap;
		}
		
		
		/*---- reset traps for next pass ----*/
		partial_trap = (close_top ? bot_trap : top_trap);
		if (close_top) top_trap = NULL;
		if (close_bot) bot_trap = NULL;

		
		/** question:  is it possible for the original normal to pass on one
		  side of the opp_vertex, but the new normal to pass on the other side
		  without it rounding to that vertex?   I don't think so, but if this
		  is the case, then there might be some serious trouble.  ***/

		if (VertEdgeIncedent(V_direction(T_rvert(trap_walk)),
				     new_normal)) {
		    if (close_top !=
			(Edge_Test(V_direction(T_rvert(trap_walk)),
				   new_normal) == ANS_ABOVE)) {
			printf("Arrange -- Trouble in 'bending edges'.\n");
			printf("  See comments in source code.\n");
			/** note:  if this is reached, it's not
			  automatically trouble.  It's only trouble
			  if the new normal actually switched to the
			  opposite side of the vert as the one we
			  thought. **/
		    }
		}
		
		
	    } else {
		/*---- "end of the line...everybody off" ----*/
		close_top = close_bot = TRUE;
		
		partial_trap->rvert = T_rvert(trap_walk);
		
		/* set partial_trap->rvtype to 'default', and then
		   it will get changed later if necessary */
		partial_trap->rvtype = T_rvtype(trap_walk);
		
		switch (T_rvtype(trap_walk)) {
		  case com:
		    if (partial_trap == top_trap)
			T_te(partial_trap)->brTrap = partial_trap;
		    else
			T_be(partial_trap)->trTrap = partial_trap;
		    break;
		    
		  case top:
		    top_trap->rvtype = com;
		    if (partial_trap == top_trap)
			T_te(partial_trap)->brTrap = partial_trap;
		    break;
		    
		  case bot:
		    bot_trap->rvtype = com;
		    if (partial_trap == bot_trap)
			T_be(partial_trap)->trTrap = partial_trap;
		    break;
		    
		  case mid:
		    top_trap->rvtype = bot;
		    bot_trap->rvtype = top;
		    break;
		}
		
		/*---- fix up right vert's references ----*/
		if (V_tlTrap(T_rvert(trap_walk)) == trap_walk)
		    T_rvert(trap_walk)->tlTrap = top_trap;
		if (V_blTrap(T_rvert(trap_walk)) == trap_walk)
		    T_rvert(trap_walk)->blTrap = bot_trap;
	    }
	}   /* matches 'for' loop */
	
	loop__edge->trTrap = top_trap;
	loop__edge->brTrap = bot_trap;
	
    } else {
#ifdef ARRANGE_DEBUG
	if (arrange_verbose)
	    printf("  reusing existing edge...\n");
#endif
    }

    FreeEdgeGeom(new_normal);

    /*--- free trap_stack ---*/
    delete_list(loop__trap_stack);

    /*--- set return variables ---*/
    (*return_edge) = loop__edge;
    
}  


/*************************************************************
  Function:  Arrange_AddEdge(the_A, start_vertex, goal_vertex,
  goal_vector, goal_normal, polys_on,
  polys_below, polys_above)
  Input:
      the_A -- a valid arrangement
      start_vertex -- TrapVert to start new edge (must be given)
      goal_vertex -- OPTIONAL parameter giving TrapVert at end
      goal_vector -- if NO goal_vertex, then this gives the
                       VertGeom for where new edge should end.
      goal_normal -- EdgeGeom describing new edge path
      polys_on -- Polygons entered only when *on* edge
      polys_below -- Polygons entered when on or below edge
      polys_above -- Polygons entered when on or above edge
  Output:
      TrapVert at end of new edge
  Effects: updates entire Arrange structure to account for the
       insertion of the new edge.
  
  Description:
  
  Arrange_AddEdge --  This is the high level routine to add a
  (possibly) new edge to the arrangement.  It does everything
  necessary.
  
  REQUIREMENT:  this routine assumes that the edge lies entirely in
  the (Z >= 0) hemisphere.  Also, it assumes that an edge lying on
  the Z=0 plane does NOT cross either of the poles. If this is not
  the case, it is up to the caller to break the edge up into pieces
  that satisfy this.
  
  Also, it assumes that the normal given truly passes through both
  endpoints.  Usually this normal should be uniquely defined by the
  points, but in the case of a full half-circle, it is ambiguous and
  hence the normal must be part of the input.
  
  For efficiency, it takes a pointer to the TrapVert which coincides
  with the start_endpoint. Also, to be consistent, the function returns
  a pointer to the TrapVert which coincides with the final_endpoint.
  
  If the goal_vector lies on the horizon, then this routine will
  guarantee that the final_endpoint returned is a vertex lying on
  the horizon.
  
  Also, if an edge is so small that the end is merged with the start
  vertex, this routine will NOT add any PolySource to the arrangement.
  If the edge is part of a chain of edges, then there is no need
  for any adjustment.   However, if for some reason, the edge is part
  of a polygon/polychain so small that it is treated as a singularity,
  then it is necessary to add the PolySource to the vertex->polys
  itself.  It is the responsibility of the CALLER of this routine to
  handle that degeneracy.
  
  **************************************************************/
TrapVert* Arrange_AddEdge(the_A, start_vertex, goal_vertex, goal_vector,
			  goal_normal, polys_on, polys_below, polys_above)
  Arrange		*the_A;
  TrapVert		*start_vertex;	/* this MUST be given */
  TrapVert		*goal_vertex;	/* this is OPTIONAL */
  struct VertGeom*	goal_vector;	/* this is used if no goal_vertex */
  struct EdgeGeom*	goal_normal;	/* this MUST be given */
  List			polys_on;	/* List of (PolySource)   */
  List			polys_below;	/* List of (PolySource)   */
  List			polys_above;	/* List of (PolySource)   */
{
    /*
      This procedure will have one main loop, which traverses from
      TrapVert to TrapVert in each pass. (notice that if the new edge
      crosses other edges, it will end up becoming several TrapEdges)
      Inside the loop, we will use a four-phase system.
      
      Phase 1 starts at a vertex and figures out what trap/edge to
      travel through.  Special care must be used to handle the case
      where you leave exactly through an edge, or very close to an
      edge.  In some of these cases, existing edges will be used
      in which case phases 2 & 3 can be skipped.
      
      Phase 2 starts walking, building a list of traps that we pass
      through, however it does NOT change the structure of these traps.
      When it reaches the final goal, or crosses an edge or existing
      vertex, it stops, and adds the new vertex if necessary.
      
      Phase 3 then walks back through the list of traps built up and
      uses that to connect the two TrapVerts with a single new TrapEdge.
      This handles splitting all the existing traps and all associated
      structures.
      
      Phase 4 simply takes care of adding the new PolySource's to the
      edge being added/reused.  Then it resets the loop variables.
      */
    
    Boolean		from_left;	/* Moving left to right? */
    VertGeom*		true_goal;	/* use this so we don't always
					   have to branch based on
					   whether goal is vert or vec */
    EdgeGeom*		loop__new_normal;
    Trap		*loop__enter_trap;
    TrapVert		*loop__enter_vertex;
    TrapEdge		*loop__edge;
    TrapVert		*loop__opp_vertex;
    List		loop__trap_stack;

    /*---- set true goal, to avoid several checks later ----*/
    true_goal = (goal_vertex ? V_direction(goal_vertex) : goal_vector);
    
    
    /*---- init loop variables ----*/
    loop__enter_vertex = start_vertex;
    
#ifdef ARRANGE_DEBUG
    if (arrange_verbose)
	printf("\n");
/*    if (arrange_verbose)
	Arrange_Print(the_A,TwoParts_NicePrint);*/
#endif
    
    /*****************************************************
     ******             Main Loop             ************
     *****************************************************/
    while ((goal_vertex) ? (loop__enter_vertex != goal_vertex) :
	   (!VertVertIncedent(V_direction(loop__enter_vertex),true_goal))) {
	
	
#ifdef ARRANGE_DEBUG
	if (loop__enter_vertex->id == 11) {
	    sleep(0);
	}
	if (arrange_verbose)
	    printf("\n");
	if (arrange_verbose) {
	    if (!Arrange_SanityCheck(the_A)) {
		printf("Sanity check failed...\n");
	    }
	    if (arrange_mike)
		Arrange_Print(the_A,TwoParts_NicePrint);
	}
#endif

	/*---- set from_left ----*/
	from_left =
	    (Thread_Test(V_direction(loop__enter_vertex),
			 true_goal) == ANS_LEFT);
    
	/*****************************************************
	 ******             phase one             ************
	 *****************************************************/
	AddEdge_Phase1(&loop__new_normal, &loop__enter_trap,
		       the_A, loop__enter_vertex, from_left,
		       true_goal, goal_normal);
	
	
	/*****************************************************
	 ******             phase two             ************
	 *****************************************************/
	
	AddEdge_Phase2(&loop__trap_stack, &loop__opp_vertex,
		       the_A, from_left, loop__enter_vertex,
		       loop__enter_trap, loop__new_normal, true_goal);
	
	
	/*****************************************************
	 ******              phase three          ************
	 *****************************************************/
	
	AddEdge_Phase3(&loop__edge, the_A, from_left, goal_normal,
		       loop__enter_vertex, loop__opp_vertex,
		       loop__trap_stack);
	
	
	/*****************************************************
	 ******             phase four            ************
	 *****************************************************/
	
#ifdef ARRANGE_DEBUG
	if (arrange_verbose)
	    printf("Starting phase four...\n");
#endif
	
	loop__edge->polys_on =
	    PolyTree_AppendList(the_A,polys_on,E_polys_on(loop__edge));
	loop__edge->polys_above =
	    PolyTree_AppendList(the_A,polys_above,E_polys_above(loop__edge));
	loop__edge->polys_below =
	    PolyTree_AppendList(the_A,polys_below,E_polys_below(loop__edge));
	
	
	/*---- reset loop variables ----*/
	loop__enter_vertex = ((from_left) ? E_rvert(loop__edge) :
			      E_lvert(loop__edge));

	FreeEdgeGeom(loop__new_normal);
	
    }	/* end of Main Loop */

    return(loop__enter_vertex);
}



enum AddArray {ADD_ON=0, ADD_BELOW, ADD_ABOVE, Num_AddArray};
enum CaseCode {CASE0 = -1, CASE1, CASE2, CASE2Bs, CASE3, CASE3s, CASEH, CASEHs};
enum HemiCode {HEMI_FIRST=0, HEMI_SECOND, HEMI_BOTH};
enum AddList {ADD_P1=0, ADD_P2, ADD_BOTH, ADD_NONE, Num_AddList};
enum VertCode {V_BEGIN=0, V_END, V_TEMP, V_TEMP_MIRROR,
		   V_NULL, V_IGNORE, Num_VertCode};
enum VecCode {VEC_END=0, VEC_BEGIN, VEC_HOR1, VEC_HOR2, Num_VecCode};


typedef int	Micro[6];

typedef struct SegCase {
    int		num_pieces;
    Micro	*pieces;
} SegCase;


static	Micro Micro_Case1[1] = {
    /*---- case 1 ----*/
    {HEMI_FIRST, V_END, V_BEGIN, V_END, VEC_END, VEC_END}
};


static	Micro Micro_Case2[2] = {
    /*---- case 2 ----*/
    {HEMI_FIRST, V_TEMP, V_BEGIN, V_NULL, VEC_HOR1, VEC_HOR2},
    {HEMI_SECOND, V_END, V_TEMP_MIRROR, V_END, VEC_END, VEC_END}
};

static	Micro Micro_Case2bs[3] = {
    /*---- case 2bs ----*/
    {HEMI_SECOND, V_TEMP, V_BEGIN, V_NULL, VEC_HOR2, VEC_HOR1},
    {HEMI_FIRST, V_END, V_TEMP_MIRROR, V_END, VEC_END, VEC_END},
    {HEMI_BOTH, V_IGNORE, V_END, V_BEGIN, VEC_BEGIN, VEC_BEGIN}
};

static	Micro Micro_Case3[3] = {
    /*---- case 3 ----*/
    {HEMI_FIRST, V_TEMP, V_BEGIN, V_NULL, VEC_HOR1, VEC_HOR2},
    {HEMI_SECOND, V_TEMP, V_TEMP_MIRROR, V_NULL, VEC_HOR2, VEC_HOR1},
    {HEMI_FIRST, V_END, V_TEMP_MIRROR, V_NULL, VEC_END, VEC_END}
};

static	Micro Micro_Case3s[3] = {
    /*---- case 3s ----*/
    {HEMI_BOTH, V_TEMP, V_BEGIN, V_NULL, VEC_HOR1, VEC_HOR2},
    {HEMI_BOTH, V_END, V_TEMP_MIRROR, V_END, VEC_END, VEC_END},
    {HEMI_SECOND, V_IGNORE, V_END, V_BEGIN, VEC_BEGIN, VEC_BEGIN}
};

static	Micro Micro_CaseH[3] = {
    /*---- case H ----*/
    {HEMI_FIRST, V_TEMP, V_BEGIN, V_NULL, VEC_HOR1, VEC_HOR2},
    {HEMI_SECOND, V_IGNORE, V_TEMP_MIRROR, V_END, VEC_BEGIN, VEC_BEGIN}
};

static	Micro Micro_CaseHs[3] = {
    /*---- case Hs ----*/
    {HEMI_BOTH, V_IGNORE, V_BEGIN, V_END, VEC_END, VEC_END}
};


/*--- Note:  Order here is based on CaseCode order ---*/
static SegCase SegMicro[7] = {
    { 1, Micro_Case1 },
    { 2, Micro_Case2 },
    { 3, Micro_Case2bs },
    { 3, Micro_Case3 },
    { 3, Micro_Case3s },
    { 2, Micro_CaseH },
    { 1, Micro_CaseHs }
};


/*************************************************************
  Function: Setup_PolyTrees()


  **************************************************************/
Setup_PolyTrees(add_results, add_lists, hemi,
		 interior, edge_upward, start_on_front, symmetric)
  List	add_results[3];
  List	add_lists[4];
  enum HemiCode  hemi;
  Boolean interior;
  Boolean edge_upward;
  Boolean start_on_front;
  Boolean symmetric;
{
    int  which_case;


    if (interior) {
	add_results[ADD_ON] = add_lists[ADD_NONE];

	which_case =
	    (symmetric ?
	     /*--- Symmetric case ---*/
	     ((hemi == HEMI_BOTH) ?
	      (edge_upward ? (ADD_P1 + 4*ADD_P2) : (ADD_P2 + 4*ADD_P1)) :
	      (edge_upward ?
	       ((start_on_front == (hemi==HEMI_FIRST)) ?
		(ADD_P1 + 4*ADD_NONE) : (ADD_NONE + 4*ADD_P2)) :
	       ((start_on_front == (hemi==HEMI_FIRST)) ?
		(ADD_NONE + 4*ADD_P1) : (ADD_P2 + 4*ADD_NONE)))) :

	     /*--- Standard case ---*/
	     (edge_upward ? (ADD_P1 + 4*ADD_NONE) : (ADD_NONE + 4*ADD_P1)));

	add_results[ADD_ABOVE] = add_lists[which_case % 4];
	add_results[ADD_BELOW] = add_lists[which_case / 4];
    } else {
	/* interior == FALSE */
	add_results[ADD_ABOVE] = add_results[ADD_BELOW] =
	    add_lists[ADD_NONE];

	add_results[ADD_ON] = 
	    ((!symmetric) ? add_lists[ADD_P1] :
	     ((hemi == HEMI_BOTH) ?
	      add_lists[ADD_BOTH] :
	      ((start_on_front == (hemi==HEMI_FIRST)) ?
	       add_lists[ADD_P1] : add_lists[ADD_P2])));
    }
    
}




/*************************************************************
  Function: Arrange_AddPolygonSeg(<much stuff>)
  Input:
      the_A -- a valid arrangement
      begin_vert -- TrapVert to start at  (NULL if hemisphere)
      begin_vec  -- VertGeom representing begining point on sphere
      end_vert -- pointer to TrapVert to end segment if known
                   (gets set otherwise)
      end_vec  -- VertGeom representing end point on sphere
      edge_vec -- EdgeGeom representing the edge to travel
      add_lists -- preecomputed polysource lists
  Output:
      Boolean -- degenerate
  Effects: Adds single polygon segment into the arrangement
  & sets end_vert if originally == NULL
  
  Description:  This is a routine to add a single polygon boundary
  into the arrangment.  It could have been done as straightline
  code as an inner loop of Arrange_AddPolygon but was separated out
  for simplicity, for a break in the data link to clarify the
  necessary data, and because it will need to be fixed to deal
  with arrangments that have two different hemispheres.
  
  **************************************************************/
Boolean Arrange_AddPolygonSeg(the_A, begin_vert, begin_vec,
			      end_vert, end_vec, edge_vec,
			      interior, add_lists)
  Arrange		*the_A;
  TrapVert		*begin_vert;
  VertGeom*		begin_vec;
  TrapVert		**end_vert;
  VertGeom*		end_vec;
  EdgeGeom*		edge_vec;
  Boolean		interior;
  List			add_lists[Num_AddList];
{
    List		add_array[Num_AddArray];
    
    Boolean		degenerate;

    Boolean		symmetric;
    
    ArrangeCell		location;
    TrapVert*		temp_vert;

    EdgeGeom*		unit_edge;
    VertGeom*		unit_begin;
    VertGeom*		unit_end;
    Boolean		flip_begin;
    Boolean		flip_end;
    Boolean		edge_upward;
#define start_on_front (!flip_begin)
    
    EdgeGeom*		edge_hor;
    VertGeom*		vec_hor1;
    VertGeom*		vec_hor2;
    Boolean		cross_hor1;
    Boolean		cross_hor2;
    
    Micro		*the_micro;
    int			piece;
    enum CaseCode	c;		/*--- which case ---*/
    TrapVert*		VertTable[Num_VertCode];
    VertGeom*		VecTable[Num_VecCode];
    

    symmetric = IS_DOUBLESYM(the_A);
    c = CASE0;

    /*---- make copies of vectors in case negation is necessary ----*/
    unit_edge = CopyEdgeGeom(edge_vec);
    
    edge_hor = GetHorizonGeom(FALSE);

    vec_hor1 =
	EdgeEdgeIntersect(unit_edge,edge_hor,NULL,NULL,NULL,NULL);

    FreeEdgeGeom(edge_hor);
    
    vec_hor2 = CopyVertGeom(vec_hor1);
    NegateVertGeom(vec_hor2);
    


    if (!begin_vert) {
	/*---- special case: full great circle ----*/
	degenerate = FALSE;
	flip_begin = FALSE;

	c = (symmetric ? CASEHs : CASEH);
	unit_begin = CopyVertGeom(vec_hor1);
	unit_end = CopyVertGeom(vec_hor2);
	
	location = Arrange_PointLocation(the_A,vec_hor1);
	temp_vert =
	    Arrange_AddVertex(the_A,vec_hor1,NULL,location);
	
	/*---- note this does NOT effect true parameter ----*/
	begin_vert = temp_vert;
	
	NegateVertGeom(vec_hor1);
	NegateVertGeom(vec_hor2);
	
	/*--- flip edge if not upward OR if on left horizon ---*/
	edge_upward =
	    (OnHorizonEdge(unit_edge) ?
	     !EdgeOnLeftHorizon(unit_edge) :
	     Upward(unit_edge));
	if (!edge_upward) NegateEdgeGeom(unit_edge);
    
    } else {
	/*---- normal case ---- */
	degenerate = TRUE;

	unit_begin = CopyVertGeom(begin_vec);
	unit_end = CopyVertGeom(end_vec);
	
	/*---
	  Note:  the following two comparisons are
	  done in an odd way for the specific reason
	  of dealing with the special behavior of
	  Closer() when two of the vertices are
	  equal.
	  ---*/
	cross_hor1 = !Closer(InqVertVector(unit_end),
			     InqEdgeVector(unit_edge),
			     InqVertVector(vec_hor1),
			     InqVertVector(unit_begin),
			     FALSE);
	cross_hor2 = !Closer(InqVertVector(unit_end),
			     InqEdgeVector(unit_edge),
			     InqVertVector(vec_hor2),
			     InqVertVector(unit_begin),
			     FALSE);
	
	
	/*---- switch so hor1 is first one we hit ----*/
	if (cross_hor2 &&
	    ((!cross_hor1) ||
	     Closer(InqVertVector(unit_begin),
		    InqEdgeVector(unit_edge),
		    InqVertVector(vec_hor2),
		    InqVertVector(vec_hor1),
		    FALSE))) {
	    NegateVertGeom(vec_hor1);
	    NegateVertGeom(vec_hor2);
	    cross_hor2 = cross_hor1;
	    cross_hor1 = TRUE;
	}
	
	
	
	/**
	  now that we've done necessary tests, let's prepare
	  all the data to be front/upward before continuing.
	  **/
	
	/*--- flip edge if not upward OR if on left horizon ---*/
	edge_upward =
	    (OnHorizonEdge(unit_edge) ?
	     !EdgeOnLeftHorizon(unit_edge) :
	     Upward(unit_edge));
	if (!edge_upward) NegateEdgeGeom(unit_edge);
	
	
	/*---
	  here we look at whether the very beginning/end part
	  of the walk is on the back hemisphere or if it is
	  ALONG the left half of the horizon.
	  
	  However we allow if the endpoint is on the border of
	  these bad regions, it just can't be that they travel
	  into the bad regions.
	  ---*/
	
	flip_begin =
	    /*---  If begin is on back part of hemisphere... ---*/
	    ((!Front(unit_begin)) ||
	     
	     /*--- or begin is on horizon AND ... ---*/
	     (OnHorizonVert(unit_begin) &&
	      
	      /*--- either edge walks along horizon and beginning
		of walk is on left horizon or edge is NOT along
		horizon and beginning of walk is onto the back ---*/
	      (VertOnLeftHorizon(unit_begin) ==
	       (OnHorizonEdge(unit_edge) ?
		(edge_upward || (!OnPoleVert(unit_begin))) :
		(!edge_upward)))));
	if (flip_begin) NegateVertGeom(unit_begin);
	
	
	flip_end =
	    /*---  If end is on back part of hemisphere... ---*/
	    ((!Front(unit_end)) ||
	     
	     /*--- or end is on horizon AND ... ---*/
	     (OnHorizonVert(unit_end) &&
	      
	      /*--- either edge walks along horizon and end
		of walk is on left horizon or edge is NOT along
		horizon and end of walk is from the back ---*/
	      (VertOnLeftHorizon(unit_end) ==
	       (OnHorizonEdge(unit_edge) ?
		((!edge_upward) || (!OnPoleVert(unit_end))) :
		edge_upward))));
	if (flip_end) NegateVertGeom(unit_end);
	
	
	
	if (ON_HORIZON_VERTEX(begin_vert) &&
	    (VertOnLeftHorizon(V_direction(begin_vert)) !=
	     VertOnLeftHorizon(unit_begin))) {
	    /*---- note this does NOT effect true parameter ----*/
	    begin_vert = V_mirror(begin_vert);
	}
	
	if ((*end_vert) && ON_HORIZON_VERTEX(*end_vert) &&
	    (VertOnLeftHorizon(V_direction(*end_vert)) !=
	     VertOnLeftHorizon(unit_end))) {
	    /*---- note this DOES effect return parameter ----*/
	    (*end_vert) = V_mirror(*end_vert);
	}
	
    }
    
    /**---- grab instructions from Micro code ----**/
    VertTable[V_BEGIN] = begin_vert;
    if (end_vert)
	VertTable[V_END] = *end_vert;
    else
	VertTable[V_END] = V_mirror(begin_vert);

    VertTable[V_TEMP] = NULL;
    VertTable[V_TEMP_MIRROR] = NULL;
    VertTable[V_NULL] = NULL;
    VertTable[V_IGNORE] = NULL;
    
    VecTable[VEC_END] = unit_end;
    VecTable[VEC_BEGIN] = unit_begin;
    VecTable[VEC_HOR1] = vec_hor1;
    VecTable[VEC_HOR2] = vec_hor2;
    
    
    /*  I made this a #define
	start_on_front = (!flip_begin);
	*/
    
    if (c == CASE0) {
	/*---- still undefined! ----*/
	if (!cross_hor1) {
	    /*---- case 1 ----*/
	    c = CASE1;
	} else if (!cross_hor2) {
	    /*---- case 2 ----*/
	    if (!symmetric ||
		(Thread_Test(unit_begin, unit_end) !=
		 Thread_Test(unit_begin,
			     (start_on_front ? vec_hor1 : vec_hor2)))) {
		/*---- case 2 ---- */
		c = CASE2;
	    } else {
		/*---- case 2bs ---- */
		c = CASE2Bs;
	    }
	} else {
	    /*---- case 3 ----*/
	    c = (symmetric ? CASE3s : CASE3);
	}
    }
    
    
    for (piece = 0; piece < SegMicro[c].num_pieces; piece++) {
	
	the_micro = (Micro*) SegMicro[c].pieces[piece];
	
	/*---------------------------------------*/
	/*---- setup the correct constraints ----*/
	Setup_PolyTrees(add_array, add_lists, (*the_micro)[0],
			interior, edge_upward, start_on_front, symmetric);
	
	if (VertTable[V_TEMP])
	    VertTable[V_TEMP_MIRROR] = V_mirror(VertTable[V_TEMP]);
	
	if (!((*the_micro[0]==HEMI_SECOND) && (IS_SINGLE(the_A))))
	    VertTable[SegMicro[c].pieces[piece][1]] =
		Arrange_AddEdge(((((*the_micro)[0]!=HEMI_SECOND) &&
				  (!symmetric)) ?
				 the_A : the_A->mirror),
				VertTable[(*the_micro)[2]],
				VertTable[(*the_micro)[3]],
				(start_on_front ?
				 VecTable[(*the_micro)[4]] :
				 VecTable[(*the_micro)[5]]),
				unit_edge, add_array[ADD_ON],
				add_array[ADD_BELOW], add_array[ADD_ABOVE]);
	
	if (degenerate &&
	    (VertTable[(*the_micro)[1]] != VertTable[(*the_micro)[2]]))
	    degenerate = FALSE;
    }
    
    if (end_vert)
	*end_vert = VertTable[V_END];

    FreeEdgeGeom(unit_edge);
    FreeVertGeom(unit_begin);
    FreeVertGeom(unit_end);
    FreeVertGeom(vec_hor1);
    FreeVertGeom(vec_hor2);

    return(degenerate);
}




/*************************************************************
  Function: Arrange_AddPolygon(the_A,poly)
  Input:
      the_A -- a valid arrangement
      poly -- a spherical polygon as defined in polygon.h
  Output: <none>
  Effects: Adds polygon into the arrangement
  
  Description:  This is the high level routine to add a 
  polygon into the arrangement.  It breaks the polygon's edges
  into parts when they cross the horizon, and it maps everything
  onto the front hemisphere, adjusting the interpretation
  of the blocking relationship as necessary.
  
  **************************************************************/
void Arrange_AddPolygon(the_A,poly)
  Arrange	*the_A;
  Polygon	*poly;
{
    List		elt_boundary;	/* traverse list of boundaries */
    List		elt_side;	/* traverse list of sides */
    PolygonBoundary*	cur_boundary;
    PolygonSeg*		cur_side;
    Boolean		degenerate;	/* collapsed to singularity */
    
    PolySource*		polys1;
    PolySource*		polys2;
    List		add_lists[43];
    List		add_on, add_below, add_above;
    
    ArrangeCell		location;
    TrapVert*		initial_vert;
    Vector*		initial_vector;
    TrapVert*		link_vert;
    Vector*		link_vector;	/* ONLY for error checking */
    Boolean		reached_end;
    TrapVert*		end_vert;
    
    VertGeom*		unit_initial;
    VertGeom*		unit_begin;
    VertGeom*		unit_end;
    EdgeGeom*		unit_edge;
    Boolean		flip_initial;
    double		scale;
    

#ifdef ARRANGE_DEBUG
    if ((((TwoParts*) poly->data)->part1->id==17) &&
	(((TwoParts*) poly->data)->part2->id==18)) {
	sleep(0);
    }
#endif

    polys1 = Make_PSource(the_A);
    polys1->poly = poly;
    polys1->flipped = FALSE;
    polys2 = Make_PSource(the_A);
    polys2->poly = poly;
    polys2->flipped = TRUE;
    add_lists[ADD_NONE] = NULL;
    add_lists[ADD_P1] = init_list(polys1);
    add_lists[ADD_P2] = init_list(polys2);
    add_lists[ADD_BOTH] = init_list(polys2);
    add_lists[ADD_BOTH] = push(add_lists[ADD_BOTH],polys1);
    
    for (elt_boundary = poly->boundaries; (elt_boundary);
	 elt_boundary = NXT(elt_boundary)) {
	
	cur_boundary = GET(elt_boundary,PolygonBoundary);
	degenerate = TRUE;
	initial_vert = NULL;
	initial_vector = NULL;
	link_vert = NULL;
	link_vector = NULL;

	unit_begin = unit_end = unit_initial = unit_edge = NULL;

	for (elt_side = cur_boundary->sides;  (elt_side);
	     elt_side = NXT(elt_side)) {
	    cur_side = GET(elt_side,PolygonSeg);
	    
	    if (!cur_side->edge) {
		if (cur_side->begin) {
		    /*---- special case: isolated vertex ----*/
		    unit_begin =
			CreateVertFromVector(*cur_side->begin,FALSE);
		    if (flip_initial = (!Front(unit_begin)))
			NegateVertGeom(unit_begin);
		    
		    location = Arrange_PointLocation(the_A,unit_begin);
		    initial_vert =
			Arrange_AddVertex(the_A,unit_begin, NULL, location);

		    /** constraints will be added later in the
		      degnerate case part of the code ***/
		}
	    } else {
		/*---- normal case or full great circle ----*/
		
		/*---- set unit_edge properly ---*/
		unit_edge = CreateEdgeFromVector(*cur_side->edge,FALSE);
		
		if (!cur_side->begin) {
		    /*---- special case: full great circle ----*/
		    degenerate =
			Arrange_AddPolygonSeg(the_A,NULL,NULL,NULL,NULL,
					      unit_edge,
					      cur_boundary->interior,
					      add_lists);
		} else {
		    /*---- normal case ---- */
		    
		    /*---- set unit_begin properly ---*/
		    if (link_vert) {
			FreeVertGeom(unit_begin);
			unit_begin = unit_end;
			if (link_vector != cur_side->begin)
			    printf("Warning:  polygon vertices not unified\n");
		    } else {
			unit_begin =
			    CreateVertFromVector(*cur_side->begin,FALSE);
			initial_vector = cur_side->begin;

			unit_initial = CopyVertGeom(unit_begin);
			
			if (flip_initial = (!Front(unit_begin)))
			    /*-- temporarily flip for PointLocation --*/
			    NegateVertGeom(unit_begin);
			
			location = Arrange_PointLocation(the_A,unit_begin);
			link_vert = initial_vert =
			    Arrange_AddVertex(the_A,unit_begin,
					      NULL,location);

			/*--- flip back if necessary ---*/
			if (flip_initial) NegateVertGeom(unit_begin);
		    }
		    
		    
		    /*---- set unit_end properly ---*/
		    link_vector = cur_side->end;
		    if (reached_end = (cur_side->end == initial_vector)) {
			unit_end = CopyVertGeom(unit_initial);
		    } else {
			unit_end = CreateVertFromVector(*cur_side->end,FALSE);
		    }
		    
		    
		    end_vert = (reached_end ?
				initial_vert : NULL);
		    
		    degenerate &=
			Arrange_AddPolygonSeg(the_A,link_vert,unit_begin,
					      &end_vert,
					      unit_end,unit_edge,
					      cur_boundary->interior,
					      add_lists);
		    
		    link_vert = end_vert;
		}
	    }

	    if (unit_edge) FreeEdgeGeom(unit_edge);
	}
	
	if (unit_initial) FreeVertGeom(unit_initial);
	if (unit_begin) FreeVertGeom(unit_begin);
	if (unit_end) FreeVertGeom(unit_end);

	/*---- if entire boundary was treated as singularity... ----*/
	if (degenerate && initial_vert) {
	    add_on = (flip_initial ? add_lists[ADD_P2] : add_lists[ADD_P1]);
	    initial_vert->polys = 
		PolyTree_AppendList(the_A,add_on, V_polys(initial_vert));
	}
    }
    
    
    
    /*---- delete lists, but do NOT delete Constraint's ----*/
    delete_list(add_lists[ADD_P1]);
    delete_list(add_lists[ADD_P2]);
    delete_list(add_lists[ADD_BOTH]);
}



/*************************************************************
  Function: CreatePerm(n)
  Input: n -- size of perm
  Output: an array with a 'random' perm of numbers [0..n-1]
  Effects: <none>
  **************************************************************/
int* CreatePerm(n)
  int	n;
{
    int*	perm;
    int		i;
    int		r;
    
    perm = ALLOCATE_ARRAY(int,n);
    
    perm[0] = 0;
    
    for (i=1; i<n; i++) {
	r = (random() % (i+1));		/* pick from [0..i] */
	perm[i] = perm[r];
	perm[r] = i;
    }
    
    return(perm);
}

/*************************************************************
  Function: Arrange_AddListOfPolygons(the_A,polylist,rand)
  Input:
     the_A -- a valid arrangement
     polylist -- a list of spherical polygons
     rand -- should we randomize order of insertion
  Output: <none>
  Effects: Adds list of polygon into the arrangement
  
  Description:  This is the high level routine to add a list of 
  polygon into the arrangement.  It randomizes the order of
  insertion depending on the rand flag.
  
  **************************************************************/
void Arrange_AddListOfPolygons(Arrange *the_A, List polylist, Boolean rand)
{
    List	elt_poly;
    int		num_polys;
    int*	perm;
    Polygon**	input;
    int		i;
    
    if (rand) {
	num_polys = length(polylist);
	perm = CreatePerm(num_polys);
	input = ALLOCATE_ARRAY(Polygon*, num_polys);
	for (i=0, elt_poly = polylist; (elt_poly); 
	     elt_poly = NXT(elt_poly),i++) {
	    input[perm[i]] = GET(elt_poly,Polygon);
	}

	if (arrange_mike) {
	    for (i=0; i<num_polys; i++) {
		Polygon_Print(stdout,input[i],TwoParts_Print);
		printf("\n");
	    }
	}

	for (i=0; i<num_polys; i++) {
	    Arrange_AddPolygon(the_A,input[i]);
	}
	FREE(perm);
	FREE(input);
    } else {
	for (elt_poly = polylist;  (elt_poly); elt_poly = NXT(elt_poly)) {
	    Arrange_AddPolygon(the_A,GET(elt_poly,Polygon));
	}
    }
}




