/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * arrange.h
 *
 * Function involving maintaining arrangements
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/

#ifndef _ARRANGE
#define _ARRANGE

#include "standards.h"
#include "arrange_type.h"
#include "polygon.h"

/*************************************************************
  Function:  Arrange_Init_Single()
  Input:     <none>
  Output:    pointer to a newly initialized arrangment
  Effects:   <none>
  
  Description:  This routine creates a single arrangement
  on a hemisphere.
  
  The current initialization creates two vertices as the
  top and bottom poles, and two edges connecting them as
  the left and right horizon.  The entire hemisphere is
  a single trap bounded from above and below by the edge,
  and having the two poles as the right and left vertices.
  
  **************************************************************/
struct Arrange*	Arrange_Init_Single();


/*************************************************************
  Function:  Arrange_Init_Symmetric()
  Input:     <none>
  Output:    pointer to a newly initialized arrangment
  Effects:   <none>
  
  Description:  This routine creates a single arrangement,
  and then 'glues' it to itself so that the "back side"
  behaves in an anti-symmetric manner to the front.
  
  The current initialization creates two vertices as the
  top and bottom poles, and two edges connecting them as
  the left and right horizon.  The entire hemisphere is
  a single trap bounded from above and below by the edge,
  and having the two poles as the right and left vertices.
  
  **************************************************************/
struct Arrange*	Arrange_Init_Symmetric();



/*************************************************************
  Function:  Arrange_Init_Double()
  Input:     <none>
  Output:    pointer to a newly initialized arrangment
  Effects:   <none>
  
  Description:  This routine creates two separate arrangments
  and then 'glues' them together using the mirror pointers
  so that they behave topologically as a sphere.
  
  The current initialization creates two vertices as the
  top and bottom poles, and two edges connecting them as
  the left and right horizon.  The entire hemisphere is
  a single trap bounded from above and below by the edge,
  and having the two poles as the right and left vertices.
  
  The hemispheres are glued so that the top poles are
  glued to each other, as are the bottom poles, and each
  horizon is glued to the 'opposite' horizon of the mirror
  hemisphere.
  
  **************************************************************/
struct Arrange*	Arrange_Init_Double();



/*************************************************************
  Function: Arrange_Free(the_A)
  Input:  arrangement
  Output: <none>
  Effects: frees structure and decendants (inlcuding mirror)
  
  Description:  Frees up all the dynamic memory, and then
  frees the Arrange structure itself.  If mirror exists,
  it is also freeed.
  
  **************************************************************/
void Arrange_Free(struct Arrange*);



/*************************************************************
  Function: Arrange_AddPolygon(the_A,poly)
  Input:
      the_A -- a valid arrangement
      poly -- a spherical polygon as defined in polygon.h
  Output: <none>
  Effects: Adds polygon into the arrangement
  
  Description:  This is the high level routine to add a 
  polygon into the arrangement.  It breaks the polygon's edges
  into parts when they cross the horizon, and it maps everything
  onto the front hemisphere, adjusting the interpretation
  of the blocking relationship as necessary.
  
  **************************************************************/
void Arrange_AddPolygon(struct Arrange*, struct Polygon*);



/*************************************************************
  Function: Arrange_AddListOfPolygons(the_A,polylist,rand)
  Input:
     the_A -- a valid arrangement
     polylist -- a list of spherical polygons
     rand -- should we randomize order of insertion
  Output: <none>
  Effects: Adds list of polygon into the arrangement
  
  Description:  This is the high level routine to add a list of 
  polygon into the arrangement.  It randomizes the order of
  insertion depending on the rand flag.
  
  **************************************************************/
void Arrange_AddListOfPolygons(struct Arrange*, List, Boolean);


#endif
