/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * arrange_debug.c
 *
 * Functions involving debugger tools
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/

#include <stdio.h>
#include "arrange.h"
#include "arrange_debug.h"
#include "arrange_geom.h"
#include "arrange_macros.h"
#include "arrange_type.h"
#include "lists.h"
#include "polygon.h"
#include "standards.h"

#ifdef ARRANGE_GRAPHICS
#include "debug_graphics.h"
#endif

#ifdef ARRANGE_DEBUG

#define DOT(A,B)  ((A).x*(B).x + (A).y*(B).y + (A).z*(B).z)
#define NORM2(A)   (DOT(A,A))
#define NEG(A)     ((A).x = -(A).x, (A).y = -(A).y, (A).z = -(A).z)

extern Boolean arrange_verbose;

/*--- special mode for creating graphics without threads ---*/
extern Boolean show_threads;

/*************************************************************/


static char*	TrapVertTypeEnum[] = { "?", "top","bot","com","mid"};
static char*	HistTypeEnum[] = { "?", "thread","edge","leaf"};
static char*	CellTypeEnum[] = { "?", "vert","edge","trap"};

static Boolean hardcopy = FALSE;

#ifdef USE_ADDRESSES
static Boolean use_addresses = TRUE;
#else
static Boolean use_addresses = FALSE;
#endif


#define NULL_ID -1
#define ID(item) (item ? (item)->id : NULL_ID)

#define Print_VertID(item)  Print_ID(ID(item))
#define Print_TvertID(item) Print_ID(ID(item))
#define Print_EdgeID(item)  Print_ID(ID(item))
#define Print_TrapID(item)  Print_ID(ID(item))
#define Print_HistID(item)  Print_ID(ID(item))



Print_ID(id)
  int	id;
{
    if (id != NULL_ID) {
	printf("%d",id);
    } else {
	printf("null");
    }
}

Print_Address(item)
  void*		item;
{
	printf("%p",item);
}



typedef union Cell_Info {
    int   id;
    void* addr;
} Cell_Info;

static Cell_Info the_cell;

#define SET_CELL(item)  if (use_addresses) the_cell.addr = item; \
                        else the_cell.id = ((item) ? (item)->id : NULL_ID)

#define Print_Cell(s1,item,s2) SET_CELL(item);Print_Cell_REAL(s1,s2)

Print_Cell_REAL(s1,s2)
  char*	s1;
  char* s2;
{
    printf("%s(",s1);

    if (use_addresses)
	Print_Address(the_cell.addr);
    else
	Print_ID(the_cell.id);

    printf(")%s\n",s2);

}


Print_PolyTree(polys,Print_Data)
  PolyTree	*polys;
  void		(*Print_Data)(FILE*,void*);
{
    PolyTree	*trav;

    for (trav=polys; (trav); trav=trav->next) {
	printf(" %c", (trav->source->flipped ? '-' : '+'));
	if (Print_Data)
	    (*Print_Data)(stdout,trav->source->poly->data);
	else
	    printf("(%p)",trav->source->poly->data);
    }
    printf("\n");
}


Print_TrapVert(the_A,vert,Print_Data)
  Arrange	*the_A;
  TrapVert	*vert;
  void		(*Print_Data)(FILE*,void*);
{
    Vector	vec;

    Print_Cell("TrapVert", vert, "");

    if ON_HORIZON_VERTEX(vert) {
	Print_Cell(" mirror", V_mirror(vert), "");
    }
    if (vert == the_A->tpole) printf(" ** Pole **\n");

    vec = InqVertVector(V_direction(vert));

    printf(" direction(%lf,%lf,%lf)\n",vec.x,vec.y,vec.z);

    printf("  polys:");
    Print_PolyTree(vert->polys,Print_Data);

    Print_Cell(" tlTrap", V_tlTrap(vert), "");
    /*Print_Cell(" trTrap", V_trTrap(vert), "");*/
    Print_Cell(" brTrap", V_brTrap(vert), "");
    /*Print_Cell(" blTrap", V_blTrap(vert), "");*/

    printf("\n");
}


Print_TrapThreadVert(the_A,tvert)
  Arrange	*the_A;
  TrapThreadVert *tvert;
{
    Print_Cell("TrapThreadVert", tvert, "");

    Print_Cell(" edge", TV_edge(tvert), "");
    Print_Cell(" source", TV_source(tvert), "");
    printf(" from_above(%s)\n",(TV_from_above(tvert)) ? "TRUE" : "FALSE");
    Print_Cell(" lTrap", TV_lTrap(tvert), "");
    Print_Cell(" rTrap", TV_rTrap(tvert), "");
    Print_Cell(" lNeigh", TV_lNeigh(tvert), "");
    Print_Cell(" rNeigh", TV_rNeigh(tvert), "");
    printf("\n");
}


Print_TrapEdge(the_A,edge,Print_Data)
  Arrange	*the_A;
  TrapEdge	*edge;
  void		(*Print_Data)(FILE*,void*);
{
    Vector	vec;

    Print_Cell("TrapEdge", edge, "");

    if ON_HORIZON_EDGE(edge) {
	Print_Cell(" mirror", E_mirror(edge), "");
    }

    vec = InqEdgeVector(E_normal(edge));

    printf(" normal(%lf,%lf,%lf)\n", vec.x, vec.y, vec.z);

    Print_Cell(" lvert", E_lvert(edge), "");
    Print_Cell(" rvert", E_rvert(edge), "");

    printf("  polys_on:");
    Print_PolyTree(edge->polys_on,Print_Data);
    printf("  polys_above:");
    Print_PolyTree(edge->polys_above,Print_Data);
    printf("  polys_below:");
    Print_PolyTree(edge->polys_below,Print_Data);

    Print_Cell(" lThrVert", E_lThrVert(edge), "");
    Print_Cell(" rThrVert", E_rThrVert(edge), "");
    Print_Cell(" tlTrap", E_tlTrap(edge), "");
    Print_Cell(" trTrap", E_trTrap(edge), "");
    Print_Cell(" blTrap", E_blTrap(edge), "");
    Print_Cell(" brTrap", E_brTrap(edge), "");

    printf("\n");
}


Print_Trap(the_A,trap)
  Arrange	*the_A;
  Trap		*trap;
{
    Print_Cell("Trap", trap, "");
    Print_Cell(" lvert", T_lvert(trap), "");
    Print_Cell(" rvert", T_rvert(trap), "");

    printf(" lvtype(%s)\n",TrapVertTypeEnum[T_lvtype(trap)]);
    printf(" rvtype(%s)\n",TrapVertTypeEnum[T_rvtype(trap)]);

    Print_Cell(" tlv", T_tlv(trap), "");
    Print_Cell(" trv", T_trv(trap), "");
    Print_Cell(" blv", T_blv(trap), "");
    Print_Cell(" brv", T_brv(trap), "");
    Print_Cell(" tln", T_tln(trap), "");
    Print_Cell(" bln", T_bln(trap), "");
    Print_Cell(" trn", T_trn(trap), "");
    Print_Cell(" brn", T_brn(trap), "");
    Print_Cell(" te", T_te(trap), "");
    Print_Cell(" be", T_be(trap), "");
    Print_Cell(" hist", T_hist(trap), "");

    printf("\n");
}


Print_TrapHist(the_A,hist)
  Arrange	*the_A;
  TrapHist	*hist;
{
    Vector	vec;

    Print_Cell("TrapHist", hist,"");

    if (hist == the_A->hist_root) printf(" ** Root **\n");

    printf(" nodetype(%s)\n",HistTypeEnum[hist->nodetype]);

    if (hist->nodetype == leaf) {
	vec.x = vec.y = vec.z = 0.0;
    } else {
	vec = ((hist->nodetype == threadtest) ?
	       InqVertVector(hist->data.v) :
	       InqEdgeVector(hist->data.e));
    }

    printf(" data(%lf,%lf,%lf)\n", vec.x, vec.y, vec.z);

    Print_Cell(" trap", hist->trap, "");
    Print_Cell(" ans[0]", hist->ans[0], "");
    Print_Cell(" ans[1]", hist->ans[1], "");

    printf("\n");
}



Translate_IDs(the_A)
  Arrange	*the_A;
{
    TrapVert		*vert;
    TrapThreadVert	*tvert;
    TrapEdge		*edge;
    Trap		*trap;
    TrapHist		*hist;

    printf("\nTrapVerts\n");
    for (vert=the_A->vert_link; (vert); vert=vert->link)
	printf("(%p) <--> (%d)\n",vert,V_id(vert));

    printf("\nTrapThreadVerts\n");
    for (tvert=the_A->tvert_link; (tvert); tvert=tvert->link)
	printf("(%p) <--> (%d)\n",tvert,TV_id(tvert));

    printf("\nTrapEdges\n");
    for (edge=the_A->edge_link; (edge); edge=edge->link)
	printf("(%p) <--> (%d)\n",edge,E_id(edge));

    printf("\nTraps\n");
    for (trap=the_A->trap_link; (trap); trap=trap->link)
	printf("(%p) <--> (%d)\n",trap,T_id(trap));

    printf("\nTrapHists\n");
    for (hist=the_A->hist_link; (hist); hist=hist->link)
	printf("(%p) <--> (%d)\n",hist,TV_id(hist));

    printf("\n");
}



Arrange_Print_Single(the_A,Print_Data)
  Arrange	*the_A;
  void		(*Print_Data)(FILE*,void*);
{
    TrapVert		*vert;
    TrapThreadVert	*tvert;
    TrapEdge		*edge;
    Trap		*trap;
    TrapHist		*hist;

    printf("\nTrapVerts\n");
    for (vert=the_A->vert_link; (vert); vert=vert->link)
	Print_TrapVert(the_A,vert,Print_Data);

    printf("\nTrapThreadVerts\n");
    for (tvert=the_A->tvert_link; (tvert); tvert=tvert->link)
	Print_TrapThreadVert(the_A,tvert);

    printf("\nTrapEdges\n");
    for (edge=the_A->edge_link; (edge); edge=edge->link)
	Print_TrapEdge(the_A,edge,Print_Data);

    printf("\nTraps\n");
    for (trap=the_A->trap_link; (trap); trap=trap->link)
	Print_Trap(the_A,trap);

    printf("\nTrapHists\n");
    for (hist=the_A->hist_link; (hist); hist=hist->link)
	Print_TrapHist(the_A,hist);

    printf("\n");
}



void Arrange_Print(the_A,Print_Data)
  Arrange		*the_A;
  Polygon_Data_Printer	Print_Data;
{
    Arrange_Print_Single(the_A,Print_Data);

    if (the_A->mirror && (the_A->mirror != the_A)) {
	printf("Arrangement Mirror:\n");
	Arrange_Print_Single(the_A->mirror,Print_Data);
    }
}



void Arrange_Draw_Graphics(Arrange *the_A, Boolean draw_threads)
{
#ifdef ARRANGE_GRAPHICS

    TrapVert		*vert;
    TrapThreadVert	*tvert;
    TrapEdge		*edge;
    Trap		*trap;
    TrapHist		*hist;
    double		scale;
    Vector		fake_edge,endpoint;
    Vector		v1,v2,v3;

    if (hardcopy)
	clear_debugger_screen_BW();
    else
	clear_debugger_screen();

    if (draw_threads) {
	for (tvert=the_A->tvert_link; (tvert); tvert=tvert->link) {
	    
	    Arrange_ThreadVert_Direction_And_Edge(the_A,tvert,
						  &fake_edge,&endpoint);
	    
	    /*--- if tvert is from above, we do NOT want 'upward' rep ---*/
	    if (TV_from_above(tvert))
		NEG(fake_edge);
	    
	    write_vector(&endpoint, (hardcopy ? db_black : db_red));
	    v1 = InqVertVector(V_direction(TV_source(tvert)));
	    write_arc(&v1, &endpoint, &fake_edge,
		      (hardcopy ? db_black : db_red),TRUE);
	}
    }

    for (vert=the_A->vert_link; (vert); vert=vert->link) {
	v1 = InqVertVector(V_direction(vert));
	write_vector(&v1, (hardcopy ? db_black : db_blue));
    }

    for (edge=the_A->edge_link; (edge); edge=edge->link) {
	v1 = InqVertVector(V_direction(E_lvert(edge)));
	v2 = InqVertVector(V_direction(E_rvert(edge)));
	v3 = InqEdgeVector(E_normal(edge));
	write_arc(&v1, &v2, &v3,
		  (hardcopy ? db_black : db_blue),FALSE);
    }

#endif
}


/*--- This relies on the fact that id's are from <1..num> ---*/
#define MARK_IND(x)  ((x)->id - 1)

Boolean Arrange_SanityCheck_Single(the_A)
  Arrange	*the_A;
{
    Boolean		failed;
    
    TrapVert		*vert1;
    TrapThreadVert	*tvert;
    TrapEdge		*edge;
    Trap		*trap;
    TrapHist		*hist;
    
    int			*count_trap;
    int			*count_vert;
    
    ArrangeCell		location;
    
    
    printf("\nStarting automated SanityCheck\n");
    
    failed = FALSE;
    
    count_vert = ALLOCATE_ARRAY(int,the_A->num_verts);
    count_trap = ALLOCATE_ARRAY(int,the_A->num_traps);
    
    /*---- set all vert/trap counters to zero to count threads ----*/
    for (vert1 = the_A->vert_link;  (vert1); vert1=V_link(vert1))
	count_vert[MARK_IND(vert1)]=0;
    for (trap=the_A->trap_link; (trap); trap=T_link(trap))
	count_trap[MARK_IND(trap)]=0;
    
    /*---- check tverts ----*/
    for (tvert=the_A->tvert_link; (tvert); tvert=TV_link(tvert)) {
	if (TV_source(tvert) && TV_edge(tvert) && TV_lTrap(tvert) &&
	    TV_rTrap(tvert)) {
	    
	    count_vert[MARK_IND(TV_source(tvert))]++;
	    count_trap[MARK_IND(TV_lTrap(tvert))]++;
	    count_trap[MARK_IND(TV_rTrap(tvert))]++;
	    
	    if (TV_from_above(tvert)) {
		if (T_be(TV_lTrap(tvert)) != TV_edge(tvert)) {
		    Print_Cell("Error: tvert", tvert,
			       "->lTrap->be != tvert->edge");
		    failed = TRUE;
		}
		if (T_be(TV_rTrap(tvert)) != TV_edge(tvert)) {
		    Print_Cell("Error: tvert", tvert,
			       "->rTrap->be != tvert->edge");
		    failed = TRUE;
		}
		if (T_brv(TV_lTrap(tvert)) != tvert) {
		    Print_Cell("Error: tvert", tvert,
			       " != tvert->lTrap->brv");
		    failed = TRUE;
		}
		if (T_blv(TV_rTrap(tvert)) != tvert) {
		    Print_Cell("Error: tvert", tvert,
			       " != tvert->rTrap->blv");
		    failed = TRUE;
		}
		if (T_brn(TV_lTrap(tvert)) != TV_rTrap(tvert)) {
		    Print_Cell("Error: tvert", tvert,
			       "->lTrap->brn != tvert->rTrap");
		    failed = TRUE;
		}
		if (T_bln(TV_rTrap(tvert)) != TV_lTrap(tvert)) {
		    Print_Cell("Error: tvert", tvert,
			       "->rTrap->bln != tvert->lTrap");
		    failed = TRUE;
		}
	    } else {
		if (T_te(TV_lTrap(tvert)) != TV_edge(tvert)) {
		    Print_Cell("Error: tvert", tvert,
			       "->lTrap->te != tvert->edge");
		    failed = TRUE;
		}
		if (T_te(TV_rTrap(tvert)) != TV_edge(tvert)) {
		    Print_Cell("Error: tvert", tvert,
			       "->rTrap->te != tvert->edge");
		    failed = TRUE;
		}
		if (T_trv(TV_lTrap(tvert)) != tvert) {
		    Print_Cell("Error: tvert", tvert,
			       " != tvert->lTrap->trv");
		    failed = TRUE;
		}
		if (T_tlv(TV_rTrap(tvert)) != tvert) {
		    Print_Cell("Error: tvert", tvert,
			       " != tvert->rTrap->tlv");
		    failed = TRUE;
		}
		if (T_trn(TV_lTrap(tvert)) != TV_rTrap(tvert)) {
		    Print_Cell("Error: tvert", tvert,
			       "->lTrap->trn != tvert->rTrap");
		    failed = TRUE;
		}
		if (T_tln(TV_rTrap(tvert)) != TV_lTrap(tvert)) {
		    Print_Cell("Error: tvert", tvert,
			       "->rTrap->tln != tvert->lTrap");
		    failed = TRUE;
		}
	    }
	    
	    
	    if (TV_lNeigh(tvert)) {
		if (TV_rNeigh(TV_lNeigh(tvert)) != tvert) {
		    Print_Cell("Error: tvert", tvert,
			       " != tvert->lNeigh->rNeigh.");
		    failed = TRUE;
		}
	    } else {
		if (E_lThrVert(TV_edge(tvert)) != tvert) {
		    Print_Cell("Error: tvert", tvert,
			       " != tvert->edge->lThrVert");
		    failed = TRUE;
		}
	    }
	    
	    if (TV_rNeigh(tvert)) {
		if (TV_lNeigh(TV_rNeigh(tvert)) != tvert) {
		    Print_Cell("Error: tvert", tvert,
			       " != tvert->rNeigh->lNeigh.");
		    failed = TRUE;
		}
	    } else {
		if (E_rThrVert(TV_edge(tvert)) != tvert) {
		    Print_Cell("Error: tvert", tvert,
			       " != tvert->edge->rThrVert");
		    failed = TRUE;
		}
	    }
	} else {
	    Print_Cell("Error: tvert", tvert,
		       " has NULL field.");
	    failed = TRUE;
	}
    }
    
    
    /*---- check edges ----*/
    for (edge=the_A->edge_link; (edge); edge=E_link(edge)) {
	if (!Upward(E_normal(edge))) {
	    Print_Cell("Error: edge", edge, "normal not upward");
	    failed = TRUE;
	}
	if (ON_HORIZON_EDGE(edge) ==
	    (OnHorizonEdge(E_normal(edge)))) {
	    Print_Cell("Warning: edge", edge,
		       "->normal does not match being on/off horizon");
	}
	if (E_tlTrap(edge) && E_trTrap(edge)) {
	    count_trap[MARK_IND(E_tlTrap(edge))]++;
	    count_trap[MARK_IND(E_trTrap(edge))]++;
	    
	    if (T_lvert(E_tlTrap(edge)) != E_lvert(edge)) {
		Print_Cell("Error: edge", edge,
			   "->tlTrap->lvert != edge->lvert");
		failed = TRUE;
	    }
	    if ((T_lvtype(E_tlTrap(edge)) != bot) &&
		(T_lvtype(E_tlTrap(edge)) != com)) {
		Print_Cell("Error: edge", edge,
			   "->tlTrap->lvtype != (bot|com)");
		failed = TRUE;
	    }
	    if (T_be(E_tlTrap(edge)) != edge) {
		Print_Cell("Error: edge", edge,
			   " != edge->tlTrap->be");
		failed = TRUE;
	    }
	} else {
	    if (!(ON_HORIZON_EDGE(edge))) {
		Print_Cell("Error: edge", edge,
			   " tlTrap or trTrap = NULL");
		failed = TRUE;
	    }
	}
	
	if (E_blTrap(edge) && E_brTrap(edge)) {
	    count_trap[MARK_IND(E_blTrap(edge))]++;
	    count_trap[MARK_IND(E_brTrap(edge))]++;
	    
	    if (T_lvert(E_blTrap(edge)) != E_lvert(edge)) {
		Print_Cell("Error: edge", edge,
			   "->blTrap->lvert != edge->lvert");
		failed = TRUE;
	    }
	    if ((T_lvtype(E_blTrap(edge)) != top) &&
		(T_lvtype(E_blTrap(edge)) != com)) {
		Print_Cell("Error: edge", edge,
			   "->blTrap->lvtype != (top|com)");
		failed = TRUE;
	    }
	    if (T_te(E_blTrap(edge)) != edge) {
		Print_Cell("Error: edge", edge,
			   " != edge->blTrap->te");
		failed = TRUE;
	    }
	} else {
	    if (!(ON_HORIZON_EDGE(edge))) {
		Print_Cell("Error: edge", edge,
			   " blTrap or brTrap = NULL");
		failed = TRUE;
	    }
	}
    }
    
    
    /*---- check verts ----*/
    for (vert1=the_A->vert_link; (vert1); vert1=V_link(vert1)) {
	if (!Front(V_direction(vert1))) {
	    Print_Cell("Error: vert", vert1, "direction not front");
	    failed = TRUE;
	}
	if (ON_HORIZON_VERTEX(vert1) ==
	    (OnHorizonVert(V_direction(vert1)))) {
	    Print_Cell("Warning: vert", vert1,
		       "->direction does not match being on/off horizon");
	}
	if ((V_tlTrap(vert1)) && (T_rvert(V_tlTrap(vert1)) != vert1)) {
	    Print_Cell("Error: vert", vert1,
		       " != vert->tlTrap->rvert");
	    failed = TRUE;
	}
	if ((V_brTrap(vert1)) && (T_lvert(V_brTrap(vert1)) != vert1)) {
	    Print_Cell("Error: vert", vert1,
		       " != vert->brTrap->lvert");
	    failed = TRUE;
	}
	if (((vert1 != the_A->tpole) && (vert1 != the_A->bpole)) &&
	    ( ON_HORIZON_VERTEX(vert1) ?
	     (count_vert[MARK_IND(vert1)] != 1) :
	     (count_vert[MARK_IND(vert1)] != 2))) {
	    Print_Cell("Error: vert", vert1,
		       "->count is wrong");
	    printf("count == %d\n",count_vert[MARK_IND(vert1)]);
	    failed = TRUE;
	}
    }
    
    
    /*---- check traps ----*/
    for (trap=the_A->trap_link; (trap); trap=T_link(trap)) {
	if (count_trap[MARK_IND(trap)] != 4) {
	    Print_Cell("Error: trap", trap,
		       "->count is wrong");
	    printf("count == %d\n",count_trap[MARK_IND(trap)]);
	    failed = TRUE;
	}
	
	if (Thread_Test(V_direction(T_lvert(trap)),
			V_direction(T_rvert(trap))) != ANS_LEFT) {
	    Print_Cell("Error: trap", trap,
		       " lvert not left of rvert");
	    failed = TRUE;
}
	
	switch (T_lvtype(trap)) {
	  case com:
	    if (E_lvert(T_be(trap)) != T_lvert(trap)) {
		Print_Cell("Error: trap", trap,
			   "->be->lvert != trap->lvert");
		failed = TRUE;
	    }
	    if (E_lvert(T_te(trap)) != T_lvert(trap)) {
		Print_Cell("Error: trap", trap,
			   "->te->lvert != trap->lvert");
		failed = TRUE;
	    }
	    if (T_tln(trap) || T_bln(trap)) {
		Print_Cell("Error: trap", trap,
			   "->(tln|bln) != NULL");
		failed = TRUE;
	    }
	    if (E_blTrap(T_te(trap)) != trap) {
		Print_Cell("Error: trap", trap,
			   " != trap->te->blTrap");
		failed = TRUE;
	    }
	    if (E_tlTrap(T_be(trap)) != trap) {
		Print_Cell("Error: trap", trap,
			   " != trap->be->tlTrap");
		failed = TRUE;
	    }
	    if (!EdgeEdge_Above(V_direction(T_lvert(trap)),
				E_normal(T_te(trap)), E_normal(T_be(trap)),
				TRUE)) {
		Print_Cell("Error: trap", trap,
			   " -- te not above be at leftvert");
		failed = TRUE;
	    }
	    break;
	    
	  case top:
	    if (E_lvert(T_te(trap)) != T_lvert(trap)) {
		Print_Cell("Error: trap", trap,
			   "->te->lvert != trap->lvert");
		failed = TRUE;
	    }
	    if (T_tln(trap) != T_bln(trap)) {
		Print_Cell("Error: trap", trap,
			   "->tln != trap->bln");
		failed = TRUE;
	    }
	    if (E_blTrap(T_te(trap)) != trap) {
		Print_Cell("Error: trap", trap,
			   " != trap->te->blTrap");
		failed = TRUE;
	    }
	    if (T_blv(trap)) {
		if (TV_rTrap(T_blv(trap)) != trap) {
		    Print_Cell("Error: trap", trap,
			       " != trap->blv->rTrap");
		    failed = TRUE;
		}
	    } else {
		Print_Cell("Error: trap", trap,
			   "->blv == NULL");
		failed = TRUE;
	    }
	    break;
	    
	  case bot:
	    if (E_lvert(T_be(trap)) != T_lvert(trap)) {
		Print_Cell("Error: trap", trap,
			   "->be->lvert != trap->lvert");
		failed = TRUE;
	    }
	    if (T_tln(trap) != T_bln(trap)) {
		Print_Cell("Error: trap", trap,
			   "->tln != trap->bln");
		failed = TRUE;
	    }
	    if (E_tlTrap(T_be(trap)) != trap) {
		Print_Cell("Error: trap", trap,
			   " != trap->be->tlTrap");
		failed = TRUE;
	    }
	    if (T_tlv(trap)) {
		if (TV_rTrap(T_tlv(trap)) != trap) {
		    Print_Cell("Error: trap", trap,
			       " != trap->tlv->rTrap");
		    failed = TRUE;
		}
	    } else {
		Print_Cell("Error: trap", trap,
			   "->tlv == NULL");
		failed = TRUE;
	    }
	    break;
	    
	  case mid:
	    if (T_blv(trap)) {
		if (TV_rTrap(T_blv(trap)) != trap) {
		    Print_Cell("Error: trap", trap,
			       " != trap->blv->rTrap");
		    failed = TRUE;
		}
	    } else {
		Print_Cell("Error: trap", trap,
			   "->blv == NULL");
		failed = TRUE;
	    }
	    if (T_tlv(trap)) {
		if (TV_rTrap(T_tlv(trap)) != trap) {
		    Print_Cell("Error: trap", trap,
			       " != trap->tlv->rTrap");
		    failed = TRUE;
		}
	    } else {
		Print_Cell("Error: trap", trap,
			   "->tlv == NULL");
		failed = TRUE;
	    }
	    break;
	}
	
	switch (T_rvtype(trap)) {
	  case com:
	    if (E_rvert(T_be(trap)) != T_rvert(trap)) {
		Print_Cell("Error: trap", trap,
			   "->be->rvert != trap->rvert");
		failed = TRUE;
	    }
	    if (E_rvert(T_te(trap)) != T_rvert(trap)) {
		Print_Cell("Error: trap", trap,
			   "->te->rvert != trap->rvert");
		failed = TRUE;
	    }
	    if (T_trn(trap) || T_brn(trap)) {
		Print_Cell("Error: trap", trap,
			   "->(trn|brn) != NULL");
		failed = TRUE;
	    }
	    if (E_brTrap(T_te(trap)) != trap) {
		Print_Cell("Error: trap", trap,
			   " != trap->te->brTrap");
		failed = TRUE;
	    }
	    if (E_trTrap(T_be(trap)) != trap) {
		Print_Cell("Error: trap", trap,
			   " != trap->be->trTrap");
		failed = TRUE;
	    }
	    if (!EdgeEdge_Above(V_direction(T_rvert(trap)),
				E_normal(T_te(trap)), E_normal(T_be(trap)),
				FALSE)) {
		Print_Cell("Error: trap", trap,
			   " -- te not above be at rightvert");
		failed = TRUE;
	    }
	    break;
	    
	  case top:
	    if (E_rvert(T_te(trap)) != T_rvert(trap)) {
		Print_Cell("Error: trap", trap,
			   "->te->rvert != trap->rvert");
		failed = TRUE;
	    }
	    if (T_trn(trap) != T_brn(trap)) {
		Print_Cell("Error: trap", trap,
			   "->trn != trap->brn");
		failed = TRUE;
	    }
	    if (E_brTrap(T_te(trap)) != trap) {
		Print_Cell("Error: trap", trap,
			   " != trap->te->brTrap");
		failed = TRUE;
	    }
	    if (T_brv(trap)) {
		if (TV_lTrap(T_brv(trap)) != trap) {
		    Print_Cell("Error: trap", trap,
			       " != trap->brv->lTrap");
		    failed = TRUE;
		}
	    } else {
		Print_Cell("Error: trap", trap,
			   "->brv == NULL");
		failed = TRUE;
	    }
	    break;
	    
	  case bot:
	    if (E_rvert(T_be(trap)) != T_rvert(trap)) {
		Print_Cell("Error: trap", trap,
			   "->be->rvert != trap->rvert");
		failed = TRUE;
	    }
	    if (T_trn(trap) != T_brn(trap)) {
		Print_Cell("Error: trap", trap,
			   "->trn != trap->brn");
		failed = TRUE;
	    }
	    if (E_trTrap(T_be(trap)) != trap) {
		Print_Cell("Error: trap", trap,
			   " != trap->be->trTrap");
		failed = TRUE;
	    }
	    if (T_trv(trap)) {
		if (TV_lTrap(T_trv(trap)) != trap) {
		    Print_Cell("Error: trap", trap,
			       " != trap->trv->lTrap");
		    failed = TRUE;
		}
	    } else {
		Print_Cell("Error: trap", trap,
			   "->trv == NULL");
		failed = TRUE;
	    }
	    break;
	    
	  case mid:
	    if (T_brv(trap)) {
		if (TV_lTrap(T_brv(trap)) != trap) {
		    Print_Cell("Error: trap", trap,
			       " != trap->brv->lTrap");
		    failed = TRUE;
		}
	    } else {
		Print_Cell("Error: trap", trap,
			   "->brv == NULL");
		failed = TRUE;
	    }
	    if (T_trv(trap)) {
		if (TV_lTrap(T_trv(trap)) != trap) {
		    Print_Cell("Error: trap", trap,
			       " != trap->trv->lTrap");
		    failed = TRUE;
		}
	    } else {
		Print_Cell("Error: trap", trap,
			   "->trv == NULL");
		failed = TRUE;
	    }
	    break;
	}
    }
    
    /*---- preform point location for each vertex ----*/
    for (vert1=the_A->vert_link; (vert1); vert1=V_link(vert1)) {
	location = Arrange_PointLocation(the_A,V_direction(vert1));
	if ((location.t != vert) ||
	    (location.d.vert != vert1)) {
	    Print_Cell("Error: Point Location for vert", vert1,
		       " failed");
	    failed = TRUE;
	}
    }
    
    if (!failed) printf("Test Passed\n");
    printf("\n");
    
    
    FREE(count_trap);
    FREE(count_vert);
    
    return(!failed);
}


Boolean Arrange_SanityCheck(the_A)
  Arrange	*the_A;
{
    /*- this any mirror arrangment independently, but does not
      check the consistency of the 'sewing together' part    -*/
    
    Arrange_SanityCheck_Single(the_A);
    
    if (the_A->mirror && (the_A->mirror != the_A)) {
	Arrange_SanityCheck_Single(the_A->mirror);
    }
}


Vector Get_InputPoint(Boolean is_vert)
{
    Vector		point;
    VertGeom*		t_vert;
    EdgeGeom*		t_edge;

    printf("Point: ");
    if ((scanf(" %lf %lf %lf",&point.x, &point.y, &point.z))!=3) {
	printf("Error reading input point...exiting...\n");
	exit(-1);
    }

    if (is_vert) {
	t_vert = CreateVertFromVector(point,TRUE);
	point = InqVertVector(t_vert);
	FreeVertGeom(t_vert);
    } else {
	t_edge = CreateEdgeFromVector(point,TRUE);
	point = InqEdgeVector(t_edge);
	FreeEdgeGeom(t_edge);
    }
    
    return(point);
}


enum ArrType {A_single, A_symmetric, A_double};

Arrange* Pick_Arrange_Init(type)
  enum ArrType	type;
{
    Arrange	*the_A;

    switch (type) {
      case A_single:
	the_A = Arrange_Init_Single();
	printf("Initializing Single Arrangement...\n");
	break;

      case A_symmetric:
	the_A = Arrange_Init_Symmetric();
	printf("Initializing Symmetric Arrangement...\n");
	break;

      case A_double:
	the_A = Arrange_Init_Double();
	printf("Initializing Double Arrangement...\n");
	break;
    }

    return(the_A);
}


Reset_CurPolys(cur_polys)
  List		*cur_polys;
{
    List	elt;

    for (elt=(*cur_polys); (elt); elt=NXT(elt)) {
	Polygon_Free(GET(elt,Polygon),TwoParts_Free);
    }
    delete_list((*cur_polys));
    (*cur_polys) = NULL;

}


/*************************************************************
  Function: DriverUsage()
  Input: <none>
  Output: <none>
  Effects: <none>
  
  **************************************************************/
#define LEN 70
static char usage[][LEN] = {
"                                                                     ",
"List of commands:                                                    ",
" ? --- this message                                                  ",
" q --- quit                                                          ",
" I --- reinialize arrangement                                        ",
" Z --- Zap current arrangement (without initializing new one)        ",
" i --- initialize from scratch                                       ",
" 1 --- set type to Single                                            ",
" 2 --- set type to Symmetric                                         ",
" 3 --- set type to Double                                            ",
" G --- Draw debugger graphics of current arrangment                  ",
" g --- Draw debugger graphics of arrangment mirror                   ",
" H --- Toggle debugger: full color/hardcopy                          ",
" p --- print entire arrangment                                       ",
" A --- Auto check for sanity                                         ",
" V --- toggle 'arrange_verbose' flag                                 ",
" F --- open inputfile used for reading polygons                      ",
" r --- Read in polygon from inputfile                                ",
" R --- Insert ALL remaining polygons from inputfile (given order)    ",
" L --- Insert ALL remaining polygons from inputfile (random order)   ",
" S --- print stack vertex address                                    ",
" v --- insert new vertex                                             ",
" e --- insert edge with two new vectors (and push end stack)         ",
" E --- insert edge from stack --> new vector                         ",
" h --- insert half circle with one horizon vector to antipode        ",
" s --- point location query  (i.e. search)                           ",
" T --- translate IDs from pointers                                   ",
" D --- set data print routine to be TwoParts_NicePrint               ",
"                                                                     "
};


DriverUsage()
{
    int lines,i,j;

    lines = sizeof(usage)/(LEN*sizeof(char));

    for (i=0; i<lines; i++) {
	for (j=0; j<LEN; j++) {
	    printf("%c",usage[i][j]);
	}
	printf("\n");
    }
}



/*************************************************************
  Function: Arrange_TestDriver()
  Input: <none>
  Output: <none>
  Effects: <none>
  
  Description: A test driver that allows simple way for user
  to send data to various routines such as the AddVertex
  or the point location query, or the total info dump.
  
  **************************************************************/
void Arrange_TestDriver()
{
    Arrange		*the_A;
    enum ArrType	Atype = A_symmetric;
    char		action;
    Vector		point1;
    Vector		point2;
    VertGeom*		t_vert;
    EdgeGeom*		t_edge;
    TrapVert		*stack;
    ArrangeCell		data;
    Boolean		done;
    Polygon		*poly;
    List		polylist;
    List		cur_polys;
    char		filename[50];
    FILE		*polyfp;
    Polygon_Data_Printer Print_Data;
    unsigned		seed;
    

    the_A = Pick_Arrange_Init(Atype);
    stack = NULL;
    polyfp = stdin;
    Print_Data = NULL;
    cur_polys = NULL;
    
    printf("Type '?' for list of commands\n");

    done = FALSE;
    while (!done) {
	printf("\nCommand: ");
	if ((scanf(" %c",&action))!=1) {
	    printf("Error reading command...\n");
	    action = ' ';
	    /*exit(-1);*/
	}
	
	switch(action) {
	  case 'q':	/* Quit */
	    printf("Quitting...\n");
	    done = TRUE;
	    break;
	    
	  case 'I':	/* re-Initialize */
	    printf("Reinitializing Arrange...\n");
	    Arrange_Free(the_A);
	    Reset_CurPolys(&cur_polys);
	    the_A = Pick_Arrange_Init(Atype);
	    stack = NULL;
	    break;
	    
	  case 'i':	/* Initialize from scratch */
	    printf("Initializing Arrange...\n");
	    the_A = Pick_Arrange_Init(Atype);
	    stack = NULL;
	    break;
	    
	  case 'T':	/* translate IDs */
	    Translate_IDs(the_A);
	    break;
	    
	  case 'p':	/* Print entire arrangement */
	    Arrange_Print(the_A, Print_Data);
	    break;
	    
	  case 'G':	/* Draw debugger graphics of arrangement */
	    Arrange_Draw_Graphics(the_A,show_threads);
	    break;
	    
	  case 'g':	/* Draw debugger graphics of arrangement mirror */
	    if (the_A->mirror)
		Arrange_Draw_Graphics(the_A->mirror,show_threads);
	    break;
	    
	  case 'H':	/* Toggle debugger: full color/hardcopy */
	    hardcopy = (!hardcopy);
	    break;
	    
	  case 'A':     /* Auto-check for sanity */
	    Arrange_SanityCheck(the_A);
	    break;
	    
	  case 's':	/* point location query */
	    point1 = Get_InputPoint(TRUE);

	    t_vert = CreateVertFromVector(point1,FALSE);
	    data = Arrange_PointLocation(the_A,t_vert);
	    FreeVertGeom(t_vert);
	    printf("Query (%lf,%lf,%lf): ",point1.x,point1.y,point1.z);
	    switch (data.t) {
	      case vert:
		Print_Cell("in vert",data.d.vert,".");
		break;
		
	      case edge:
		Print_Cell("in edge",data.d.edge,".");
		break;
		
	      case trap:
		Print_Cell("in trap",data.d.trap,".");
		break;
	    }
	    break;
	    
	  case 'v':	/* insert new vertex */
	    point1 = Get_InputPoint(TRUE);
	    printf("Inserting (%lf,%lf,%lf)\n",point1.x,point1.y,point1.z);
	    t_vert = CreateVertFromVector(point1,FALSE);
	    data = Arrange_PointLocation(the_A,t_vert);
	    stack = Arrange_AddVertex(the_A,t_vert,NULL,data);
	    FreeVertGeom(t_vert);
	    break;
	    
	  case 'S':	/* print stack vertex address */
	    printf("Stack vertex(");
	    Print_VertID(stack);
	    printf(").\n");
	    break;
	    
	  case 'e':	/* insert new edge with two new vectors */
	    /* push resulting endpoint on stack */
	    point1 = Get_InputPoint(TRUE);
	    printf("Endpoint1 (%lf,%lf,%lf)\n",point1.x,point1.y,point1.z);
	    t_vert = CreateVertFromVector(point1,TRUE);
	    data = Arrange_PointLocation(the_A,t_vert);
	    stack = Arrange_AddVertex(the_A,t_vert,NULL,data);
	    FreeVertGeom(t_vert);
	    point2 = Get_InputPoint(TRUE);
	    printf("Endpoint2 (%lf,%lf,%lf)\n",point2.x,point2.y,point2.z);
	    t_vert = CreateVertFromVector(point2,FALSE);
	    t_edge = EdgeThroughVerts(V_direction(stack),t_vert,NULL);
	    stack =
		Arrange_AddEdge(the_A, stack, NULL, t_vert,t_edge,
				NULL,NULL,NULL);
	    FreeVertGeom(t_vert);
	    FreeEdgeGeom(t_edge);
	    break;
	    
	  case 'E':	/* insert new edge with one new vector */
	    /* using stack as START endpoint */
	    /* push resulting endpoint on stack */
	    if (stack) {
		point1 = InqVertVector(V_direction(stack));
		printf("Endpoint1 (%lf,%lf,%lf)\n", point1.x,
		       point1.y, point1.z);
		
		point2 = Get_InputPoint(TRUE);
		printf("Endpoint2 (%lf,%lf,%lf)\n",point2.x,point2.y,point2.z);
		t_vert = CreateVertFromVector(point2,FALSE);
		t_edge = EdgeThroughVerts(V_direction(stack),t_vert,NULL);
		stack =
		    Arrange_AddEdge(the_A, stack, NULL, t_vert, t_edge,
				    NULL,NULL,NULL);
		FreeVertGeom(t_vert);
		FreeEdgeGeom(t_edge);
	    }
	    break;
	    
	  case 'h':	/* insert a half circle with one new
			   vector and it's antipode */
	    point1 = Get_InputPoint(TRUE);
	    t_vert = CreateVertFromVector(point1,FALSE);
	    t_edge = GetHorizonGeom(FALSE);
	    SnapVertToEdge(t_vert,t_edge,NULL,NULL);
	    FreeEdgeGeom(t_edge);
	    printf("Endpoint1 (%lf,%lf,%lf)\n",point1.x,point1.y,point1.z);
	    data = Arrange_PointLocation(the_A,t_vert);
	    stack = Arrange_AddVertex(the_A,t_vert,NULL,data);
	    FreeVertGeom(t_vert);
	    point2 = Get_InputPoint(FALSE);
	    printf("Normal (%lf,%lf,%lf)\n",point2.x,point2.y,point2.z);
	    t_vert = CreateVertFromVector(point2,FALSE);
	    stack =
		Arrange_AddEdge(the_A, stack, V_mirror(stack),
				V_direction(V_mirror(stack)),
				t_vert, NULL,NULL,NULL);
	    FreeVertGeom(t_vert);
	    break;
	    
	  case 'F':	/* open inputfile for polygon reading */
	    if ((polyfp) && (polyfp!=stdin)) fclose(polyfp);
	    printf("filename: ");
	    scanf(" %s",filename);
	    if ((polyfp=fopen(filename,"r"))==NULL) {
		printf("Couldn't open file...using stdin\n");
		polyfp = stdin;
	    }
	    break;
	    
	  case 'r':	/* Read in polygon from file and add to arrange */
	    poly = Polygon_Read(polyfp,TwoParts_Read);
	    if (poly) {
		printf("\nAdding: ");
		TwoParts_NicePrint(stdout,poly->data);
		printf("\n");
		Arrange_AddPolygon(the_A,poly);
		cur_polys = push(cur_polys,poly);
	    }
	    break;
	    
	  case 'R':	/* Read to end of file */
	    printf("Reading to end of file...\n");
	    while ((poly = Polygon_Read(polyfp,TwoParts_Read))!=NULL) {
		printf("\nAdding: ");
		TwoParts_NicePrint(stdout,poly->data);
		printf("\n");
		Arrange_AddPolygon(the_A,poly);
		cur_polys = push(cur_polys,poly);
	    }
	    break;
	    
	  case 'L':	/* Read to end of file & randomize list */
	    printf("Reading to end of file...\n");
	    polylist = NULL;
	    while ((poly = Polygon_Read(polyfp,TwoParts_Read))!=NULL) {
		polylist = push(polylist,poly);
	    }
	    polylist = reverse(polylist,NULL);

	    printf("Seed: ");
	    scanf(" %ld",&seed);
	    srandom(seed);
	    Arrange_AddListOfPolygons(the_A,polylist,TRUE);
	    cur_polys = concatenate(cur_polys,polylist);
	    break;
	    
	  case 'D':	/* Set data print to TwoParts_NicePrint */
	    Print_Data = (Polygon_Data_Printer) TwoParts_NicePrint;
	    break;
	    
	  case 'V':	/* Toggle arrange_verbose flag */
	    arrange_verbose = (!arrange_verbose);
	    break;
	    
	  case 'Z':	/* zap current arrangement */
	    printf("Freeing Arrangement...\n");
	    Arrange_Free(the_A);
	    Reset_CurPolys(&cur_polys);
	    the_A = NULL;
	    break;
	    
	  case '1':	/* set type to Single */
	    Atype = A_single;
	    break;
	    
	  case '2':	/* set type to Symmetric */
	    Atype = A_symmetric;
	    break;
	    
	  case '3':	/* set type to Double */
	    Atype = A_double;
	    break;

	  case '?':	/* usage */
	    DriverUsage();
	    break;

	  default:
	    printf("unrecognized command: '%c'\n",action);
	}
	
    }
}

#endif   /* ARRANGE_DEBUG */
    
