/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * arrange_geom.h
 *
 * Abstract definitions for geometric features
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/


/*===============================================================*/
/*=								=*/
/*=		Geometric Primitives on the Sphere		=*/
/*=							   	=*/
/*===============================================================*/

/*************************************************************
  This file is used as a toolkit for geometric tests needed
  by an application.  It has some simple comparison routines
  such as "above/below" or "left/right" as well as some more
  complicated routines.

  This specific file deals with points on the unit sphere in R^3.
  If necessary, a similar file could be made with the same
  functionality for planar geometry.

  ****************************************************************
                    Spherical Model
  ****************************************************************
  Organizationally, this module thinks of the sphere as two
  hemispheres, the "front" and "back", based on whether the
  z-coordinate is positive or negative, and they join at the
  "horizon" (z=0), which belongs to both hemispheres.  And
  ALL of the functions here deal with only ONE of the hemispheres
  at a time.  (so in fact, this is really a file handling the
  geometry of a hemisphere!)

  For piece of mind, although all facets are on the sphere, we will
  define terms such as "up", "down", "above", "below", "left", and
  "right", and use them consistently throughout the implementation.

  The directional terms are based on the following:
    "top" pole    --- the vector (0,1,0)    <positive Y-axis>
    "bottom" pole --- the vector (0,-1,0)   <negative Y-axis>
    "vertical thread" --- a great circal passing through the poles.
    "horizon" ---- the great circle with normal (0,0,1) or (0,0,-1)

  ****************************************************************
                    Homogeneous Coordinates
  ****************************************************************
  Although the calculations in this module will be done efficiently
  using straight-forward vector calculus, there is a correspondance
  between the spherical geometry and the plane geometry that is
  based on the common theory of homogeneous coordinates.  We refer
  to a directional vector as a "point", and we refer to a great
  circle as an "edge".  Note that a "point" is simply the
  intersection between a ray centered at the origin and the sphere,
  and an "edge" is the intersection between a plane through the
  origin and the sphere.

  If the (x,y,z) coordinates of the sphere are considered to be
  homogeneous with the z-coord being the weight, then the hemisphere
  can be mapped to the plane simply by projecting from the origin
  onto the Z=1 plane.   That is the (x,y,z) cooridnates are mapped to
  the planar point (X,Y) where X=x/z  and Y=y/z.

  Under this mapping, all of our edges become line segments in the
  XY-plane, and all the directional notions defined earlier become
  exactly the standard definitions in the plane.


  ****************************************************************
                  Degeneracy Conventions
  ****************************************************************
  In addition, this file handles several degeneracies robustly, by
  relying on a lexicographical perturbation.  Most importantly,
  any two distinct points that are 'vertical' are perturbed so
  that the "higher" one is treated as farther right of the other.

  A thread is an ideal vertical line through a specific point,
  the source, and threads are treated as the only truly vertical
  features.

  To be consistent with this pertubation, all edges that are
  vertical are treated as if they were perturbed slightly so that
  they head slightly left to right as they go upward.  Therefore,
  points to the left of the edge are considered "above" the edge,
  and points to the right are considered "below" the edge.

  As a special case of this rule, edges involving the horizon
  may take two forms.  An edge lying on the "right" horizion
  will have an upward pointing normal, (0,0,1), so that all
  other Front points will appear as if they are above the edge.
  For this reason the right horizon is often refered to
  synonomously as the "bottom" horizon.

  Similarly, all edges lying on the "left" horizion will have the
  upward normal (0,0,-1) so that all other Front points will appear
  as if they are below the edge.  For this reason, the left horizon
  is sometimes refered to synonomously as the "top" horizon.

  One final note:  there is a function that compares two points
  (or two edges), and decides if they are "equal"  or compares a
  point and an edge, and decides if the point "lies" on the edge.
  With real numbers, all of these test boil down to a comparison
  with 0.0, however because of the inherent inacuracy with floating
  point numbers, this function can be called with a flag 'exact'
  that can decide whether to compare with 0.0 or EPS_1, a small
  constant.


*************************************************************/


#ifndef _ARRANGE_GEOM
#define _ARRANGE_GEOM

#include "standards.h"
#include <math.h>
#include "spheregeom.h"

typedef void VertGeom;
typedef void EdgeGeom;


/*   int codes for answers to geometric primitive test */
#define ANS_RIGHT  0
#define ANS_LEFT   1

#define ANS_ABOVE  0
#define ANS_BELOW  1


/*===============================================================*/
/*=								=*/
/*=             Geometric Test Functions			=*/
/*=								=*/
/*===============================================================*/

/*************************************************************
  Function: OnHorizonVert(V)
  Input:
      V --- VertGeom for point in question
  Output:
      TRUE --- V is on the horizon
      FALSE --- V is not on the horizon
**************************************************************/
Boolean OnHorizonVert(VertGeom *);


/*************************************************************
  Function: OnPoleVert(V)
  Input:
      V --- VertGeom for point in question
  Output:
      TRUE --- V is on a pole
      FALSE --- V is not on a pole
**************************************************************/
Boolean OnPoleVert(VertGeom *);


/*************************************************************
  Function: OnHorizonEdge(E)
  Input:
      E --- EdgeGeom for edge in question
  Output:
      TRUE --- E is on the horizon
      FALSE --- E is not on the horizon
**************************************************************/
Boolean OnHorizonEdge(EdgeGeom *);



/*************************************************************
  Function: VertVertIncedent(U,V)
  Input:
      U --- VertGeom for point
      V --- VertGeom for point
  Output:
      TRUE --- U and V are considered the same
      FALSE --- U and V are not considered the same
**************************************************************/
Boolean VertVertIncedent(VertGeom *, VertGeom *);


/*************************************************************
  Function: VertEdgeIncedent(V,E)
  Input:
      V --- VertGeom for point
      E --- EdgeGeom for edge
  Output:
      TRUE --- V is considered to lie on E
      FALSE --- V is not considered to lie on E
**************************************************************/
Boolean VertEdgeIncedent(VertGeom *, EdgeGeom *);


/*************************************************************
  Function: EdgeEdgeIncedent(E,F)
  Input:
      E --- EdgeGeom for edge
      F --- EdgeGeom for edge
  Output:
      TRUE --- E and F are considered to be the same
      FALSE --- E and F are not considered to be the same
**************************************************************/
Boolean EdgeEdgeIncedent(VertGeom *, VertGeom *);


/*************************************************************
  Function:  Front(v)
  Input:
      V ---  VertGeom in question
  Output:
      TRUE if V is in front hemisphere
      FALSE if V is not in front hemisphere
**************************************************************/
Boolean Front(VertGeom *);


/*************************************************************
  Function:  Upward(E)
  Input:
      E --- EdgeGeom in question
  Output:
      TRUE if E has an "upward" normal
      FALSE if E has a "downward" normal
      <for vertical E, conventionally left is "upward">
**************************************************************/
Boolean Upward(EdgeGeom *);


/*************************************************************
  Function:  VertOnLeftHorizon(V)
  Input:
      V --- VertGeom in question
  Assumptions:  V "on" horizon
  Output:
      TRUE if V is on left horizon (or bottom pole)
      FALSE if V is on right horizon (or top pole)
**************************************************************/
Boolean VertOnLeftHorizon(VertGeom *);


/*************************************************************
  Function:  EdgeOnLeftHorizon(E)
  Input:
      E --- EdgeGeom in question
  Assumptions:  edge E part of horizon
  Output:
      TRUE if E is on left side of horizon
      FALSE if E is on right side of horizon
**************************************************************/
Boolean EdgeOnLeftHorizon(EdgeGeom*);


/*************************************************************
  Function:  Thread_Test(P,Q,front,unit)
  Input:
      P --- VertGeom
      Q--- VertGeom
  Output:
      ANS_RIGHT == point P is on/right of thread through Q
      ANS_LEFT  == point P is left of thread through Q
      <note: as described in the introduction, we perturb
             lexicographically, so that P only lies on a
	     thread if it is exactly equal to Q>
**************************************************************/
int Thread_Test(VertGeom*, VertGeom*);



/*************************************************************
  Function:  Edge_Test(P,E,front,unit)
  Input:
      P ---  VertGeom
      E ---  EdgeGeom
  Output:
      ANS_ABOVE ==  point P is on/above edge E
      ANS_BELOW ==  point P is below edge E
**************************************************************/
int  Edge_Test(VertGeom*, EdgeGeom*);



/*************************************************************
  Function:  EdgeEdge_Above(T,E1,E2,right)
  Input:
      T --- VertGeom sitting at intersection of edges
      E1 --- EdgeGeom for first edge
      E2 --- EdgeGeom for second edge
      right --- interest in relationship to right or left of T?
  Output:
    TRUE  ==  E1 is "above" E2 (or EXACTLY equal)
    FALSE ==  E1 is "below" E2
    <where above is relative to the requested side>

**************************************************************/
Boolean EdgeEdge_Above(VertGeom*, EdgeGeom*,
		       EdgeGeom*, Boolean);


/*===============================================================*/
/*=								=*/
/*=                     Inquire functions			=*/
/*=								=*/
/*===============================================================*/

/*************************************************************
  Function: InqVertVector(V)
  Input:
      V --- VertGeom
  Output:
      vector --- vector of point being represented by V
**************************************************************/
Vector InqVertVector(VertGeom*);


/*************************************************************
  Function: InqEdgeVector(E)
  Input:
      E --- EdgeGeom
  Output:
      vector --- vector of edge being represented by E
**************************************************************/
Vector InqEdgeVector(EdgeGeom*);



/*===============================================================*/
/*=								=*/
/*=             Data Modifying/Generating Functions		=*/
/*=								=*/
/*===============================================================*/

/*************************************************************
  Function: FreeVertGeom(V)
  Input:
      V --- VertGeom
  Output: <none>
**************************************************************/
void FreeVertGeom(VertGeom*);


/*************************************************************
  Function: FreeEdgeGeom(E)
  Input:
      E --- EdgeGeom
  Output: <none>
**************************************************************/
void FreeEdgeGeom(EdgeGeom*);


/*************************************************************
  Function: GetPoleGeom(bottom)
  Input:
      bottom ---  interested in top pole or bottom pole?
  Output:
      VertGeom --- new VertGeom representing the desired pole
**************************************************************/
VertGeom*  GetPoleGeom(Boolean);


/*************************************************************
  Function: GetHorizonGeom(left)
  Input:
      left---  interested in left horizon or right horizon?
  Output:
      EdgeGeom --- new EdgeGeom representing the desired horizon
**************************************************************/
EdgeGeom*  GetHorizonGeom(Boolean);


/*************************************************************
  Function: CopyVertGeom(V)
  Input:
      V --- VertGeom to copy
  Output:
      VertGeom --- new VertGeom structure
**************************************************************/
VertGeom*  CopyVertGeom(VertGeom*);


/*************************************************************
  Function: CopyEdgeGeom(E)
  Input:
      E --- EdgeGeom to copy
  Output:
      EdgeGeom --- new EdgeGeom structure
**************************************************************/
EdgeGeom*  CopyEdgeGeom(EdgeGeom*);


/*************************************************************
  Function: NegateVertGeom(V)
  Input:
      V --- VertGeom to Negate
  Output:
      <none>
  Effects:  negates the VertGeom

**************************************************************/
void NegateVertGeom(VertGeom*);


/*************************************************************
  Function: NegateEdgeGeom(e)
  Input:
      E --- EdgeGeom to Negate
  Output:
      <none>
  Effects:  negates the EdgeGeom

**************************************************************/
void NegateEdgeGeom(EdgeGeom*);


/*************************************************************
  Function: CreateVertFromVector(vec, orient)
  Input:
      vec ---  Vector to use as input data
      orient --- force VertGeom to front?
  Output:
      VertGeom -- newly created VertGeom

**************************************************************/
VertGeom*  CreateVertFromVector(struct Vector, Boolean);


/*************************************************************
  Function: CreateEdgeFromVector(vec, orient)
  Input:
      vec ---  Vector to use as input data
      orient --- force EdgeGeom to upward?
  Output:
      EdgeGeom -- newly created EdgeGeom

**************************************************************/
EdgeGeom*  CreateEdgeFromVector(struct Vector, Boolean);


/*************************************************************
  Function:  EdgeThroughVerts(U,V,E)
  Input:   
     U --- VertGeom
     V --- VertGeom
     E --- EdgeGeom
  Assumptions:  U & V not identical/anitpodal.  E is a "suggestion"
  Output:
    EdgeGeom --- the unique upward edge passing through both points.

  Description:
     This routine returns a new edge E satifsyfing,
        VertEdgeIncedent(U,E) == TRUE
        VertEdgeIncedent(V,E) == TRUE

**************************************************************/
EdgeGeom* EdgeThroughVerts(VertGeom*, VertGeom*, EdgeGeom*);


/*************************************************************
  Function:  EdgeEdgeIntersect(E1,E2,L1,R1,L2,R2)
  Input:   
      E1 --- first edge's EdgeGeom
      E2 --- second edge's EdgeGeom

  Optional Input:   (if NULL it will be ignored)
      L1 --- VertGeom for a left boudary for edge segment E1
      R1 --- VertGeom for a right boudary for edge segment E1
      L2 --- VertGeom for a left boudary for edge segment E2
      R2 --- VertGeom for a right boudary for edge segment E2
  Assumptions:  E1 & E2 not identical
  Output:
    VertGeom --- the unique front point that lies on both edges.

  Description:
     This routine returns a new VertGeom V satifsyfing,
        VertEdgeIncedent(V,E1) == TRUE
        VertEdgeIncedent(V,E2) == TRUE
	ThreadTest(V,L1) == ANS_RIGHT
	ThreadTest(V,L2) == ANS_RIGHT
	ThreadTest(R1,V) == ANS_RIGHT
	ThreadTest(R2,V) == ANS_RIGHT

**************************************************************/
VertGeom* EdgeEdgeIntersect(EdgeGeom*, EdgeGeom*, VertGeom*,
			    VertGeom*, VertGeom*, VertGeom*);


/*************************************************************
  Function:  SnapVertToEdge(V, E, L1, R1)
  Input:
    V --- VertGeom represnting a point
    E --- EdgeGeom represnting an edge

  Optional Input:   (if NULL it will be ignored)
    L1 --- VertGeom for a left boudary for edge segment E
    R1 --- VertGeom for a right boudary for edge segment E
  Assumptions:  V is not parallel to E
  Output: <none>
  Effects:
     V is projected so that by end of routine,
        VertEdgeIncedent(V,E) == TRUE

     And if the projection lies between L1 and R1 on E, then
	ThreadTest(V,L1) == ANS_RIGHT
	ThreadTest(R1,V) == ANS_RIGHT

  Description:
      Given a point, presumably "near" an edge, this function
      projects the point so that it lies "exactly" on the edge.
      (that is, as close as FP precision will allow)

**************************************************************/
void SnapVertToEdge(VertGeom*, EdgeGeom*, VertGeom*, VertGeom*);



/*************************************************************
  Function:  SnapEdgeToVert(E,V)
  Input:
      E --- EdgeGeom to adjust
      V --- VertGeom
  Assumptions:  V is not parallel to E
  Output: <none>
  Effects:
     E is projected so that by end of routine,
        VertEdgeIncedent(V,E) == TRUE

  Description:
      Given an edge, presumably "near" a point, this function
      projects the edge so that it goes "exactly" through the point.
      (that is, as close as FP precision will allow)

**************************************************************/
void SnapEdgeToVert(EdgeGeom*, VertGeom*);


#endif

