/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * arrange_geom_float.c
 *
 * Floating point implementation of geometric features
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/


/*===============================================================*/
/*=								=*/
/*=		Geometric Primitives on the Sphere		=*/
/*=							   	=*/
/*===============================================================*/

#include <math.h>
#include <values.h>
#include "arrange_geom.h"
#include "spheregeom.h"
#include "standards.h"

/*===============================================================*/
/*=								=*/
/*=                       EdgeGeom/VertGeom			=*/
/*=								=*/
/*=	This typedef is used as the VertGeom structure		=*/
/*===============================================================*/
typedef struct Vert {
    Vector	d;
    float	longit;		/*  X=x/z result */
    int		epsilon;	/* tie breaker! */
} Vert;

typedef struct Edge {
    Vector	n;
} Edge;


#define SMALL .001

/*===============================================================*/
/*=								=*/
/*=             Geometric Test Functions			=*/
/*=								=*/
/*===============================================================*/

/*************************************************************
  Function: OnHorizonVert(V)
  Input:
      V --- VertGeom for point in question
  Output:
      TRUE --- V is on the horizon
      FALSE --- V is not on the horizon
**************************************************************/
Boolean OnHorizonVert(V)
  VertGeom*	V;
{
    return(ON_HORIZON_VEC(((Vert*) V)->d));
}



/*************************************************************
  Function: OnPoleVert(V)
  Input:
      V --- VertGeom for point in question
  Output:
      TRUE --- V is on a pole
      FALSE --- V is not on a pole
**************************************************************/
Boolean OnPoleVert(V)
  VertGeom*	V;
{
    return(ON_POLE_VEC(((Vert*) V)->d));
}



/*************************************************************
  Function: OnHorizonEdge(E)
  Input:
      E --- EdgeGeom for edge in question
  Output:
      TRUE --- E is on the horizon
      FALSE --- E is not on the horizon
**************************************************************/
Boolean OnHorizonEdge(E)
  EdgeGeom*	E;
{
    return(ON_HORIZON_NORM(((Edge*) E)->n));
}



/*************************************************************
  Function: VertVertIncedent(U,V)
  Input:
      U --- VertGeom for point
      V --- VertGeom for point
  Output:
      TRUE --- U and V are considered the same
      FALSE --- U and V are not considered the same
**************************************************************/
Boolean VertVertIncedent(U,V)
  VertGeom*	U;
  VertGeom*	V;
{
    Boolean result;
    Vert*	u;
    Vert*	v;

    u = (Vert*) U;
    v = (Vert*) V;

    result = VectorTest(u->d, v->d, FALSE);

    result = result &&
	(ON_HORIZON_VEC(u->d) == ON_HORIZON_VEC(v->d)) &&
	    (ON_POLE_VEC(u->d) == ON_POLE_VEC(v->d));

    return(result);
}



/*************************************************************
  Function: VertEdgeIncedent(V,E)
  Input:
      V --- VertGeom for point
      E --- EdgeGeom for edge
  Output:
      TRUE --- V is considered to lie on E
      FALSE --- V is not considered to lie on E
**************************************************************/
Boolean VertEdgeIncedent(V,E)
  VertGeom*	V;
  EdgeGeom*	E;
{
    Boolean result;
    Edge*	e;
    Vert*	v;

    e = (Edge*) E;
    v = (Vert*) V;


    result = VectorTest(v->d, e->n, TRUE);

    result = result &&
	(!ON_HORIZON_NORM(e->n) ||
	 (ON_HORIZON_VEC(v->d) &&
	  (ON_POLE_VEC(v->d) ||
	   (VertOnLeftHorizon(v) == EdgeOnLeftHorizon(e)))));

    return(result);
}



/*************************************************************
  Function: EdgeEdgeIncedent(E,F)
  Input:
      E --- EdgeGeom for edge
      F --- EdgeGeom for edge
  Output:
      TRUE --- E and F are considered to be the same
      FALSE --- E and F are not considered to be the same
**************************************************************/
Boolean EdgeEdgeIncedent(E,F)
  EdgeGeom*	E;
  EdgeGeom*	F;
{
    Boolean result;
    Edge*	e;
    Edge*	f;

    e = (Edge*) E;
    f = (Edge*) F;

    result = VectorTest(((Edge*) E)->n, ((Edge*) F)->n,FALSE);

    result = result &&
	(ON_HORIZON_NORM(e->n) == ON_HORIZON_NORM(f->n));

    return(result);

}


/*************************************************************
  Function:  Front(v)
  Input:
      V ---  VertGeom in question
  Output:
      TRUE if V is in front hemisphere
      FALSE if V is not in front hemisphere
**************************************************************/
Boolean Front(V)
  VertGeom* V;
{
    Vert* v;

    v = (Vert*) V;

    return(v->d.z >= 0.0);
}



/*************************************************************
  Function:  Upward(E)
  Input:
      E --- EdgeGeom in question
  Output:
      TRUE if E has an "upward" normal
      FALSE if E has a "downward" normal
      <for vertical E, conventionally left is "upward">
**************************************************************/
Boolean Upward(E)
  EdgeGeom* E;
{
    Edge* e;

    e = (Edge*) E;
    return(
	   ((e->n.y>0.0) ||
	    ((e->n.y==0.0) && (e->n.x <= 0.0)))
	   );
}


/*************************************************************
  Function:  VertOnLeftHorizon(V)
  Input:
      V --- VertGeom in question
  Assumptions:  V "on" horizon
  Output:
      TRUE if V is on left horizon (or bottom pole)
      FALSE if V is on right horizon (or top pole)
**************************************************************/
Boolean VertOnLeftHorizon(VertGeom* V)
{
    Vert* v;

    v = (Vert*) V;

    return(
	   ((v->d.x < 0.0) ||
	    ((v->d.x == 0.0) && (v->d.y < 0.0)))
	   );
}



/*************************************************************
  Function:  EdgeOnLeftHorizon(E)
  Input:
      E --- EdgeGeom in question
  Assumptions:  edge E part of horizon
  Output:
      TRUE if E is on left side of horizon
      FALSE if E is on right side of horizon
**************************************************************/
Boolean EdgeOnLeftHorizon(EdgeGeom* E)
{
    return((((Edge*) E)->n.z < 0.0));
}



/*************************************************************
  Function:  Thread_Test(P,Q,front,unit)
  Input:
      P --- VertGeom
      Q--- VertGeom
  Output:
      ANS_RIGHT == point P is on/right of thread through Q
      ANS_LEFT  == point P is left of thread through Q
      <note: as described in the introduction, we perturb
             lexicographically, so that P only lies on a
	     thread if it is exactly equal to Q>
**************************************************************/
int Thread_Test(VertGeom* P,VertGeom* Q)
{
    int		answer;
    double	store;
    Boolean	p_right;
    Boolean	q_right;
    Vert*	p;
    Vert*	q;

    p = (Vert*) P;
    q = (Vert*) Q;

    if (ON_HORIZON_VEC(p->d)) {
	p_right = (p->d.x > 0.0) || ((p->d.x == 0.0) && (p->d.y > 0.0));
	if (ON_HORIZON_VEC(q->d)) {
	    q_right = (q->d.x > 0.0) || ((q->d.x == 0.0) && (q->d.y > 0.0));
	    /*---- both P & Q on horizon ----*/
	    if (p_right != q_right) {
		/* P & Q on opposite horizons */
		answer = (p_right ?
			  /** P on right, Q on left **/
			  ANS_RIGHT :
			  /** P on left, Q on right **/
			  ANS_LEFT);
	    } else {
		/* P & Q on same horizon */
		answer = ((p->d.y >= q->d.y) ?
			  ANS_RIGHT : ANS_LEFT);
	    }
	} else {

	    /*---- just P on horizon ----*/
	    answer = (p_right ?
		      /** P on right horizon **/
		      ANS_RIGHT :
		      /** P on left horizon **/
		      ANS_LEFT);
	}
    } else if (ON_HORIZON_VEC(q->d)) {
	q_right = (q->d.x > 0.0) || ((q->d.x == 0.0) && (q->d.y > 0.0));
	/*---- just Q on horizon ----*/
	answer = (q_right ?
		  /** Q on right horizon **/
		  ANS_LEFT :
		  /** Q on left horizon **/
		  ANS_RIGHT);
    } else {


	/*---- neither on horizon ----*/
	store = (p->longit - q->longit);
	if (store > 0.0) {
	    answer = ANS_RIGHT;
	} else if (store < 0.0) {
	    answer = ANS_LEFT;
	} else {	/* "vertical" -- must perturb */
	    if ((p->epsilon - q->epsilon) > 0) {
		answer = ANS_RIGHT;
	    } else if ((p->epsilon - q->epsilon) < 0) {
		answer = ANS_LEFT;
	    } else {	/* REALLY vertical, even after epsilon */
		answer = ((p->d.y >= q->d.y) ?
			  ANS_RIGHT : ANS_LEFT);
	    }
	}
    }

    return(answer);
}


/*************************************************************
  Function:  Edge_Test(P,E,front,unit)
  Input:
      P ---  VertGeom
      E ---  EdgeGeom
  Output:
      ANS_ABOVE ==  point P is on/above edge E
      ANS_BELOW ==  point P is below edge E
**************************************************************/
int  Edge_Test(VertGeom* P, EdgeGeom* E)
{
    int		answer;
    double	store;
    Vert*	p;
    Edge*	e;

    p = (Vert*) P;
    e = (Edge*) E;

    if ((store=DOT(p->d,e->n)) > 0.0) {
	answer = ANS_ABOVE;
    } else if (store < 0.0) {
	answer = ANS_BELOW;
    } else {	/* on edge .. check for weird horizon cases */
	if (!ON_HORIZON_NORM(e->n)) {
	    answer = ANS_ABOVE;
	} else {
	    answer = (((p->d.x > 0.0) || ((p->d.x == 0.0) && (p->d.y > 0.0))) ?
		      ((e->n.z > 0.0) ?
		       /** both P & E refer to bot horizon **/
		       ANS_ABOVE :
		       /** P is bot, E is top **/
		       ANS_BELOW) :
		      /** P is top, E is either **/
		      ANS_ABOVE);

	}
    }

    return(answer);
}


/*************************************************************
  Function:  EdgeEdge_Above(T,E1,E2,right)
  Input:
      T --- VertGeom sitting at intersection of edges
      E1 --- EdgeGeom for first edge
      E2 --- EdgeGeom for second edge
      right --- interest in relationship to right or left of T?
  Output:
    TRUE  ==  E1 is "above" E2 (or EXACTLY equal)
    FALSE ==  E1 is "below" E2
    <where above is relative to the requested side>

**************************************************************/
Boolean EdgeEdge_Above(VertGeom* T, EdgeGeom* E1, EdgeGeom* E2,
		       Boolean right)
{
    Boolean	test;
    Vert*	t;
    Edge*	e1;
    Edge*	e2;
    

    t = (Vert*) T;
    e1 = (Edge*) E1;
    e2 = (Edge*) E2;

    test = (ON_HORIZON_VEC(t->d) ?
	    (e1->n.z <= e2->n.z) :
	    (right == (e1->n.x * e2->n.y < e2->n.x * e1->n.y)));

    return(test);
}

/*===============================================================*/
/*=								=*/
/*=                     Inquire functions			=*/
/*=								=*/
/*===============================================================*/

/*************************************************************
  Function: InqVertVector(V)
  Input:
      V --- VertGeom
  Output:
      vector --- vector of point being represented by V
**************************************************************/
Vector InqVertVector(V)
  VertGeom*	V;
{
    return( ((Vert*) V)->d);
}


/*************************************************************
  Function: InqEdgeVector(E)
  Input:
      E --- EdgeGeom
  Output:
      vector --- vector of edge being represented by E
**************************************************************/
Vector InqEdgeVector(E)
  EdgeGeom*	E;
{
    return( ((Edge*) E)->n);
}



/*===============================================================*/
/*=								=*/
/*=             Data Modifying/Generating Functions		=*/
/*=								=*/
/*===============================================================*/


/*************************************************************
  Function: FreeVertGeom(V)
  Input:
      V --- VertGeom
  Output: <none>
**************************************************************/
void FreeVertGeom(V)
  VertGeom*	V;
{
    FREE((Vert*) V);
}



/*************************************************************
  Function: FreeEdgeGeom(E)
  Input:
      E --- EdgeGeom
  Output: <none>
**************************************************************/
void FreeEdgeGeom(E)
  EdgeGeom*	E;
{
    FREE((Edge*) E);
}



/*************************************************************
  Function: GetPoleGeom(bottom)
  Input:
      bottom ---  interested in top pole or bottom pole?
  Output:
      VertGeom --- new VertGeom representing the desired pole
**************************************************************/
VertGeom* GetPoleGeom(Boolean bottom)
{
    Vert*	v;
    v = ALLOCATE(Vert);

    v->d.x = v->d.z = 0.0;
    v->d.y = 1.0;
    v->longit = 0.0;	/* undefined for verts on horizon */
    v->epsilon = 0;

    if (bottom)
	NEG(v->d);

    return((VertGeom*) v);
}



/*************************************************************
  Function: GetHorizonGeom(left)
  Input:
      left---  interested in left horizon or right horizon?
  Output:
      EdgeGeom --- new EdgeGeom representing the desired horizon
**************************************************************/
EdgeGeom* GetHorizonGeom(Boolean left)
{
    Edge*	e;
    e = ALLOCATE(Edge);

    e->n.x = e->n.y = 0.0;
    e->n.z = 1.0;

    if (left)
	NEG(e->n);

    return((EdgeGeom*) e);
}



/*************************************************************
  Function: CopyVertGeom(V)
  Input:
      V --- VertGeom to copy
  Output:
      VertGeom --- new VertGeom structure
**************************************************************/
VertGeom*  CopyVertGeom(V)
  VertGeom*	V;
{
    Vert*	v;
    Vert*	new;

    v = (Vert*) V;

    new = ALLOCATE(Vert);

    (*new) = (*v);

    return(new);
}


/*************************************************************
  Function: CopyEdgeGeom(E)
  Input:
      E --- EdgeGeom to copy
  Output:
      EdgeGeom --- new EdgeGeom structure
**************************************************************/
EdgeGeom*  CopyEdgeGeom(E)
  EdgeGeom*	E;
{
    Edge*	e;
    Edge*	new;

    e = (Edge*) E;

    new = ALLOCATE(Edge);

    (*new) = (*e);

    return(new);
}


/*************************************************************
  Function: NegateVertGeom(V)
  Input:
      V --- VertGeom to Negate
  Output:
      <none>
  Effects:  negates the VertGeom

**************************************************************/
void NegateVertGeom(V)
  VertGeom*	V;
{
    Vert*	v;

    v = (Vert*) V;

    NEG(v->d);
}


/*************************************************************
  Function: NegateEdgeGeom(e)
  Input:
      E --- EdgeGeom to Negate
  Output:
      <none>
  Effects:  negates the EdgeGeom

**************************************************************/
void NegateEdgeGeom(E)
  EdgeGeom*	E;
{
    Edge*	e;

    e = (Edge*) E;

    NEG(e->n);
}


/*************************************************************
  Function: CreateVertFromVector(vec, orient)
  Input:
      vec ---  Vector to use as input data
      orient --- force VertGeom to front?
  Output:
      VertGeom -- newly created VertGeom

**************************************************************/
VertGeom*  CreateVertFromVector(Vector vec, Boolean orient)
{
    Vert*	v;
    v = ALLOCATE(Vert);

    v->d = NormalizeVector(vec);

    if (orient &&(!Front(v)))
	NEG(v->d);

    if (!ON_HORIZON_VEC(v->d))
	v->longit = v->d.x/v->d.z;

    v->epsilon = 0;

    return((VertGeom*) v);
}



/*************************************************************
  Function: CreateEdgeFromVector(vec, orient)
  Input:
      vec ---  Vector to use as input data
      orient --- force EdgeGeom to upward?
  Output:
      EdgeGeom -- newly created EdgeGeom

**************************************************************/
EdgeGeom*  CreateEdgeFromVector(Vector vec, Boolean orient)
{
    Edge*	e;
    e = ALLOCATE(Edge);

    e->n = NormalizeVector(vec);

    if (orient &&(!Upward(e)))
	NEG(e->n);

    return((EdgeGeom*) e);
}



/*************************************************************
  Function:  EdgeThroughVerts(U,V,E)
  Input:   
     U --- VertGeom
     V --- VertGeom
     E --- EdgeGeom
  Assumptions:  U & V not identical/anitpodal.  E is a "suggestion"
  Output:
    EdgeGeom --- the unique upward edge passing through both points.

  Description:
     This routine returns a new edge E satifsyfing,
        VertEdgeIncedent(U,E) == TRUE
        VertEdgeIncedent(V,E) == TRUE

**************************************************************/
EdgeGeom* EdgeThroughVerts(U,V,E)
  VertGeom*	U;
  VertGeom*	V;
  EdgeGeom*	E;
{
    Edge*	new;
    Vector	result;
    Vert*	u;
    Vert*	v;
    Boolean	lefthigher;

    /*--- NOTE:  for floating point, we simply ignore E.  We may as
      well try to find the most precise answer from scratch ---*/

    u = (Vert*) U;
    v = (Vert*) V;

    CROSS(result,u->d,v->d);
    result = NormalizeVector(result);

    new = ALLOCATE(Edge);
    new->n = result;

    if (!Upward(new))
	NEG(new->n);

    if (OnHorizonEdge(new)) {
	/*--- edge is on horizon (or zero?) ---*/
	if (EdgeOnLeftHorizon(new) !=
	    (!ON_POLE_VEC(u->d) ?
	     VertOnLeftHorizon(u) :
	     VertOnLeftHorizon(v))) {
	    /*---- gave us the wrong side of horizon ----*/
	    NEG(new->n);
	}
    } else {
	/*--- also beware of floating point problems in
	  which points are vertical or almost vertical
	  and edge should either be vertical, or almost
	  exactly reversed depending on which point is
	  considered leftendpoint and which is right ---*/



	if (new->n.y <= SMALL) {

	    /*--- which endpoint is the 'higher' one? ---*/
	    lefthigher = ((Thread_Test(u,v) == ANS_LEFT) ?
			  (u->d.y > v->d.y) :
			  (u->d.y < v->d.y));

	    if ((new->n.x > 0.0) && (!lefthigher)) {
		NEG(new->n);
		new->n.y = 0.0;
	    } else if ((new->n.x < 0.0) && (lefthigher)) {
		NEG(new->n);
		new->n.y = MINFLOAT;
	    }
	}
    }

    return((EdgeGeom*) new);
}


/*************************************************************
  Function:  EdgeEdgeIntersect(E1,E2,L1,R1,L2,R2)
  Input:   
      E1 --- first edge's EdgeGeom
      E2 --- second edge's EdgeGeom

  Optional Input:   (if NULL it will be ignored)
      L1 --- VertGeom for a left boudary for edge segment E1
      R1 --- VertGeom for a right boudary for edge segment E1
      L2 --- VertGeom for a left boudary for edge segment E2
      R2 --- VertGeom for a right boudary for edge segment E2
  Assumptions:  E1 & E2 not identical
  Output:
    VertGeom --- the unique front point that lies on both edges.

  Description:
     This routine returns a new VertGeom V satifsyfing,
        VertEdgeIncedent(V,E1) == TRUE
        VertEdgeIncedent(V,E2) == TRUE
	ThreadTest(V,L1) == ANS_RIGHT
	ThreadTest(V,L2) == ANS_RIGHT
	ThreadTest(R1,V) == ANS_RIGHT
	ThreadTest(R2,V) == ANS_RIGHT

**************************************************************/
VertGeom* EdgeEdgeIntersect(E1,E2,L1,R1,L2,R2)
  EdgeGeom*	E1;
  EdgeGeom*	E2;
  VertGeom*	L1;
  VertGeom*	R1;
  VertGeom*	L2;
  VertGeom*	R2;
{
    Vert*	new;
    Vector	result;
    Edge*	e1;
    Edge*	e2;
    Vert*	l1;
    Vert*	r1;
    Vert*	l2;
    Vert*	r2;

    Vert*	innerleft;
    Vert*	innerright;

    e1 = (Edge*) E1;
    e2 = (Edge*) E2;
    l1 = (Vert*) L1;
    r1 = (Vert*) R1;
    l2 = (Vert*) L2;
    r2 = (Vert*) R2;

    new = ALLOCATE(Vert);
    CROSS(new->d,e1->n,e2->n);
    new->d = NormalizeVector(new->d);
    new->epsilon = 0;
    if (!Front(new))
	NEG(new->d);

    if (ZERO_VEC(new->d)) {
	/*--- edges were coincedent ---*/
	if (l1 && l2) {
	    new->d = ((Thread_Test(l1,l2) == ANS_LEFT) ?
		      l2->d : l1->d);
	} else {
	    if (ON_HORIZON_NORM(e1->n)) {
		/*--- set to pole ---*/
		new->d.y = 1.0;
	    } else {
		new->d.x = e1->n.y;
		new->d.y = - e1->n.x;
		new->d = NormalizeVector(new->d);
	    }
	}
    } else {
	/*--- legitimate cross product ---*/

	if (!ON_HORIZON_VEC(new->d)) {
	    new->longit = (new->d.x / new->d.z);
	}

	/*--- now ensure that if intersection point is within
	  both segments, then it will end up testing correctly ---*/
	
	if (l1 && l2 && r1 && r2) {

	    innerleft = ((Thread_Test(l1,l2) == ANS_LEFT) ?
			 l2 : l1);
	    innerright = ((Thread_Test(r1,r2) == ANS_LEFT) ?
			  r1 : r2);


	    /*--- be careful of two edges which meet exactly at
	      an endpoint.  Make sure result is at endpoint ---*/

	    /*--- this is a big hack!  Not sure of better way ---*/
	    if (l1==l2) {
		new->d = l1->d;
		new->longit = l1->longit;
	    } else if (r1==r2) {
		new->d = r1->d;
		new->longit = r1->longit;
	    } else {

		/*--- also beware of floating point problems in
		  which intersection point does not fall in the
		  longitude range of the segment endpoints.
		  (this is a problem with 'vertical' edges) ---*/

		if (Closer(l1->d,e1->n,new->d,r1->d,TRUE) &&
		    Closer(l2->d,e2->n,new->d,r2->d,TRUE)) {
		    if (Thread_Test(new,innerleft) == ANS_LEFT) {
			if (new->longit == innerleft->longit) {
			    new->epsilon = innerleft->epsilon + 1;
			} else {
			    new->longit == innerleft->longit;
			}
		    }
		    if (Thread_Test(innerright,new) == ANS_LEFT) {
			if (new->longit == innerright->longit) {
			    new->epsilon = innerright->epsilon - 1;
			} else {
			    new->longit == innerright->longit;
			}
		    }
		}

/**
		if (((Thread_Test(new,innerleft) == ANS_LEFT) ||
		     (Thread_Test(innerright,new) == ANS_LEFT)) &&
		    Closer(l1->d,e1->n,new->d,r1->d,TRUE) &&
		    Closer(l2->d,e2->n,new->d,r2->d,TRUE)
		    ) {
**/
		    /*--- floating point problems ---*/
/**
		    new->longit = innerleft->longit;
		}
**/
	    }
	}
    }

    return((VertGeom*) new);
}



/*************************************************************
  Function:  SnapVertToEdge(V, E, L1, R1)
  Input:
    V --- VertGeom represnting a point
    E --- EdgeGeom represnting an edge

  Optional Input:   (if NULL it will be ignored)
    L1 --- VertGeom for a left boudary for edge segment E
    R1 --- VertGeom for a right boudary for edge segment E
  Assumptions:  V is not parallel to E
  Output: <none>
  Effects:
     V is projected so that by end of routine,
        VertEdgeIncedent(V,E) == TRUE

     And if the projection lies between L1 and R1 on E, then
	ThreadTest(V,L1) == ANS_RIGHT
	ThreadTest(R1,V) == ANS_RIGHT

  Description:
      Given a point, presumably "near" an edge, this function
      projects the point so that it lies "exactly" on the edge.
      (that is, as close as FP precision will allow)

**************************************************************/
void SnapVertToEdge(VertGeom* V, EdgeGeom* E,
	       VertGeom* L1, VertGeom* R1)
{
    Vector	result;
    Vector	subtract;
    double	scale;
    Vert*	v;
    Vert*	l1;
    Vert*	r1;
    Edge*	e;

    e = (Edge*) E;
    v = (Vert*) V;
    l1 = (Vert*) L1;
    r1 = (Vert*) R1;

    scale = DOT(v->d,e->n)/SIZE(e->n);
    subtract = e->n;
    SCALE(scale,subtract);

    SUB(result,v->d,subtract);
    
    result = NormalizeVector(result);

    v->d = result;

    if (!Front(v))
	NEG(v->d);

    if (!ON_HORIZON_VEC(v->d)) {
	v->longit = v->d.x / v->d.z;
    }

    if (l1 && r1) {
	if (((Thread_Test(v,L1) == ANS_LEFT) ||
	     (Thread_Test(v,R1) == ANS_RIGHT)) &&
	    Closer(l1->d,e->n,v->d,r1->d,TRUE)
	    ) {
	    /*--- floating point problems ---*/
	    v->longit = l1->longit;
	}
    }
}



/*************************************************************
  Function:  SnapEdgeToVert(E,V)
  Input:
      E --- EdgeGeom to adjust
      V --- VertGeom
  Assumptions:  V is not parallel to E
  Output: <none>
  Effects:
     E is projected so that by end of routine,
        VertEdgeIncedent(V,E) == TRUE

  Description:
      Given an edge, presumably "near" a point, this function
      projects the edge so that it goes "exactly" through the point.
      (that is, as close as FP precision will allow)

**************************************************************/
void SnapEdgeToVert(EdgeGeom* E, VertGeom* V)
{
    Vector	result;
    Vector	subtract;
    double	scale;
    Edge*	e;
    Vert*	v;

    e = (Edge*) E;
    v = (Vert*) V;

    scale = DOT(v->d,e->n)/SIZE(e->n);
    subtract = v->d;
    SCALE(scale,subtract);

    SUB(result,e->n,subtract);

    result = NormalizeVector(result);

    e->n = result;

    if (!Upward(e))
	NEG(e->n);
}






