/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * arrange_geom_int.c
 *
 * Integer implementation of geometric features
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/


/*===============================================================*/
/*=								=*/
/*=		Geometric Primitives on the Sphere		=*/
/*=							   	=*/
/*===============================================================*/

#include <math.h>
#include <values.h>
#include "arrange_geom.h"
#include "spheregeom.h"
#include "standards.h"

/*===============================================================*/
/*=								=*/
/*=                       EdgeGeom/VertGeom			=*/
/*=								=*/
/*=	This typedef is used as the VertGeom structure		=*/
/*===============================================================*/
typedef struct Vert {
    long	x;
    long	y;
    long	z;
} Vert;

typedef struct Edge {
    long	x;
    long	y;
    long	z;
} Edge;


#define MAX_GRID 1000

/*===============================================================*/
/*=								=*/
/*=             Integer Vector Macros				=*/
/*=								=*/
/*===============================================================*/
#define ON_HORIZON_IVEC(v) ((v)->z == 0)
#define ON_POLE_IVEC(v)    (((v)->z == 0) && ((v)->x == 0))

#define ON_HORIZON_INORM(e) (((e)->x == 0) && ((e)->y == 0))

#define ZERO_IVEC(v) (((v)->x==0) && ((v)->y==0) && ((v)->z==0))


/*===============================================================*/
/*=								=*/
/*=             Geometric Test Functions			=*/
/*=								=*/
/*===============================================================*/

/*************************************************************
  Function: OnHorizonVert(V)
  Input:
      V --- VertGeom for point in question
  Output:
      TRUE --- V is on the horizon
      FALSE --- V is not on the horizon
**************************************************************/
Boolean OnHorizonVert(V)
  VertGeom*	V;
{
    return(ON_HORIZON_IVEC((Vert*) V));
}



/*************************************************************
  Function: OnPoleVert(V)
  Input:
      V --- VertGeom for point in question
  Output:
      TRUE --- V is on a pole
      FALSE --- V is not on a pole
**************************************************************/
Boolean OnPoleVert(V)
  VertGeom*	V;
{
    return(ON_POLE_IVEC((Vert*) V));
}



/*************************************************************
  Function: OnHorizonEdge(E)
  Input:
      E --- EdgeGeom for edge in question
  Output:
      TRUE --- E is on the horizon
      FALSE --- E is not on the horizon
**************************************************************/
Boolean OnHorizonEdge(E)
  EdgeGeom*	E;
{
    return(ON_HORIZON_INORM((Edge*) E));
}



/*************************************************************
  Function: VertVertIncedent(U,V)
  Input:
      U --- VertGeom for point
      V --- VertGeom for point
  Output:
      TRUE --- U and V are considered the same
      FALSE --- U and V are not considered the same
**************************************************************/
Boolean VertVertIncedent(U,V)
  VertGeom*	U;
  VertGeom*	V;
{
    Vert*	u;
    Vert*	v;

    u = (Vert*) U;
    v = (Vert*) V;

    return(
	   /*--- same vector? ---*/
	   ((u->x*v->y == u->y*v->x) && (u->y*v->z == u->z*v->y)) &&
	   /*--- same orientation? ---*/
	   ((u->z*v->z > 0) || (u->y*v->y > 0) || (u->x*v->x > 0))
	   );
}



/*************************************************************
  Function: VertEdgeIncedent(V,E)
  Input:
      V --- VertGeom for point
      E --- EdgeGeom for edge
  Output:
      TRUE --- V is considered to lie on E
      FALSE --- V is not considered to lie on E
**************************************************************/
Boolean VertEdgeIncedent(V,E)
  VertGeom*	V;
  EdgeGeom*	E;
{
    Edge*	e;
    Vert*	v;
    Boolean	result;

    e = (Edge*) E;
    v = (Vert*) V;


    if ((e->x == 0) && (e->y == 0)) {
	result = ((v->z == 0) &&
		  ((v->x == 0) ||
		   ((e->z > 0) == (v->x > 0))));
    } else {
	result = ((e->x*v->x + e->y*v->y + e->z*v->z) == 0);
    }

    return(result);
}



/*************************************************************
  Function: EdgeEdgeIncedent(E,F)
  Input:
      E --- EdgeGeom for edge
      F --- EdgeGeom for edge
  Output:
      TRUE --- E and F are considered to be the same
      FALSE --- E and F are not considered to be the same
**************************************************************/
Boolean EdgeEdgeIncedent(E,F)
  EdgeGeom*	E;
  EdgeGeom*	F;
{
    Boolean result;
    Edge*	e;
    Edge*	f;

    e = (Edge*) E;
    f = (Edge*) F;

    return(
	   /*--- same vector? ---*/
	   ((e->x*f->y == e->y*f->x) && (e->y*f->z == e->z*f->y)) &&
	   /*--- same orientation? ---*/
	   ((e->z*f->z > 0) || (e->y*f->y > 0) || (e->x*f->x > 0))
	   );
}


/*************************************************************
  Function:  Front(v)
  Input:
      V ---  VertGeom in question
  Output:
      TRUE if V is in front hemisphere
      FALSE if V is not in front hemisphere
**************************************************************/
Boolean Front(V)
  VertGeom* V;
{
    return(((Vert*) V)->z >= 0);
}



/*************************************************************
  Function:  Upward(E)
  Input:
      E --- EdgeGeom in question
  Output:
      TRUE if E has an "upward" normal
      FALSE if E has a "downward" normal
      <for vertical E, conventionally left is "upward">
**************************************************************/
Boolean Upward(E)
  EdgeGeom* E;
{
    Edge* e;

    e = (Edge*) E;
    return(
	   ((e->y>0) ||
	    ((e->y==0) && (e->x <= 0)))
	   );
}


/*************************************************************
  Function:  VertOnLeftHorizon(V)
  Input:
      V --- VertGeom in question
  Assumptions:  V "on" horizon
  Output:
      TRUE if V is on left horizon (or bottom pole)
      FALSE if V is on right horizon (or top pole)
**************************************************************/
Boolean VertOnLeftHorizon(VertGeom* V)
{
    Vert* v;

    v = (Vert*) V;

    return(
	   ((v->x < 0) ||
	    ((v->x == 0) && (v->y < 0)))
	   );
}



/*************************************************************
  Function:  EdgeOnLeftHorizon(E)
  Input:
      E --- EdgeGeom in question
  Assumptions:  edge E part of horizon
  Output:
      TRUE if E is on left side of horizon
      FALSE if E is on right side of horizon
**************************************************************/
Boolean EdgeOnLeftHorizon(EdgeGeom* E)
{
    return((((Edge*) E)->z < 0));
}



/*************************************************************
  Function:  Thread_Test(P,Q,front,unit)
  Input:
      P --- VertGeom
      Q--- VertGeom
  Output:
      ANS_RIGHT == point P is on/right of thread through Q
      ANS_LEFT  == point P is left of thread through Q
      <note: as described in the introduction, we perturb
             lexicographically, so that P only lies on a
	     thread if it is exactly equal to Q>
**************************************************************/
int Thread_Test(VertGeom* P,VertGeom* Q)
{
    int		answer;
    double	store;
    Boolean	p_right;
    Boolean	q_right;
    Vert*	p;
    Vert*	q;

    p = (Vert*) P;
    q = (Vert*) Q;

    if (ON_HORIZON_IVEC(p)) {
	p_right = (p->x > 0) || ((p->x == 0) && (p->y > 0));
	if (ON_HORIZON_IVEC(q)) {
	    q_right = (q->x > 0) || ((q->x == 0) && (q->y > 0));
	    /*---- both P & Q on horizon ----*/
	    if (p_right != q_right) {
		/* P & Q on opposite horizons */
		answer = (p_right ?
			  /** P on right, Q on left **/
			  ANS_RIGHT :
			  /** P on left, Q on right **/
			  ANS_LEFT);
	    } else {
		/* P & Q on same horizon */
		answer = ((p_right == (p->y*q->x >= q->y*p->x)) ?
			  ANS_RIGHT : ANS_LEFT);
	    }
	} else {

	    /*---- just P on horizon ----*/
	    answer = (p_right ?
		      /** P on right horizon **/
		      ANS_RIGHT :
		      /** P on left horizon **/
		      ANS_LEFT);
	}
    } else if (ON_HORIZON_IVEC(q)) {
	q_right = (q->x > 0) || ((q->x == 0) && (q->y > 0));
	/*---- just Q on horizon ----*/
	answer = (q_right ?
		  /** Q on right horizon **/
		  ANS_LEFT :
		  /** Q on left horizon **/
		  ANS_RIGHT);
    } else {


	/*---- neither on horizon ----*/
	store = (p->x*q->z - q->x*p->z);
	if (store > 0) {
	    answer = ANS_RIGHT;
	} else if (store < 0) {
	    answer = ANS_LEFT;
	} else {	/* "vertical" -- must perturb */
	    answer = ((p->y*q->z >= q->y*p->z) ?
		      ANS_RIGHT : ANS_LEFT);
	}
    }

    return(answer);
}


/*************************************************************
  Function:  Edge_Test(P,E,front,unit)
  Input:
      P ---  VertGeom
      E ---  EdgeGeom
  Output:
      ANS_ABOVE ==  point P is on/above edge E
      ANS_BELOW ==  point P is below edge E
**************************************************************/
int  Edge_Test(VertGeom* P, EdgeGeom* E)
{
    int		answer;
    double	store;
    Vert*	p;
    Edge*	e;

    p = (Vert*) P;
    e = (Edge*) E;

    if ((store=(p->x*e->x + p->y*e->y + p->z*e->z)) > 0) {
	answer = ANS_ABOVE;
    } else if (store < 0) {
	answer = ANS_BELOW;
    } else {	/* on edge .. check for weird horizon cases */
	if (!ON_HORIZON_INORM(e)) {
	    answer = ANS_ABOVE;
	} else {
	    answer = (((p->x > 0) || ((p->x == 0) && (p->y > 0))) ?
		      ((e->z > 0) ?
		       /** both P & E refer to bot horizon **/
		       ANS_ABOVE :
		       /** P is bot, E is top **/
		       ANS_BELOW) :
		      /** P is top, E is either **/
		      ANS_ABOVE);
	}
    }

    return(answer);
}


/*************************************************************
  Function:  EdgeEdge_Above(T,E1,E2,right)
  Input:
      T --- VertGeom sitting at intersection of edges
      E1 --- EdgeGeom for first edge
      E2 --- EdgeGeom for second edge
      right --- interest in relationship to right or left of T?
  Output:
    TRUE  ==  E1 is "above" E2 (or EXACTLY equal)
    FALSE ==  E1 is "below" E2
    <where above is relative to the requested side>

**************************************************************/
Boolean EdgeEdge_Above(VertGeom* T, EdgeGeom* E1, EdgeGeom* E2,
		       Boolean right)
{
    Boolean	test;
    Vert*	t;
    Edge*	e1;
    Edge*	e2;
    

    t = (Vert*) T;
    e1 = (Edge*) E1;
    e2 = (Edge*) E2;

    if (ON_HORIZON_IVEC(t)) {
	if (ON_POLE_IVEC(t)) {
	    test = (e1->x*e2->z <= e1->z*e2->x);
	} else {
	    test = (e1->z*e2->y <= e1->y*e2->z);
	}
    } else {
	test = (right == (e1->x * e2->y < e2->x * e1->y));
    }

    return(test);
}

/*===============================================================*/
/*=								=*/
/*=                     Inquire functions			=*/
/*=								=*/
/*===============================================================*/

/*************************************************************
  Function: InqVertVector(V)
  Input:
      V --- VertGeom
  Output:
      vector --- vector of point being represented by V
**************************************************************/
Vector InqVertVector(V)
  VertGeom*	V;
{
    Vert*	v;
    Vector	new;

    v = (Vert*) V;

    new.x = v->x;
    new.y = v->y;
    new.z = v->z;

    new = NormalizeVector(new);
    
    return(new);
}


/*************************************************************
  Function: InqEdgeVector(E)
  Input:
      E --- EdgeGeom
  Output:
      vector --- vector of edge being represented by E
**************************************************************/
Vector InqEdgeVector(E)
  EdgeGeom*	E;
{
    Edge*	e;
    Vector	new;

    e = (Edge*) E;

    new.x = e->x;
    new.y = e->y;
    new.z = e->z;

    new = NormalizeVector(new);
    
    return(new);
}



/*===============================================================*/
/*=								=*/
/*=             Data Modifying/Generating Functions		=*/
/*=								=*/
/*===============================================================*/


/*************************************************************
  Function: FreeVertGeom(V)
  Input:
      V --- VertGeom
  Output: <none>
**************************************************************/
void FreeVertGeom(V)
  VertGeom*	V;
{
    FREE((Vert*) V);
}



/*************************************************************
  Function: FreeEdgeGeom(E)
  Input:
      E --- EdgeGeom
  Output: <none>
**************************************************************/
void FreeEdgeGeom(E)
  EdgeGeom*	E;
{
    FREE((Edge*) E);
}



/*************************************************************
  Function: GetPoleGeom(bottom)
  Input:
      bottom ---  interested in top pole or bottom pole?
  Output:
      VertGeom --- new VertGeom representing the desired pole
**************************************************************/
VertGeom* GetPoleGeom(Boolean bottom)
{
    Vert*	v;
    v = ALLOCATE(Vert);

    v->x = v->z = 0;
    v->y = (bottom ? -1 : 1);

    return((VertGeom*) v);
}



/*************************************************************
  Function: GetHorizonGeom(left)
  Input:
      left---  interested in left horizon or right horizon?
  Output:
      EdgeGeom --- new EdgeGeom representing the desired horizon
**************************************************************/
EdgeGeom* GetHorizonGeom(Boolean left)
{
    Edge*	e;
    e = ALLOCATE(Edge);

    e->x = e->y = 0;
    e->z = (left ? -1 : 1);

    return((EdgeGeom*) e);
}



/*************************************************************
  Function: CopyVertGeom(V)
  Input:
      V --- VertGeom to copy
  Output:
      VertGeom --- new VertGeom structure
**************************************************************/
VertGeom*  CopyVertGeom(V)
  VertGeom*	V;
{
    Vert*	v;
    Vert*	new;

    v = (Vert*) V;

    new = ALLOCATE(Vert);

    (*new) = (*v);

    return(new);
}


/*************************************************************
  Function: CopyEdgeGeom(E)
  Input:
      E --- EdgeGeom to copy
  Output:
      EdgeGeom --- new EdgeGeom structure
**************************************************************/
EdgeGeom*  CopyEdgeGeom(E)
  EdgeGeom*	E;
{
    Edge*	e;
    Edge*	new;

    e = (Edge*) E;

    new = ALLOCATE(Edge);

    (*new) = (*e);

    return(new);
}


/*************************************************************
  Function: NegateVertGeom(V)
  Input:
      V --- VertGeom to Negate
  Output:
      <none>
  Effects:  negates the VertGeom

**************************************************************/
void NegateVertGeom(V)
  VertGeom*	V;
{
    Vert*	v;

    v = (Vert*) V;

    v->x = -1*v->x;
    v->y = -1*v->y;
    v->z = -1*v->z;
}


/*************************************************************
  Function: NegateEdgeGeom(e)
  Input:
      E --- EdgeGeom to Negate
  Output:
      <none>
  Effects:  negates the EdgeGeom

**************************************************************/
void NegateEdgeGeom(E)
  EdgeGeom*	E;
{
    Edge*	e;

    e = (Edge*) E;

    e->x = -1*e->x;
    e->y = -1*e->y;
    e->z = -1*e->z;
}


/*************************************************************
  Function: CreateVertFromVector(vec, orient)
  Input:
      vec ---  Vector to use as input data
      orient --- force VertGeom to front?
  Output:
      VertGeom -- newly created VertGeom

**************************************************************/
VertGeom*  CreateVertFromVector(Vector vec, Boolean orient)
{
    Vert*	v;
    double	max;

    /*--- what is a good algorithm for fitting a lattice??? ---*/

    v = ALLOCATE(Vert);

    max = ((vec.x > 0) ? vec.x : -vec.x);

    if (vec.y > 0) max = vec.y;
    if (-vec.y > 0) max = -vec.y;
    if (vec.z > 0) max = vec.z;
    if (-vec.z > 0) max = -vec.z;

    v->x = (MAX_GRID * vec.x / max);
    v->y = (MAX_GRID * vec.y / max);
    v->z = (MAX_GRID * vec.z / max);

    /*--- just for efficiency, check for unit vectors ---*/
    if ((v->x==0) && (v->y==0) && (v->z!=0)) {
	v->z /= abs(v->z);
    } else if ((v->x==0) && (v->z==0) && (v->y!=0)) {
	v->y /= abs(v->y);
    } else if ((v->z==0) && (v->y==0) && (v->x!=0)) {
	v->x /= abs(v->x);
    }

    if (orient && (!Front(v)))
	NegateVertGeom(v);

    return((VertGeom*) v);
}



/*************************************************************
  Function: CreateEdgeFromVector(vec, orient)
  Input:
      vec ---  Vector to use as input data
      orient --- force EdgeGeom to upward?
  Output:
      EdgeGeom -- newly created EdgeGeom

**************************************************************/
EdgeGeom*  CreateEdgeFromVector(Vector vec, Boolean orient)
{
    Edge*	e;
    double	max;

    /*--- what is a good algorithm for fitting a lattice??? ---*/

    e = ALLOCATE(Edge);

    max = ((vec.x > 0) ? vec.x : -vec.x);

    if (vec.y > 0) max = vec.y;
    if (-vec.y > 0) max = -vec.y;
    if (vec.z > 0) max = vec.z;
    if (-vec.z > 0) max = -vec.z;

    e->x = (MAX_GRID * vec.x / max);
    e->y = (MAX_GRID * vec.y / max);
    e->z = (MAX_GRID * vec.z / max);


    if (orient && (!Upward(e)))
	NegateEdgeGeom(e);

    return((EdgeGeom*) e);
}



/*************************************************************
  Function:  EdgeThroughVerts(U,V,E)
  Input:   
     U --- VertGeom
     V --- VertGeom
     E --- EdgeGeom
  Assumptions:  U & V not identical/anitpodal.  E is a "suggestion"
  Output:
    EdgeGeom --- the unique upward edge passing through both points.

  Description:
     This routine returns a new edge E satifsyfing,
        VertEdgeIncedent(U,E) == TRUE
        VertEdgeIncedent(V,E) == TRUE

**************************************************************/
EdgeGeom* EdgeThroughVerts(U,V,E)
  VertGeom*	U;
  VertGeom*	V;
  EdgeGeom*	E;
{
    Edge*	new;
    Vert*	u;
    Vert*	v;
    EdgeGeom*	New;

    /*--- NOTE:  for int, we will first see if the suggestion is
      correct.  That will avoid calculating a new one which may
      have blown up the size of the integers. ---*/

    u = (Vert*) U;
    v = (Vert*) V;


    if (E && VertEdgeIncedent(V,E) && VertEdgeIncedent(U,E)) {
	New = CopyEdgeGeom(E);
    } else {
	new = ALLOCATE(Edge);
	new->x = (u->y*v->z - u->z*v->y);
	new->y = (u->z*v->x - u->x*v->z);
	new->z = (u->x*v->y - u->y*v->x);
	
	if (!Upward(new))
	    NegateEdgeGeom(new);
	
	if (OnHorizonEdge(new)) {
	    /*--- edge is on horizon (or zero?) ---*/
	    if (EdgeOnLeftHorizon(new) !=
		((!OnPoleVert(u)) ?
		 VertOnLeftHorizon(u) :
		 VertOnLeftHorizon(v))) {
		/*---- gave us the wrong side of horizon ----*/
		NegateEdgeGeom(new);
	    }
	}
	
	New = (EdgeGeom*) new;
    }

    return(New);
}


/*************************************************************
  Function:  EdgeEdgeIntersect(E1,E2,L1,R1,L2,R2)
  Input:   
      E1 --- first edge's EdgeGeom
      E2 --- second edge's EdgeGeom

  Optional Input:   (if NULL it will be ignored)
      L1 --- VertGeom for a left boudary for edge segment E1
      R1 --- VertGeom for a right boudary for edge segment E1
      L2 --- VertGeom for a left boudary for edge segment E2
      R2 --- VertGeom for a right boudary for edge segment E2
  Assumptions:  E1 & E2 not identical
  Output:
    VertGeom --- the unique front point that lies on both edges.

  Description:
     This routine returns a new VertGeom V satifsyfing,
        VertEdgeIncedent(V,E1) == TRUE
        VertEdgeIncedent(V,E2) == TRUE
	ThreadTest(V,L1) == ANS_RIGHT
	ThreadTest(V,L2) == ANS_RIGHT
	ThreadTest(R1,V) == ANS_RIGHT
	ThreadTest(R2,V) == ANS_RIGHT

**************************************************************/
VertGeom* EdgeEdgeIntersect(E1,E2,L1,R1,L2,R2)
  EdgeGeom*	E1;
  EdgeGeom*	E2;
  VertGeom*	L1;
  VertGeom*	R1;
  VertGeom*	L2;
  VertGeom*	R2;
{
    /*--- note:  L1,R1,L2,R2 are completely irrelevant, since the
      calculations here are exact. ---*/

    Vert*	new;
    Edge*	e1;
    Edge*	e2;

    e1 = (Edge*) E1;
    e2 = (Edge*) E2;

    new = ALLOCATE(Vert);
    new->x = (e1->y*e2->z - e1->z*e2->y);
    new->y = (e1->z*e2->x - e1->x*e2->z);
    new->z = (e1->x*e2->y - e1->y*e2->x);

    if (!Front(new))
	NegateVertGeom(new);

    if (ZERO_IVEC(new)) {
	/*--- edges were coincedent ---*/
	if (ON_HORIZON_INORM(e1)) {
	    /*--- set to pole ---*/
	    new->y = 1;
	} else {
	    new->x = e1->y;
	    new->y = - e1->x;
	}
    }

    return((VertGeom*) new);
}



/*************************************************************
  Function:  SnapVertToEdge(V, E, L1, R1)
  Input:
    V --- VertGeom represnting a point
    E --- EdgeGeom represnting an edge

  Optional Input:   (if NULL it will be ignored)
    L1 --- VertGeom for a left boudary for edge segment E
    R1 --- VertGeom for a right boudary for edge segment E
  Assumptions:  V is not parallel to E
  Output: <none>
  Effects:
     V is projected so that by end of routine,
        VertEdgeIncedent(V,E) == TRUE

     And if the projection lies between L1 and R1 on E, then
	ThreadTest(V,L1) == ANS_RIGHT
	ThreadTest(R1,V) == ANS_RIGHT

  Description:
      Given a point, presumably "near" an edge, this function
      projects the point so that it lies "exactly" on the edge.
      (that is, as close as FP precision will allow)

**************************************************************/
void SnapVertToEdge(VertGeom* V, EdgeGeom* E,
	       VertGeom* L1, VertGeom* R1)
{
    /*--- this function was not necessary with exact precision ---*/
}


/*************************************************************
  Function:  SnapEdgeToVert(E,V)
  Input:
      E --- EdgeGeom to adjust
      V --- VertGeom
  Assumptions:  V is not parallel to E
  Output: <none>
  Effects:
     E is projected so that by end of routine,
        VertEdgeIncedent(V,E) == TRUE

  Description:
      Given an edge, presumably "near" a point, this function
      projects the edge so that it goes "exactly" through the point.
      (that is, as close as FP precision will allow)

**************************************************************/
void SnapEdgeToVert(EdgeGeom* E, VertGeom* V)
{
    /*--- this function was not necessary with exact precision ---*/
}






