/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * arrange_macros.h
 *
 * Arrangement macros for accessing structure fields
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/

#ifndef _ARRANGE_MACROS
#define _ARRANGE_MACROS

#include "arrange_type.h"

/*************************************************************/
/*  All of the arrangement types are written as objects, and
    applications may want to query many different pieces of
    information from the objects.   The problem is that some of
    these queries are actually stored as true fields in the
    structures, and others are simply indirect combination of
    other fields.

    Although, application can be written which access the fields
    of the structured DIRECTLY, these applications will be very
    susceptible to upward compatability problems with future
    versions and bug fixes.  That is, if the fields of the structures
    are changed for any reason, the code written based on those
    fields will no longer compile without modification.

    The true solution to this is provided by languages such as
    C++ where structures can have methods that are provided by
    the structure and can be accessed in a postfix manner.
    However C does not provide any clean solution to this.
    The queries can be made to be functions, but those cannot
    be compiled as in line code.

    Chosen here, we define macros for all the query methods.  This
    serves the goal of efficiency, and upward compatability.  The
    serious downside is a loss of readability in the code due to
    the prefix nature of nested macros.

    That is, simply queries such as (the_vert->id) becoming V_ID(the_vert)
    are easy to read, but when two or three levels are combined,
    the_vert->brTrap->te->lvert   becomes
    E_lvert(T_te(V_brTrap(the_vert)))  which will be unnatural for
    most C programmers.   In C++, methods can be given in postfix
    order, so that the expression becomes the_vert.brTrap.te.lvert, even
    though the methods are not necessarily direct field references.

    If the rest of the project is using C++, I would strongly suggest
    writing a simple C++ file on top of this, making the type classes
    into objects, and making true methods based on these macros.
    Then this code could be used from C++ in it's natural fashion.
    If someone does this, I would be very happy if they would send
    me the C++ overlay so that I could add that file as an additional
    tool for others to use.

/*************************************************************/




#define V_id(v)		(v)->id
#define V_mirror(v)	(v)->mirror
#define V_direction(v)	(v)->direction
#define V_polys(v)	(v)->polys
#define V_tlTrap(v)	(v)->tlTrap
#define V_trTrap(v)	(v)->trTrap
#define V_blTrap(v)	(v)->blTrap
#define V_brTrap(v)	(v)->brTrap
#define V_topTV(v)      (v)->topTV
#define V_botTV(v)      (v)->botTV

#ifdef ARRANGE_DEBUG
#define V_link(v)	(v)->link
#endif



#define TV_id(tv)	(tv)->id
#define TV_from_above(tv) (tv == tv->source->botTV)
#define TV_edge(tv)	(TV_from_above(tv) ? tv->source->brTrap->be : \
                                             tv->source->trTrap->te)
#define TV_source(tv)	(tv)->source
#define TV_lTrap(tv)	(TV_from_above(tv) ? tv->source->blTrap : \
                                             tv->source->tlTrap)
#define TV_rTrap(tv)	(TV_from_above(tv) ? tv->source->brTrap : \
                                             tv->source->trTrap)
#define TV_lNeigh(tv)	(tv)->lNeigh
#define TV_rNeigh(tv)	(tv)->rNeigh

#ifdef ARRANGE_DEBUG
#define TV_link(tv)	(tv)->link
#endif



#define E_id(e)		(e)->id
#define E_mirror(e)	(e)->mirror
#define E_normal(e)	(e)->normal
#define E_lvert(e)	((e)->tlTrap ? (e)->tlTrap->lvert : (e)->blTrap->lvert)
#define E_rvert(e)	((e)->trTrap ? (e)->trTrap->rvert : (e)->brTrap->rvert)
#define E_polys_on(e)	(e)->polys_on
#define E_polys_above(e) (e)->polys_above
#define E_polys_below(e) (e)->polys_below
#define E_lThrVert(e)	(e)->lThrVert
#define E_rThrVert(e)	(e)->rThrVert
#define E_tlTrap(e)	(e)->tlTrap
#define E_trTrap(e)	(e)->trTrap
#define E_blTrap(e)	(e)->blTrap
#define E_brTrap(e)	(e)->brTrap

#ifdef ARRANGE_DEBUG
#define E_link(e)	(e)->link
#endif



#define T_id(t)		(t)->id
#define T_lvert(t)	(t)->lvert
#define T_rvert(t)	(t)->rvert
#define T_lvtype(t)	(t)->lvtype
#define T_rvtype(t)	(t)->rvtype
#define T_tlv(t)	(V_ON_TOP((t)->lvtype) ? NULL : (t)->lvert->topTV )
#define T_trv(t)	(V_ON_TOP((t)->rvtype) ? NULL : (t)->rvert->topTV )
#define T_blv(t)	(V_ON_BOT((t)->lvtype) ? NULL : (t)->lvert->botTV )
#define T_brv(t)	(V_ON_BOT((t)->rvtype) ? NULL : (t)->rvert->botTV )
#define T_tln(t)	((t)->lvtype==com ? NULL : ((t)->lvtype==top ? \
		            (t)->lvert->blTrap : (t)->lvert->tlTrap))
#define T_trn(t)	((t)->rvtype==com ? NULL : ((t)->rvtype==top ? \
		            (t)->rvert->brTrap : (t)->rvert->trTrap))
#define T_bln(t)	((t)->lvtype==com ? NULL : ((t)->lvtype==bot ? \
		            (t)->lvert->tlTrap : (t)->lvert->blTrap))
#define T_brn(t)	((t)->rvtype==com ? NULL : ((t)->rvtype==bot ? \
		            (t)->rvert->trTrap : (t)->rvert->brTrap))
#define T_te(t)		(t)->te
#define T_be(t)		(t)->be
#define T_hist(t)	(t)->hist

#ifdef ARRANGE_DEBUG
#define T_link(t)	(t)->link
#endif


#endif
