/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * arrange_type.h
 *
 * Arrangements and supporting data structures
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/

#ifndef _ARRANGE_TYPE
#define _ARRANGE_TYPE

#include "standards.h"



/*   enumerated type for nodes of history dag */
enum HistType {threadtest = 1, edgetest, leaf };

/*   enumerated type for point location query */
enum CellType {vert = 1, edge, trap };

/*   enumerated type for vertex classification */
enum TrapVertType {top=1, bot, com, mid };

/*   Macros to check if vertex is on top and/or bot edge of trap */
#define V_ON_TOP(vtype)  (vtype & 1)
#define V_ON_BOT(vtype)  (vtype & 2)


/*************************************************************/
/*  macros to check edge/vertex for horizon special case     */

#define ON_HORIZON_VERTEX(v)  (((v)->tlTrap==NULL) || ((v)->brTrap==NULL))
#define ON_HORIZON_EDGE(e)  (((e)->tlTrap==NULL) || ((e)->brTrap==NULL))

/*************************************************************/



/*************************************************************/
/*  A TrapVert represents a "true" vertex from the input data.
    These is either a vertex of a polygon, or the intersection of
    two edges from polygons.

    TrapVert methods
      mirror -- if the vertex is on horizon, this is the antipode
      direction -- the direction of the true point on the sphere
                    (Assumption:  direction is on "front" hemisphere)
      polys  --  polygons which are entered *only* when exactly on
                   the vertex.  (these are in addition to many other
		   polygons which are inferred from the incedent edges,
		   as well as enclosing faces)
      tlTrap -- the trap immediately left of the *upward* thread
      trTrap -- the trap immediately right of the *upward* thread
      blTrap -- the trap immediately left of the *downward* thread
      brTrap -- the trap immediately right of the *downward* thread
      topTV  --- the threadvert at end of *upward* thread (if exists)
      botTV  --- the threadvert at end of *downward* thread (if exists)
*/
typedef struct TrapVert {
    int			id;
    struct TrapVert	*mirror;
    struct VertGeom*	direction;	/* Assumption:  Front(direction) */
    struct PolyTree	*polys;
    struct Trap		*tlTrap;
    struct Trap		*trTrap;
    struct Trap		*blTrap;
    struct Trap		*brTrap;
    struct TrapThreadVert *topTV;
    struct TrapThreadVert *botTV;

    
#ifdef ARRANGE_DEBUG
    struct TrapVert	*link;		/* keep global linked list */
#endif
} TrapVert;



/*************************************************************/
/*  A TrapThreadVert represents an "artificial" vertex that was
    created by the intersection of a thread and an original edge.
    Notice that by the degeneracy assumptions, two threads cannot
    meet from opposite sides at a TrapThreadVert.

    TrapThreadVert methods
      edge -- what (unique) TrapEdge does this lie on
      source -- vertex which defines the source of thread
      from_above -- is the thread coming from above or below
      lTrap -- the trap on the left of the thread coming in
      rTrap -- the trap on the right of the thread coming in
      lNeigh -- the next ThreadVert to the left on the edge
                   (NULL if left neighbor is the "true" vert)
      rNeigh -- the next ThreadVert to the right on the edge
                   (NULL if right neighbor is the "true" vert)
*/
typedef struct TrapThreadVert {
    int			id;
    struct TrapVert	*source;
    struct TrapThreadVert *lNeigh;
    struct TrapThreadVert *rNeigh;

#ifdef ARRANGE_DEBUG
    struct TrapThreadVert	*link;	/* keep global linked list */
#endif
} TrapThreadVert;



/*************************************************************/
/*  A TrapEdge represents the section of an original polygon edge
    between two TrapVert's.  That is, if an original edge is not
    intersected by other edges, then it will become a single
    TrapEdge, but if another edge intersects it, then the point of
    intersection becomes a TrapVert, and the edge is broken into
    two TrapEdge's.

    TrapEdge fields
      mirror -- if the edge is on horizon, this is the antiedge
      normal -- Vector specifying normal to great circle of edge
                 (Assumption: normal points conventionally "upward")
      lvert --  Left endpoint TrapVert
      rvert --  Right endpoint TrapVert
      polys_on -- polygons which are entered *only* while on the edge
                   but not on either side.
      polys_below -- polygons which are entered while on or locally
                      below the edge, but not above
      polys_above -- polygons which are entered while on or locally
                      above the edge, but not below
      lThrVert -- the leftmost TrapThreadVert on this edge
                   (NULL if no threads hit this edge)
      rThrVert -- the rightmost TrapThreadVert on this edge
                   (NULL if no threads hit this edge)
      tlTrap -- trap above this edge incident to the left vertex
      trTrap -- trap above this edge incident to the right vertex
      blTrap -- trap below this edge incident to the left vertex
      brTrap -- trap below this edge incident to the right vertex
*/
typedef struct TrapEdge {
    int			id;
    struct TrapEdge	*mirror;
    struct EdgeGeom*	normal;		/* Assumption:  Upward(normal) */
    struct PolyTree	*polys_on;
    struct PolyTree	*polys_below;
    struct PolyTree	*polys_above;
    struct TrapThreadVert *lThrVert;	/* Left most ThreadVert */
    struct TrapThreadVert *rThrVert;	/* Right most ThreadVert */
    struct Trap		*tlTrap;
    struct Trap		*trTrap;
    struct Trap		*blTrap;
    struct Trap		*brTrap;

#ifdef ARRANGE_DEBUG
    struct TrapEdge	*link;		/* keep global linked list */
#endif
} TrapEdge;



/*************************************************************/
/*  A Trap represents a trapezoid in the vertical decomposition.
    Each trap will *always* have an edge above it and below it,
    and it will *always* have a TrapVert bounding the left extent
    and the right extent.   It also may or may not have several
    TrapThreadVert's depending on the case.

    Trap fields
      lvert -- the TrapVert defining the left side
      rvert -- the TrapVert defining the right side
      lvtype -- left vert lies on (top edge, neither edge,
                                   bottom edge, or both)
      rvtype -- right vert list on (top | middle | bottom | both)
      tlv -- top-left TrapThreadVert  (may be NULL)
      trv -- top-right TrapThreadVert  (may be NULL)
      blv -- bottom-left TrapThreadVert  (may be NULL)
      brv -- bottom-right TrapThreadVert  (may be NULL)
      tln -- neighbor trap to left (and above lvert if lvtype=mid)
      bln -- neighbor trap to left (and below lvert if lvtype=mid)
      trn -- neighbor trap to right (and above rvert if rvtype=mid)
      brn -- neighbor trap to right (and below rvert if rvtype=mid)
      te --  edge bounding top of trap
      be --  edge bounding bottom of trap
      hist -- back-pointer to leaf in history DAG leading to trap
*/
typedef struct Trap {
    int			id;
    struct TrapVert	*lvert;
    struct TrapVert	*rvert;
    enum TrapVertType	lvtype;
    enum TrapVertType	rvtype;
    struct TrapEdge	*te;		/* top edge */
    struct TrapEdge	*be;
    struct TrapHist	*hist;		/* back-pointer into history */

#ifdef ARRANGE_DEBUG
    struct Trap		*link;		/* keep global linked list */
#endif
} Trap;

/*************************************************************/


typedef struct ArrangeCell {
    enum CellType 	t;
    union {
	struct Trap	*trap;
	struct TrapEdge	*edge;
	struct TrapVert	*vert;
    } d;
    struct Trap		*ref_trap;	/* nearby trap */
} ArrangeCell;

typedef struct TrapHist {
    int			id;
    enum HistType	nodetype;	/* what comparison */
    union {
	struct VertGeom*	v;
	struct EdgeGeom*	e;
    } data;				/* if a test...*/
    Trap		*trap;		/* if a leaf... */
    struct TrapHist	*ans[2];	/* Children nodes */

    /*  The order of children is based on the constants
	ANS_RIGHT, ANS_LEFT  or ANS_ABOVE, ANS_BELOW as
	defined in spheregeom.h */


#ifdef ARRANGE_DEBUG
    struct TrapHist	*link;		/* keep global linked list */
#endif
} TrapHist;

/*************************************************************/
/*
  special case: the horizon

  Properly handling the horizon edges takes quite a bit of special
  care.  The current implementation creates two mirror copies of all
  edges and vertices that are part of the horizon.  One behaves as
  the part on the left horizon and one behaves as the part on the
  right horizon.

  Notice that the horizon may be broken into many separate edges as
  time goes by.  Although it is initialized as a single TrapEdge
  structure for each side, if any later vertex lies on the horizon,
  then it breaks both sides into two or more edges.

  However, the way arrange.c is currently written, it so happens that
  the piece of the broken great circle, immediately left of the North
  pole, (or immediately right of the south pole) has some special
  properties. First of all, ALL thread verts which hit the horizon
  will actually hit this specific edge.  All other edges on the horizon
  will have NO thread verts.

*/


typedef struct Arrange {
  List			polys;		/* List of (Polygon *)  */
  struct TrapHist	*hist_root;
  struct TrapVert	*tpole;
  struct TrapVert	*bpole;
  struct Arrange	*mirror;
  
/**  List of VertGeom/EdgeGeom structs allocated **/
  List			vgeom;
  List			egeom;

/**  Dynamic memory stacks for extra efficiency  **/
  struct Dheap		*Dheap_verts;
  struct Dheap		*Dheap_tverts;
  struct Dheap		*Dheap_edges;
  struct Dheap		*Dheap_traps;
  struct Dheap		*Dheap_hist;
  struct Dheap		*Dheap_ptree;

  int	num_verts;
  int	num_tverts;
  int	num_edges;
  int	num_traps;
  int	num_hists;

#ifdef ARRANGE_DEBUG
  struct TrapVert	*vert_link;
  struct TrapThreadVert *tvert_link;
  struct TrapEdge	*edge_link;
  struct Trap		*trap_link;
  struct TrapHist	*hist_link;
#endif
} Arrange;



/*************************************************************/
/*  Some 'useful' functions for the serious Arrange clients  */
/*************************************************************/

/*************************************************************
  Function:  Arrange_vert_nexttrap(the_A, the_vert, the_trap, ccw,
  result_trap, result_edge)
  Input:
      the_A    --- an arrangement
      the_vert --- TrapVert to walk around
      the_trap --- current Trap in walk
      ccw      --- walk counterclockwise?  (or else clockwise)
  Output:
      result_trap --- next trap walking (counter)clockwise
      result_edge --- next edge walking (counter)clockwise
  Effects:    <none>
  
  Description:  This routine simulates a (counter)clockwise circular
  list around a TrapVert, even though no such list exists explicitly.
  
  Given a trap, this will return the next trap seen when walking
  around the vert in a (counter)clockwise order.  It will also
  return the edge that is crossed to get there if it exists.
  (Otherwise, NULL will be returned for that parameter)
  
  If the walk hits the horizon, then result_edge will be set to
  the edge that was hit, and result_trap will be set to NULL.
  **************************************************************/
void Arrange_vert_nexttrap(Arrange*, TrapVert*, Trap*,
		      Boolean, Trap**, TrapEdge**);



/*************************************************************
  Function:  Arrange_ThreadVert_Direction(the_A,tvert,answer)
  Input:	
      the_A --- an arrangement
      tvert --- a TrapThreadVert
  Output:
      answer --- the vector direction of tvert
  Effects:    <none>
  
  Description:  This routine calculates the normalized
  directional vector for a threadvert, and places the
  result in 'answer' upon return.
  
  **************************************************************/
void Arrange_ThreadVert_Direction(Arrange*, TrapThreadVert*, Vector*);


/*************************************************************
  Function:  Arrange_ThreadVert_Direction_And_Edge(the_A,tvert,
  fake_edge, endpoint)
  Input:
      the_A --- an arrangement
      tvert --- a TrapThreadVert
  Output:
      fake_edge --- the upward edge simulating a thread
      endpoint  --- the vector direction of tvert
  Effects:    <none>
  
  Description:  This routine calculates the upward normal to the
  "fake" edge that would like immediately overlapping the thread
  through tvert, and it then calculates the normalized direction
  of tvert by intersecting this edge with the true edge which
  the thread hits.
  
  **************************************************************/
void Arrange_ThreadVert_Direction_And_Edge(Arrange*,
					   TrapThreadVert*,
					   Vector*, Vector*);

/*************************************************************
  Function: Arrange_PointLocation(the_A,p)
  Input: arrangement & VertGeom p
  Output: ArrangeCell with respect to query P
  Effects: <none>
  
  Description:  This uses the History to locate a trap
  containing the query point, and then afterwards checks
  the boundary verts/edges explicitly.
  
  Uses a precision error in checking boundaries.
  
  **************************************************************/
ArrangeCell	Arrange_PointLocation(Arrange*,struct VertGeom*);


/*************************************************************
  Function: Arrange_AddVertex(the_A,direction,polys_on,location)
  Input:
      the_A -- a valid arrangement
      direction -- VertGeom representing new vertex
      polys_on -- polygons entered *only* when on new vertex
      location -- result of previous point location query.
  Output:
      new TrapVert based on direction
  Effects: updates entire Arrange structure to reflect change
  
  Description:
  
  Arrange_AddVertex --  This is the high level routine to add a
  (possibly) new vertex to the arrangement.  It does everything
  necessary depending on what type of cell (trap/edge/vert) the
  new vert lies on top of.

  This routine will also handle any details necessary because of
  any mirrors.
  
  **************************************************************/
TrapVert* Arrange_AddVertex(Arrange*, struct VertGeom*, List, ArrangeCell);


/*************************************************************
  Function:  Arrange_AddEdge(the_A, start_vertex, goal_vertex,
  goal_vector, goal_normal, polys_on,
  polys_below, polys_above)
  Input:
      the_A -- a valid arrangement
      start_vertex -- TrapVert to start new edge (must be given)
      goal_vertex -- OPTIONAL parameter giving TrapVert at end
      goal_vector -- if NO goal_vertex, then this gives the
                       VertGeom for where new edge should end.
      goal_normal -- EdgeGeom describing new edge path
      polys_on -- Polygons entered only when *on* edge
      polys_below -- Polygons entered when on or below edge
      polys_above -- Polygons entered when on or above edge
  Output:
      TrapVert at end of new edge
  Effects: updates entire Arrange structure to account for the
       insertion of the new edge.
  
  Description:
  
  Arrange_AddEdge --  This is the high level routine to add a
  (possibly) new edge to the arrangement.  It does everything
  necessary.
  
  REQUIREMENT:  this routine assumes that the edge lies entirely in
  the (Z >= 0) hemisphere.  Also, it assumes that an edge lying on
  the Z=0 plane does NOT cross either of the poles. If this is not
  the case, it is up to the caller to break the edge up into pieces
  that satisfy this.
  
  Also, it assumes that the normal given truly passes through both
  endpoints.  Usually this normal should be uniquely defined by the
  points, but in the case of a full half-circle, it is ambiguous and
  hence the normal must be part of the input.
  
  For efficiency, it takes a pointer to the TrapVert which coincides
  with the start_endpoint. Also, to be consistent, the function returns
  a pointer to the TrapVert which coincides with the final_endpoint.
  
  If the goal_vector lies on the horizon, then this routine will
  guarantee that the final_endpoint returned is a vertex lying on
  the horizon.
  
  Also, if an edge is so small that the end is merged with the start
  vertex, this routine will NOT add any PolySource to the arrangement.
  If the edge is part of a chain of edges, then there is no need
  for any adjustment.   However, if for some reason, the edge is part
  of a polygon/polychain so small that it is treated as a singularity,
  then it is necessary to add the PolySource to the vertex->polys
  itself.  It is the responsibility of the CALLER of this routine to
  handle that degeneracy.
  
  **************************************************************/
TrapVert* Arrange_AddEdge(Arrange*, TrapVert*, TrapVert*, struct VertGeom*,
			  struct EdgeGeom*, List, List, List);

#endif

