/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * debug_graphics.c
 *
 * Graphical display of arrangement
 *
 * Author:  Bruce Romney
 * Modified:  Jan 1, 1995 --- Original release (Romney)
 * Modified:  Jan 15, 1995 --- Added code to create B/W printouts(mhg)
 *******************************************************/

#include <math.h>
#include "arrange.h"
#include "debug_graphics.h"
#include "lists.h"
#include "standards.h"

int debugger_default_size = 500;
#define DB_RADIUS (2*debugger_default_size/5)
#define DB_HEIGHT debugger_default_size
#define DB_WIDTH debugger_default_size
/*  unit of DB_STEPSIZE is degrees per step */
#define DB_STEPSIZE (360.0/((float)debugger_default_size))


#define PRECISION                0.001
#define PI 3.1415926535898
#define TWO_PI (PI*2)

extern Display *dpy;	
extern XFontStruct *the_font_struct;

GC     db_red, db_blue, db_grey, db_green, db_white, db_black, db_yellow;
Window db_window;
int    db_x0, db_y0;		/* center of the hemisphere */

/*--------------------------------------*/
/*-  Utils functions			*/
/*--------------------------------------*/

void specify_vector(v, x, y, z)
     Vector *v;
     double x, y, z;
{
   v->x = x;
   v->y = y;
   v->z = z;
}

void copy_vector(Vector *new, Vector *old)
{
  new->x = old->x;
  new->y = old->y;
  new->z = old->z;
}

void scale_vector(v, factor)
     Vector *v;
     double factor;
{
   v->x *= factor;
   v->y *= factor;
   v->z *= factor;
}

double norm_of_vec(v)
     Vector *v;
{
return (  sqrt(v->x*v->x + v->y*v->y + v->z*v->z ));
}

Boolean equal(n1, n2)
     double n1, n2;
{
  return(n1 < n2 + PRECISION && n2 < n1 + PRECISION);
}

double sq_norm_of_vec(v)
     Vector *v;
{
return (v->x*v->x + v->y*v->y + v->z*v->z );
}

void normalize_vector(v)
     Vector *v;
{
   double mag;
   mag = norm_of_vec(v);
   if (mag<PRECISION) {
       printf("Cannot normalize a zero vector.\n");
       exit(-1);
   }
   if (!equal(mag,1.0))
     scale_vector(v,1.0/mag);
}

/*----------------------------------------------*/
Vector *cross_vec_product(Vector *product, Vector *v1, Vector *v2)
{
   product->x = v1->y*v2->z - v1->z*v2->y ;
   product->y = v1->z*v2->x - v1->x*v2->z ;
   product->z = v1->x*v2->y - v1->y*v2->x ;
   return product;
}

/*---------------------------------------------*/
Vector *normalized_cross_product(Vector *product, Vector *v1, Vector *v2)
{
   double mag;
   cross_vec_product(product,v1,v2);
   normalize_vector(product);
   return product;
}

double dot_product(v1, v2)
     Vector *v1, *v2;
{
   return (  v2->x*v1->x + v2->y*v1->y + v2->z*v1->z );
}


/*--------------------------------------*/
/*-  Window functions			*/
/*--------------------------------------*/
#define DEFAULT_WINDOW_WIDTH      500
#define DEFAULT_WINDOW_HEIGHT     500
#define PALLET_SIZE 50		/* maximum number of distinct colors */
#define INTERESTING_EVENTS ExposureMask |StructureNotifyMask

extern GC BASIC_COLORS;

XColor backgrd_xc2;
char FG_COLOR[51];      	/*the foreground color string*/
char BG_COLOR[51];      	/*the background color*/
char FONTNAME[51];
char *DISPLAY_NAME;     	/*the display name string*/
XColor pallet_xc[PALLET_SIZE], backgrd_xc;
GC pallet_gc[PALLET_SIZE];

XColor pallet_xc[PALLET_SIZE], backgrd_xc;
GC pallet_gc[PALLET_SIZE];
char pallet_name[PALLET_SIZE][25];
int num_colors=0;


/*---------------------------------------------------------------------
 * initialize_display_options  -  sets default display options
 */
void initialize_display_options() 
{
   FG_COLOR[0] = '\0';			/* no constant FG color */
   strncpy(BG_COLOR, "DimGrey", 50);
   strncpy(FONTNAME, "fixed", 50);
   DISPLAY_NAME = NULL;
}   

/*****************************************************************/
/**								**/
/**		Program initialization				**/
/**								**/
/*****************************************************************/

XColor allocate_rgb_color(unsigned short r, unsigned short g, unsigned short b)
     /* Generate an XColor from RGB values (non-motif) */
     /* r,g,b range from 0 - 65535		       */
{
   Colormap cmap;
   XColor xc;
   cmap = DefaultColormap(dpy, DefaultScreen(dpy));
   xc.red = r;
   xc.green = g;
   xc.blue = b;
   if (XAllocColor(dpy, cmap, &xc) == 0)
     {
	printf("Problem allocating RGB color (%d, %d, %d)\n", r, g, b);
	exit(0);
     }
   return xc;
}

XColor allocate_named_color(char *name)	/* Generate an XColor from a name */
{					/* (non-motif) */
   XColor xc, exact;
   if (!XAllocNamedColor(dpy, DefaultColormap(dpy, DefaultScreen(dpy)), 
			 name, &xc, &exact))
     {
	printf("\nSorry, the color %s is not supported on this display.\n",
	       name);
	exit(0);
     }
   return xc;
}   
   
GC allocate_GC(XColor xc)		/* Generate a GC from an XColor */
{					/* (non-motif) */
   GC answer;
   XGCValues gcvalues;
   gcvalues.function = GXcopy;
   gcvalues.font = the_font_struct->fid;
   gcvalues.background = backgrd_xc2.pixel;
   gcvalues.foreground = xc.pixel;
   answer = XCreateGC(dpy, db_window,
		      GCFunction | GCForeground | GCFont |
		      GCBackground, &gcvalues);
   return answer;
}   


/*---------------------------------------------------*/
/* Functions to find a color's "GC" from its name,   */
/*   and to create it if it doesn't exist yet.	     */
/* These functions are for both motif & non-motif.   */
/*---------------------------------------------------*/

int lookup_color(char *name)
     /* Return the index of the named color, or -1 if not found */
{
   int i, answer = -1;
   for (i=0; (i<num_colors) && (answer<0); ++i)
     if (strcmp(pallet_name[i], name)==0) answer = i;
   return answer;
}

GC identify_color_from_name(char *name)
     /* Find a color (i.e., a GC) from its name.  Perform color */
     /* allocation if necessary.				*/
{
   int i, indx;
   float r, g, b;
   unsigned short rr, gg, bb;
   indx = lookup_color(name);
   if (indx<0)		/* if this is a new color */
     {
	indx = num_colors++;
	strcpy(pallet_name[indx], name);
	
	if (strncmp(name,"RGB",3)==0)
	    {
		sscanf(name+4, "%f %f %f", &r, &g, &b);
		rr = (unsigned short) (r*65535);
		gg = (unsigned short) (g*65535);
		bb = (unsigned short) (b*65535);
		pallet_xc[indx] = allocate_rgb_color(rr,gg,bb);
	    }
	else
	    pallet_xc[indx] = allocate_named_color(name);
	
	if (db_window != 0)
	    pallet_gc[indx] = allocate_GC(pallet_xc[indx]);

     }
   return pallet_gc[indx];
}


/*--------------------------------------*/
/*-  Initialization			*/
/*--------------------------------------*/

initialize_basic_colors()
     /* Initialize the colors found in the 'BASIC_COLORS' definition */
     /* (formats.h) */
{
   grey_gc = identify_color_from_name("grey");
   black_gc = identify_color_from_name("black");
   green_gc = identify_color_from_name("green");
   red_gc   = identify_color_from_name("red");
   blue_gc  = identify_color_from_name("blue");
   yellow_gc  = identify_color_from_name("yellow");
   white_gc  = identify_color_from_name("white");
   violet_gc  = identify_color_from_name("violet");
   orange_gc  = identify_color_from_name("orange");
   lightblue_gc  = identify_color_from_name("LightBlue");
}


cant_open_display()	/* Can't open display, so print an error message. */
{
   if (DISPLAY_NAME == NULL)
     {
	printf("\nXOpenDisplay failed on your DISPLAY environment ");
	printf("variable display.\n");
	printf("Try setting it to something else, or maybe you haven't ");
	printf("xhosted the machine\nyou are running this from.\n\n");
     }
   else
     {
	printf("\nXOpenDisplay failed on Display: %s\n", DISPLAY_NAME);
	printf("Try setting DISPLAY to something else, or maybe you ");
	printf("haven't xhosted the\n");
	printf("machine you are running this from.\n");
	printf("\n");
     }
   exit(0);
}

void initialize_prog()
{
    Cursor the_cursor, menu_cursor;
    XSetWindowAttributes winvals;
    XGCValues gcvalues;
    XWindowAttributes window_attr;
    XColor backgrd_xc;
    
    if ((dpy = XOpenDisplay (DISPLAY_NAME)) == NULL)
      cant_open_display();
    
    if ((dpy = XOpenDisplay (DISPLAY_NAME)) == NULL)
      cant_open_display();

    if ((the_font_struct = XLoadQueryFont(dpy, FONTNAME)) == NULL) {
	printf("Couldn't allocate font.\n\n");
	exit(-1);
    }

    gcvalues.font = the_font_struct->fid;

    identify_color_from_name(BG_COLOR);		/* sets up pallet_xc[], */
    backgrd_xc = pallet_xc[0];			/* but not pallet_gc yet */
    
     winvals.event_mask = INTERESTING_EVENTS;
     winvals.backing_store = NotUseful;
     winvals.background_pixel = backgrd_xc.pixel;

    db_window = XCreateWindow(dpy, RootWindow(dpy, DefaultScreen(dpy)), 0, 
		    0, DB_WIDTH, DB_HEIGHT, 0, 0, 
		    CopyFromParent, CopyFromParent, CWEventMask | CWBackPixel, 
		    &winvals);

    /* !!! this is non-MOTIF */
    initialize_basic_colors();

    
   XFlush(dpy);

}


void init_db_window()
{
    XSetWindowAttributes winvals;
    XGCValues gcvalues;
    XWindowAttributes window_attr;
    XColor backgrd_xc;

    if (dpy==NIL) {
	printf("Arrange debugger cannot be the 1st window opened.\n");
	exit(-1);
    }

    gcvalues.font = the_font_struct->fid;
    backgrd_xc = allocate_named_color("black");
    winvals.event_mask = 0;
    winvals.backing_store = NotUseful;
    winvals.background_pixel = backgrd_xc.pixel;

    db_black = allocate_GC(backgrd_xc);
    db_grey = allocate_GC(allocate_named_color("grey"));
    db_black = allocate_GC(allocate_named_color("black"));
    db_green = allocate_GC(allocate_named_color("green"));
    db_red   = allocate_GC(allocate_named_color("red"));
    db_blue  = allocate_GC(allocate_named_color("blue"));
    db_yellow  = allocate_GC(allocate_named_color("yellow"));
    db_white  = allocate_GC(allocate_named_color("white"));
    
    XSetStandardProperties(dpy, db_window, "Arrange Graphics",
			   "Arrange Graphics", None, NULL, 0, NULL);
    XMapWindow(dpy, db_window);

    do
      XGetWindowAttributes(dpy, db_window, &window_attr);
    while (window_attr.map_state == IsUnmapped);

    XClearWindow(dpy, db_window);
    db_x0 = DB_WIDTH/2;
    db_y0 = DB_HEIGHT/2;
    XFlush(dpy);
}

void draw_db_axes()
{
   int x0 = DB_WIDTH * 0.87;
   int y0 = DB_HEIGHT * 0.13;
   int size = DB_WIDTH * 0.07;
   int r = 4;
   XDrawLine(dpy, db_window, db_white, x0, y0-size, x0, y0);
   XDrawLine(dpy, db_window, db_white, x0+size, y0, x0, y0);
   XDrawArc(dpy, db_window, db_white, x0-r, y0-r, r*2, r*2, 0, 360*64); 
   XDrawString(dpy, db_window, db_white, x0+size+2, y0+1, "x", 1);
   XDrawString(dpy, db_window, db_white, x0+2, y0-size+1, "y", 1);
   XDrawString(dpy, db_window, db_white, x0+4, y0+8, "z", 1);
   XFlush(dpy);
}

void draw_db_grid()
{
   int i, n=10;		/* Number of grid spaces across the window */
   int xmin, ymin, xmax, ymax, gridsize;
   char str[5];
   xmin = db_x0 - DB_RADIUS;
   ymin = db_y0 - DB_RADIUS;
   xmax = db_x0 + DB_RADIUS;
   ymax = db_y0 + DB_RADIUS;
   gridsize = 2*DB_RADIUS / n;

   for (i=0; i<=n; ++i)
     {
	sprintf(str,"%4.1lf",2.0*i/n - 1.0);
	XDrawLine(dpy, db_window, db_blue,
		  xmin, ymax-i*gridsize,   xmax, ymax-i*gridsize);
	XDrawString(dpy, db_window, db_white,
		    xmin-28, ymax - i*gridsize + 4, str, 4);
	
	XDrawLine(dpy, db_window, db_blue,
		  xmin+i*gridsize, ymin,   xmin+i*gridsize, ymax);
	XDrawString(dpy, db_window, db_white,
		    xmin + i*gridsize - 12, ymax+16, str, 4);
     }
}

void draw_db_hemisphere()
{
   XFillArc(dpy, db_window, db_grey,
	    (db_x0 - DB_RADIUS), (db_y0 - DB_RADIUS),
	    2*DB_RADIUS, 2*DB_RADIUS, 0, 360*64);
}

void set_to_white()
{
   XFillRectangle(dpy, db_window, db_white,
		  db_x0 - DB_WIDTH/2, db_y0 - DB_HEIGHT/2,
		  DB_WIDTH, DB_HEIGHT);
}

void clear_debugger_screen()
{
   XClearWindow(dpy, db_window);
   draw_db_grid();
   draw_db_hemisphere();
   draw_db_axes();
   XFlush(dpy);
}
 
void clear_debugger_screen_BW()
{
   XClearWindow(dpy, db_window);
   set_to_white();
   XFlush(dpy);
}
 
void init_arrange_debugger()
{

   initialize_display_options();
   initialize_prog();

   init_db_window();
   draw_db_grid();
   draw_db_hemisphere();
   draw_db_axes();
   XFlush(dpy);
}
     
/*===============================================================*/
/*=								=*/
/*=		N D B G  Functions				=*/
/*=								=*/
/*===============================================================*/

/*-----------------------------------------------*/
/*  Non-graphical support functions for the NDBG */
/*-----------------------------------------------*/

transform_pt(int *graph_x, int *graph_y, Vector *original)
     /* Convert from unit-sphere coordinates to graphical coordinates */
{
   *graph_x = (int) (db_x0 + DB_RADIUS * original->x);
   *graph_y = (int) (db_y0 - DB_RADIUS * original->y);
}

Boolean verify_unit_vec(Vector *v)
     /* Verify that this is indeed a unit vector.  If not, print error msg. */
{
   Boolean answer;
   answer = equal(sq_norm_of_vec(v), 1.0);
   if (!answer)
     {
	printf("Arrange debugger error: cannot process a non-unit vector\n");
	printf("  Vector in question = (%g, %g, %g)\n", v->x, v->y, v->z);
     }
   return answer;
}

verify_upper_hemi(Vector *v)
     /* Verify that this vector is not in the lower hemisphere. */
     /* If it is, print an error message. */
{
   if (v->z < 0.0)
     {
	printf("Arrange debugger warning: vector is in the lower hemisphere.\n");
	printf("  Vector in question = (%g, %g, %g)\n", v->x, v->y, v->z);
     }
}

Boolean verify_ok_normal(Vector *start, Vector *end, Vector *norm)
     /* Verify that 'norm' is perpendicular to 'start' and 'end', */
     /* and that it has unit length.				  */
{
   Boolean answer;
   answer = equal(dot_product(start, norm), 0.0);
   answer = answer && equal(dot_product(end, norm), 0.0);
   answer = answer && verify_unit_vec(norm);
   if (!answer)
     {
	printf("Arrange debugger error: invalid normal vector.\n");
	printf("  Vector in question = (%g, %g, %g)\n",
	       norm->x, norm->y, norm->z);
     }
}

obtain_correct_normal(Vector *answer, Vector *start, Vector *end, Vector *n)
     /* Find the correct normal vector, according to the following 	*/
     /* rules:								*/
     /*   If start & end are not colinear, use their normalized		*/
     /* cross product.							*/
     /*   If start & end are colinear, use 'n', and run some error-	*/
     /* checking on it.							*/
{
   double dot;
   dot = dot_product(start, end);
   if ((dot > -1.0 + PRECISION) &&
       (dot < -1.0 - PRECISION))
       normalized_cross_product(answer, start, end);
   else
     {
	if (n==NIL)
	  {
	     printf("Arrange debugger error: arc has colinear endpoints and\n");
	     printf("  normal vector was not specified.\n");
	     return;
	  }
	else
	  {
	     copy_vector(answer, n);
	     if (!verify_ok_normal(start, end, answer)) answer = NIL;
	  }
     }
}   

double find_angle(Vector *start, Vector *end)
     /* Find the angle from 'start' to 'end' (always positive).       */
{
   double c;		/* cosine of the angle */
   c = dot_product(start, end);

   /* Handle floating-point errors */
   if (c >= 0.999999) return 0;
   if (c <= -0.999999) return PI;

   return acos(c);
}

rotate_vec(Vector *answer, Vector *start, Vector *axis, double c, double s)
     /* Rotate 'start' about 'axis' by angle theta, where 'c'=cos(theta) */
     /*   and 's'=sin(theta).  Store the result in 'answer'.		 */
     /* (Rodriques' formula, from Craig p.64)				 */
{
   double dot;
   dot = dot_product(start, axis);
   answer->x = start->x * c +
     (axis->y * start->z - axis->z * start->y) * s +
       (1-c) * dot * start->x;
   answer->y =
     start->y * c +
       (axis->z * start->x - axis->x * start->z) * s +
	 (1-c) * dot * start->y;
   answer->z =
     start->z * c +
       (axis->x * start->y - axis->y * start->x) * s +
	 (1-c) * dot * start->z;
}

/*--------------------------------------*/
/*  NDBG graphics functions		*/
/*--------------------------------------*/

void write_vector(Vector *v, GC colr)
     /* Draw a point on the NDBG hemisphere */
{
   int r=2;		/* radius of surrounding circle, in pixels */
   int x,y;
   if (!verify_unit_vec(v)) return;
   verify_upper_hemi(v);

   transform_pt(&x, &y, v);
   XDrawPoint(dpy, db_window, colr, x, y);
   XDrawArc(dpy, db_window, colr, x-r/2, y-r/2, r, r, 0,360*64); 
   XDrawArc(dpy, db_window, colr, x-r, y-r, r*2, r*2, 0,360*64); 
   XFlush(dpy);
}

void write_line(Vector *v1, Vector *v2, GC colr)
     /* Draw a straight line from v1 to v2 */
     /* (Requires a subsequent XFlush) */
{
   int x1, y1, x2, y2;
   if (!verify_unit_vec(v1) || !verify_unit_vec(v2)) return;
   verify_upper_hemi(v1);
   verify_upper_hemi(v2);

   transform_pt(&x1, &y1, v1);
   transform_pt(&x2, &y2, v2);
   XDrawLine(dpy, db_window, colr, x1, y1, x2, y2);
}

void write_arc(Vector *start, Vector *end, Vector *norm, GC colr, Boolean dotted)
{
   double ang;				/* angle between start & end */
   double dtheta = DB_STEPSIZE*TWO_PI/360.0;	/* angular step size (radians) */
   double c, s;				/* cos & sin of dtheta */
   int i, nsteps, m, lastm;
   Vector v[2], normal;		/* 'normal' = normal w/correct orientation */

   /*--- Error-check the input ---*/
   verify_upper_hemi(start);
   verify_upper_hemi(end);
   if (!verify_unit_vec(start) || !verify_unit_vec(end)) return;
   obtain_correct_normal(&normal, start, end, norm);
   /*-----------------------------*/
   
   c = cos(dtheta);
   s = sin(dtheta);
   ang = find_angle(start, end);
   nsteps = (int)(ang/dtheta);
   copy_vector(&v[0], start);

   /* Basic idea: take 'nsteps' steps, each of angle dtheta. */
   /* At each step, compute the new vector by rotating the   */
   /* old one by dtheta about the normal.  Store the new     */
   /* vector in either v[0] or v[1] (alternating), and	     */
   /* draw a line from v[0] to v[1].			     */

   for (i=0, m=0; i<nsteps; ++i)
     {
	lastm = m;
	m = 1-m;
	rotate_vec(&v[m], &v[lastm], &normal, c, s);
	/*--- if dotted, skip ever other segment ---*/
	if (!(dotted && (i%2==1)))
	    write_line(&v[0], &v[1], colr);
     }
   if (!(dotted && (i%2==1)))
       write_line(&v[m], end, colr);
   XFlush(dpy);
}

void debugger_practice_run()	/* (for use in debugging this module) */
{
   Vector q, q1, q2, n;
   init_arrange_debugger();
   specify_vector(&q, 0.57735025, 0.57735027, 0.57735027);
   write_vector(&q, db_green);

   specify_vector(&q1, -1.0, 0.0, 0.0);
   write_vector(&q1, db_yellow);

   specify_vector(&q2, 1.0, 0.0, 0.0);
   write_vector(&q2, db_red);
   XFlush(dpy);

   write_arc(&q, &q1, NIL, db_red,TRUE);
   XFlush(dpy);

   specify_vector(&n, 0.0, 0.5, 0.866);
   write_arc(&q1, &q2, &n, db_green,FALSE);
   XFlush(dpy);

   write_arc(&q, &q2, NIL, db_yellow,TRUE);
   XFlush(dpy);
}   


/*---------------------------------------------------------------------
 * handle_ConfigureNotify_event - this function handles the resize event 
 *                        passed to it.
 */
void handle_ConfigureNotify_event()
{
    XFlush(dpy);
}


/*---------------------------------------------------------------------
 * handle_Expose_event - this function handles the expose event passed to
 *                       it.
 */
void handle_Expose_event(the_A)
     struct Arrange	*the_A;
{
    Arrange_Draw_Graphics(the_A);
}


/*-------------------------------------------------------------------------
 * process_event - this function takes an event off the event queue and
 *             calls the appropriate routine to handle this type of event.
 *-------------------------------------------------------------------------
 */
void process_event(the_event, the_A)
     XEvent *the_event;
     struct Arrange *the_A;
{
    switch (the_event->type)
    {
        case Expose:
                handle_Expose_event(the_A);
		break;
        case ConfigureNotify:
                handle_ConfigureNotify_event();
		break;
        default:
		break;
     }
}
