/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/***********************************************************************/
/**                                                                   **/
/**                            lists.c                                **/
/**                 Functions for operating on lists.                 **/
/**                                                                   **/
/**                   by Bruce Romney, October 1992                   **/
/**             (based on the ACT solid modelling package)            **/
/**                                                                   **/
/***********************************************************************/
#include <stdio.h>
#include "lists.h"


/*----------------------------------------------------------------*/
/*                 List initialization function                   */
/*----------------------------------------------------------------*/

List init_list(p)       /* Initialize a list with a given element */
     Ptr p;
{
   List l;
   l = MAKE_ELT;       /* (allocate memory) */
   l->next = NIL;
   l->prev = NIL;
   l->item = p;
   return l;
}

/*----------------------------------------------------------------*/
/*                 List interrogation functions                   */
/*----------------------------------------------------------------*/

List last(l)
     List l;
     /* Find the last element in the list. */
{
   List elt;
   for (elt=l; NXT(elt)!=NIL; elt=NXT(elt));
   return elt;
}

List first(l)
     List l;
     /* Find the first element in the list containing the given element. */
{
   List elt;
   for (elt=l; PRV(elt)!=NIL; elt=PRV(elt));
   return elt;
}

Boolean memberp(p, l)
     Ptr p;
     List l;
     /* Does list l contain p? */
{
   List elt;
   Boolean answer=FALSE;
   for (elt=l; (elt!=NIL) && (!answer); elt=NXT(elt))
     answer = (GET0(elt)==p);
   return answer;
}

Boolean firstp(l)
     List l;
     /* Is the given element at the head of a list? */
{
   return ( (l->prev)==NIL );
}

Boolean lastp(l)
     List l;
     /* Is the given element at the end of a list? */
{
   return ( (l->next)==NIL );
}

int length(l)
     List l;
     /* Find the length of a list */
{
   List elt;
   int len=0;
   for (elt=l; elt!=NIL; elt=NXT(elt),++len);
   return len;
}

/*----------------------------------------------------------------*/
/*                  List construction functions                   */
/*----------------------------------------------------------------*/

List push(l, p)
     List l;
     Ptr p;
     /* Add an element to the front of a list; return the new list. */
{
   List newl;
   newl = init_list(p);
   if (l!=NIL) l->prev = newl;
   newl->next = l;
   return newl;
}

List add_to_end(l, p)
     List l;
     Ptr p;
     /* Add an element to the end of a list; return the new list. */
{
   List newl=NIL, tail=NIL;
   if (l==NIL)
     {
	newl = init_list(p);
	newl->prev = tail;
	return newl;
     }
   else
     {
	tail = last(l);
	newl = init_list(p);
	tail->next = newl;
	newl->prev = tail;
	return l;
     }
}

List concatenate(l1, l2)
     List l1;
     List l2;
     /* Concatenate one list onto a second; return the new list. */
{
   List tail1;
   if (l2==NIL) return l1;
   if (l1==NIL) return l2;
   tail1 = last(l1);
   tail1->next = l2;
   l2->prev = tail1;
   return l1;
}

void set(elt, p)
     List elt;
     Ptr p;
     /* Replace an element of the list with something new. */
{
   elt->item = p;
}

List copy_list(start, finish)
     List start, finish;
     /* Copy a list from elements 'start' to 'finish'.  */
     /* If finish==NIL, copy the whole list.  Return the new list. */
{
   List answer=NIL, tail, elt, Next;
   if (start==NIL) return answer;
   Next = NIL;
   if (finish!=NIL) Next = NXT(finish);
   
   answer = init_list(GET0(start));          /* the new list */
   tail = answer;                           /* last element of the new list */
   for (elt=NXT(start); elt!=Next; elt=NXT(elt))
     {
	add_to_end(tail,GET0(elt));
	tail = NXT(tail);
     }
   return answer;
}
	
List reverse(start, finish)
     List start, finish;
     /* Reverse the elements of a sublist, from 'start' to 'finish' */
     /* inclusive.  If finish==NIL, reverse the whole list.       */
     /* Return the new list.                                        */
{
   List elt, Previous, Next, prv, nex, last;
   if (start==NIL) return NIL;   
   Previous = PRV(start);
   Next = NIL;
   if (finish!=NIL) Next = NXT(finish);
   
   for (elt=start; elt!=Next; elt=nex)
     {
	last = elt;
	prv = PRV(elt);
	nex = NXT(elt);
	elt->prev = nex;
	elt->next = prv;
     }
   if (Previous!=NIL) Previous->next = last;
   last->prev = Previous;
   if (Next!=NIL) Next->prev = start;
   start->next = Next;
   return last;
}

/*----------------------------------------------------------------*/
/*                    List deletion functions                     */
/*----------------------------------------------------------------*/

void delete_list(l)
     List l;
     /* Delete a list (but not the objects it contains) */
{
   List elt, nxt;
   for (elt=l; elt!=NIL; elt=nxt)
     {
	nxt = NXT(elt);
	FREE(elt);
     }
}   

void delete_list_of_lists(cc)
     List cc;
     /* Delete a list of lists. */
{
   List onelist, elt;
   for (elt=cc; elt!=NIL; elt=NXT(elt))
     {
	onelist = (List)(elt->item);
	FREE(onelist);
     }
   FREE(cc);
}

List delete_elt(first, elt)
     List first, elt;
     /* Remove a single element from the list which starts at 'first'; */
     /* return the new list.					       */
{
   List pr,nx;
   if ((first==NIL) || (elt==NIL)) return first;
   pr = PRV(elt);
   nx = NXT(elt);
   
   if (pr==NIL)    /* If it's at the head of the list */
     {
	if (nx!=NIL) nx->prev = NIL;
	FREE(elt);
	return nx;
     }
   else             /* If it's not at the head of the list */
     {
	pr->next = nx;
	if (nx!=NIL) nx->prev = pr;
	FREE(elt);
	return first;
     }
}

List delete_obj(p, l)
     Ptr p;
     List l;
     /* Delete all instances of object p from list l (but don't delete */
     /* p itself).  Return a pointer to the modified list.             */
{
   List newlist, elt;
   newlist = l;
   for (elt=l; elt!=NIL; elt=NXT(elt))
     if (GET0(elt)==p) newlist = delete_elt(newlist, elt);
   return newlist;
}

Ptr pop_element(List *l)
     /* Delete the top element off the list, and return the object */
     /* that it used to point to.				   */
     /*   In conjunction with the POP macro in lists.h, we can use
	this function as in the following example:
	{ Vector *v;  List vectorlist;  v = POP(vectorlist, Vector); }  */
{
   Ptr answer;
   answer = GET0(*l);
   *l = delete_elt(*l, *l);
   return answer;
}

List clip(List first, List elt)
     /* Remove 'elt' from the list beginning with 'first', and	*/
     /*   return the new list.					*/
     /* In other words, this is just like 'delete_elt()'	*/
     /*   except it doesn't free anything.			*/
{
   List pr, nx;
   
   pr = PRV(elt);
   nx = NXT(elt);
   elt->next = elt->prev = NIL;

   if (pr==NIL)    /* If it's at the head of the list */
     {
	if (nx!=NIL) nx->prev = NIL;
	return nx;
     }
   else             /* If it's not at the head of the list */
     {
	pr->next = nx;
	if (nx!=NIL) nx->prev = pr;
	return first;
     }
}   

/*-----------------*/
/*  Main routines  */
/*-----------------*/
/*
void print_list_of_ints(l)
     List l;
{
   List elt;
   for (elt=l; elt!=NIL; elt=NXT(elt))
     printf("%d\n",*(GET(elt,int)));
}

main()
{
   List a=NIL, b=NIL, c;
   int n[10], i;
   for (i=0;i<10;++i)
     n[i]=i;

   for (i=1;i<10;++i)
     a= add_to_end(a,&n[i]);
   print_list_of_ints(a);
   printf("---\n");
   a = delete_obj(&n[4],a);
   print_list_of_ints(a);
}
*/
  
