/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * memory_debug.c
 *
 * A *debug* package to monitor memory management
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/

#include "memory_debug.h"

#ifndef NULL
#define NULL 0
#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef Boolean
#define Boolean char
#endif



extern  Boolean mem_debug;
extern  void*   bug_ptr;

/*--- do full verification after EVERY operation??? ---*/
static Boolean  do_verify = FALSE;

typedef struct Block {
    unsigned int	size;
    void*		ptr;
} Block;

typedef struct Item* Finger;

typedef struct Item {
    Block		block;
    struct Item	*l;	/* left child */
    struct Item	*r;	/* right child */
    struct Item *p;	/* parent */
} Item;


typedef struct Dictionary {
    int		num_items;
    Finger	root;
} Dictionary;

typedef int Bool;

/*-----------------------------*/
/*--- Dictionary Prototypes ---*/
/*-----------------------------*/
Finger D_Insert(Dictionary*,Block*);
Finger D_Search(Dictionary*,void*);
D_Remove(Dictionary*,Finger);
Finger D_Leftmost(Dictionary*);
Finger D_Rightmost(Dictionary*);
Finger D_Succ(Dictionary*,Finger);
Finger D_Pred(Dictionary*,Finger);

Boolean D_Traverse(Dictionary*,Boolean);
#define D_Verify(d)  D_Traverse(d,FALSE)
#define D_Print(d)   D_Traverse(d,TRUE)

/*-----------------------------------------------------------------*/

static struct Dictionary the_d = { 0, NULL };

/*--- trace a specific memory block ---*/
static Finger	bug_item = NULL;
static Finger	bug_pred = NULL;

print_debug_message(char *msg)
{
   printf("%s", msg);
}


Boolean In_Block(block,ptr)
  Block		*block;
  void*		ptr;
{
    Boolean	result;

    result = ((ptr >= block->ptr) &&
	      (ptr <  ((void*) (((unsigned long) block->ptr) +
				block->size))));

    return(result);
}

void* malloc__debug(size)
  unsigned	size;
{
    void	*new_ptr;
    new_ptr = (void*) malloc(size);

    if (mem_debug)
	new_block(new_ptr,size);

    return(new_ptr);
}


void* calloc__debug(num,size)
  unsigned	num;
  unsigned	size;
{
    void	*new_ptr;

    new_ptr = (void*) calloc(num,size);

    if (mem_debug)
	new_block(new_ptr,num*size);

    return(new_ptr);
}


void free__debug(ptr)
  void* ptr;
{
    Finger	f;
    char errormsg[80];

    if (mem_debug) {
	f = D_Search(&the_d, ptr);

	if (f == NULL) {
	    sprintf(errormsg,"Memory Error --- Freeing (%p), but not owned.\n",ptr);
	    print_debug_message(errormsg);
	} else {
	    if (In_Block(&f->block,bug_ptr)) {
		bug_item = bug_pred = NULL;
		print_debug_message("Mem_Trace --> Freeing bug_ptr\n");
	    } else if (f == bug_pred) {
		bug_pred = D_Pred(&the_d,bug_item);
		print_debug_message("Mem_Trace --> Pred deleted\n");
	    }
	    D_Remove(&the_d, f);
	}




	if (do_verify) {
	    if (!(D_Verify(&the_d))) {
		sprintf(errormsg,"Memory Error --- Verification failed!!!\n");
		print_debug_message(errormsg);
	    }
	}
    }

    free(ptr);
}


new_block(ptr,size)
  void		*ptr;
  unsigned int	size;
{
    Block	n_block;
    Finger	f,g;
    char errormsg[80];

    n_block.size = size;
    n_block.ptr = ptr;

    f = D_Insert(&the_d, &n_block);

    /*--- this next line will create MUCH output ---*/
    /*D_Print(&the_d);*/

    if (In_Block(&n_block,bug_ptr)) {
	if (bug_item != NULL) {
	    print_debug_message("Mem_Trace --> About to allocate...but not freed yet.\n");
	}
	bug_item = f;
	bug_pred = D_Pred(&the_d,f);
	print_debug_message("Mem_Trace -->  Allocating bug_ptr\n");
    }

    if (do_verify) {
	if (!D_Verify(&the_d)) {
	    sprintf(errormsg,"Memory Error --- Verification failed!!!\n");
	    print_debug_message(errormsg);
	}

	g = D_Search(&the_d, ptr);
	if (f!=g) {
	    print_debug_message("Problem..can't find what I just inserted!!!\n");
	}	
    }
    
}



/*-----------------------------------------------------------------*/

#define KEY(b)   ((b).ptr)



Finger	D_Insert(d,b)
  Dictionary	*d;
  Block		*b;
{
    Item	*new_item;
    Item	*trav1,*trav2;

    d->num_items ++;
    new_item = (Item *) malloc(sizeof(Item));

    new_item->block.size  = b->size;
    new_item->block.ptr  = b->ptr;
    new_item->l = new_item->r = new_item->p = NULL;

    for (trav1=trav2=d->root; (trav2!=NULL); ) {
	trav1 = trav2;
	
	trav2 = ((KEY(*b) < KEY(trav1->block)) ?
		 trav1->l : trav1->r);
    }
    
    new_item->p = trav1;

    if (trav1 == NULL) {
	d->root = new_item;
    } else {
	if (KEY(*b) < KEY(trav1->block))
	    trav1->l = new_item;
	else
	    trav1->r = new_item;
    }
}




Finger	D_Search(d,key)
  Dictionary	*d;
  void*		key;
{
    Item	*trav, *trav2;

    trav = trav2 = d->root;
    while ((trav!=NULL) && (key != KEY(trav->block))) {
       trav2 = trav;
       trav = ((key < KEY(trav->block)) ?
	       trav->l : trav->r);
    }
    
    return(trav);
}



D_Remove(d,f)
  Dictionary	*d;
  Finger	f;
{
    Item	*to_delete;
    Item	*repl;
    Boolean	rchild;

    d->num_items--;

    to_delete = f;

    if ((to_delete->l != NULL) && (to_delete->r != NULL)) {
	/*--- two children ---*/
	repl = D_Succ(d,to_delete);
	to_delete->block  = repl->block;

	/*---    now we just need to delete successor    ---*/
	/*--- (which is guaranteed to have 0 or 1 child) ---*/
	to_delete = repl;
	repl = NULL;
    }


    /*--- at this point, to_delete has at most one child ---*/

    if ((to_delete->l != NULL) || (to_delete->r != NULL)) {
	/*--- one child ---*/
	rchild = (to_delete->r != NULL);
	repl = (rchild ? to_delete->r : to_delete->l);

	/*--- fix up child's parent pointer ---*/
	repl->p = to_delete->p;

	/*--- we make parent point to 'repl' later ---*/
    } else {
	/*--- no child ---*/

	/*--- we make parent point to 'repl' later ---*/
	repl = NULL;
    }


    /*--- for ALL cases, we still need to fixup parent's child pointer ---*/
    /*--- (we assume 'repl' is set properly for the specific case) ---*/
    
    if (to_delete->p) {
	if (to_delete == to_delete->p->l) {
	    /*-- to_delete is a left child --*/
	    to_delete->p->l = repl;
	} else {
	    /*-- to_delete is a right child --*/
	    to_delete->p->r = repl;
	}
    } else {
	/*--- must be root! ---*/
	d->root = repl;
    }

    free(to_delete);
}



Finger	D_Leftmost(d)
  Dictionary	*d;
{
    Item	*trav;

    if (d->root) {
	for (trav = d->root;  (trav->l != NULL); trav=trav->l);
    } else {
	trav = NULL;
    }

    return(trav);
}


Finger	D_Rightmost(d)
  Dictionary	*d;
{
    Item	*trav;

    if (d->root) {
	for (trav = d->root;  (trav->r != NULL); trav=trav->r);
    } else {
	trav = NULL;
    }

    return(trav);
}


Finger	D_Succ(d,f)
  Dictionary	*d;
  Finger	f;
{

    Item	*trav;

    trav = f;

    if (trav->r) {
	for (trav = trav->r;  (trav->l != NULL); trav=trav->l);
    } else {
	for ( ; (trav->p && (trav == trav->p->r)); trav=trav->p);
	trav = trav->p;
    }

    return(trav);
}


Finger	D_Pred(d,f)
  Dictionary	*d;
  Finger	f;
{

    Item	*trav;

    trav = f;

    if (trav->l) {
	for (trav = trav->l;  (trav->r != NULL); trav=trav->r);
    } else {
	for ( ; ((trav->p) && (trav == trav->p->l)); trav=trav->p);
	trav = trav->p;
    }

    return(trav);
}


Boolean D_Traverse(Dictionary *d, Boolean print_me)
{
    Item	*trav;
    Item	*succ;
    int		count;
    Boolean	okay;

    if (print_me) printf("Memory dump...\n");

    okay = TRUE;

    count = 0;
    for(trav=D_Leftmost(d); (trav!=NULL); trav = succ) {

	succ = D_Succ(d,trav);
	count++;

	if (print_me)
	    printf("D_Item: (%p) with size (%d)\n",
		   trav->block.ptr,trav->block.size);

	if (succ && (KEY(trav->block) > KEY(succ->block)))
	    okay = FALSE;
    }

    if (count != d->num_items) okay = FALSE;

    if (print_me && !okay) {
	printf("\nFailed D_Verify\n");
    }

    if (print_me) printf("\n\n");

    return(okay);
}



