/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * polygon.c
 *
 * Functions for operating on polygons
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/

#include <stdio.h>
#include <strings.h>
#include "lists.h"
#include "polygon.h"
#include "spheregeom.h"
#include "standards.h"


/*************************************************************/
/**
  static variables containing vectors for
  bot horizon and top pole.
**/
Vector	vec_horizon = {0.0, 0.0, 1.0};
Vector	vec_pole =    {0.0, 1.0, 0.0};
Vector  vec_zero =    {0.0, 0.0, 0.0};

/*************************************************************/

/*************************************************************
    Function:  Polygon_Interior(poly,point)
    Input:     poly -- a spherical polygon
               point -- a point on the sphere
    Output:    TRUE -- point is in the interior of poly
               FALSE -- point is not in the interior of poly
    Effects:   <none>

    Description:  This routine is meant to test whether a point
      is in the *interior* of a polygon.  It is somewhat unpredictable
      for points lying exactly on the boundary.  This is due
      somewhat in part to the inherent floating point error.

      This routine does not even bother to check any polygon
      boundary with interior=FALSE, rather it simply ignores it.

**************************************************************/
Boolean Polygon_Interior(poly,point)
  Polygon*	poly;
  Vector	point;
{

    /**
      For the first effort, this will be written in a very brute
      force manner.  There are probably more efficient methods
      relying on preprocessing.  Also, there are probably methods
      less susceptible to floating point error.

      This will pick a vertex of some boundary which has an interior,
      and then will find the closest polygon edge crossed on the
      way from the query point to that vertex.  Once the closest
      edge is known, then it is easy to see if the path is crossing
      the edge from left to right or right to left.
    **/

    Boolean		result;

    List		elt_boundary;	/* traverse list of boundaries */
    List		elt_side;	/* traverse list of sides */
    PolygonBoundary*	cur_boundary;
    PolygonSeg*		cur_side;

    Vector	walk_goal;
    Vector	walk_edge;
    Vector	closest_point;
    PolygonSeg*	closest_edge;

    Vector	intersect1,intersect2;
    Boolean	hit1,hit2;



    result = FALSE;

    /*--- find the first boundary with interior=TRUE, if exists ---*/
    for (elt_boundary = poly->boundaries;
	 ((elt_boundary) &&
	  (!(cur_boundary =
	     GET(elt_boundary,PolygonBoundary))->interior));
	 elt_boundary = NXT(elt_boundary));

    if (elt_boundary) {
	/*-- found an interior part of the polygon --*/
	cur_side = GET(cur_boundary->sides,PolygonSeg);

	if (!cur_side) {
	    /*---- must be FullSphere ---*/
	    result = TRUE;
	    return(result);
	}


	if (cur_side->begin) {
	    walk_goal = *cur_side->begin;
	} else {
	    /*-- must be a hemisphere --*/
	    if ON_HORIZON_NORM(*cur_side->edge) {
		walk_goal = vec_pole;
	    } else {
		walk_goal.x = cur_side->edge->y;
		walk_goal.y = - cur_side->edge->x;
		walk_goal.z = 0.0;
	    }
	}


	/*--- directed edge is the 'shortest' route to walk_goal ---*/
	CROSS(walk_edge,point,walk_goal);

	if (ZERO_VEC(walk_edge)) {
	    /*--- cross product failed...pick path by hand ---*/
	    if (ON_HORIZON_VEC(walk_goal)) {
		walk_edge = vec_horizon;
	    } else {
		CROSS(walk_edge,point,vec_pole);
	    }
	}

	closest_edge = cur_side;
	closest_point = walk_goal;
	

	/*--- Now, we've got a path to walk, so find the closest
	  intersection along that path ---*/
	for ( /* elt_boundary is already set */ ; (elt_boundary);
	     elt_boundary = NXT(elt_boundary)) {
	    cur_boundary = GET(elt_boundary,PolygonBoundary);

	    if (cur_boundary->interior) {
		for (elt_side = cur_boundary->sides; (elt_side);
		     elt_side = NXT(elt_side)) {
		    cur_side = GET(elt_side,PolygonSeg);

		    /*--
		      must find (first) intersection
		      between path and cur_side
		    --*/
		    CROSS(intersect1, walk_edge, *cur_side->edge);
		    if (!ZERO_VEC(intersect1)) {
			intersect2 = intersect1;
			NEG(intersect2);
			hit1 = Closer(point,walk_edge,
				      intersect1,walk_goal,FALSE);
			hit2 = Closer(point,walk_edge,
				      intersect2,walk_goal,FALSE);
			    
			if (hit2 &&
			    (!hit1 || Closer(point,walk_edge,
					     intersect2, intersect1,
					     FALSE))) {
			    /*--- intersect2 was first hit ---*/
			    NEG(intersect1);
			    hit1 = TRUE;
			}

			/*---
			  So at this point, intersect1 is set to be
			  the first hit (if any), and hit1=TRUE if it
			  did, indeed, hit the current side.
			---*/

			if (hit1 &&
			    Closer(point,walk_edge,intersect1,
				   closest_point,FALSE)) {
			    /*--- we found a closer point (or a tie) ---*/	
			    /*---   <this tiebreak code is a mess> ---*/

			    if ((!VectorSame(intersect1, closest_point)) ||
				(!cur_side->begin) ||
				(!closest_edge->begin) ||
				(SmallerAngle(walk_edge,
					      intersect1,
					      *cur_side->edge,
					      VectorSame(intersect1,
							 *cur_side->begin),
					      *closest_edge->edge,
					      VectorSame(intersect1,
							 *closest_edge->begin),
					      FALSE))) {

				/*--- found a new closer point ---*/
				closest_point = intersect1;
				closest_edge = cur_side;
			    }
			}
		    }
		}
	    }
	}

	result = Cross_Left(walk_edge, *closest_edge->edge,
			    closest_point, FALSE);

    }

    return(result);
}



/*************************************************************
    Function:  Compile_Vector_List(poly)
    Input:     poly -- a spherical polygon
    Output:    <none>
    Effects:   Gathers the Vectors, removing duplicates

    Description:  This routine is designed to gather all
      vector poitners into one list, removing duplicates.

**************************************************************/
List	Compile_Vector_List(poly)
  Polygon*	poly;
{
    List		vectors;
    List		elt_boundary;	/* traverse list of boundaries */
    List		elt_side;	/* traverse list of sides */
    PolygonSeg*		cur_side;


    vectors = NULL;
    for (elt_boundary = poly->boundaries; (elt_boundary);
	 elt_boundary = NXT(elt_boundary)) {
	
	for (elt_side = GET(elt_boundary,PolygonBoundary)->sides;
	     (elt_side); elt_side = NXT(elt_side)) {
	    cur_side = GET(elt_side,PolygonSeg);

	    if (cur_side->begin && (!memberp(cur_side->begin, vectors)))
		vectors = push(vectors,cur_side->begin);
	    if (cur_side->edge && (!memberp(cur_side->edge, vectors)))
		vectors = push(vectors,cur_side->edge);
	    if (cur_side->end && (!memberp(cur_side->end, vectors)))
		vectors = push(vectors,cur_side->end);
	}
    }

    vectors = reverse(vectors,NULL);
    return(vectors);
}



/*************************************************************
    Function:  Polygon_Free(poly)
    Input:     poly -- a spherical polygon
    Output:    <none>
    Effects:   Frees up all memory from a polygon

    Description:  This routine is designed to free all memory
      that is a part of the polygon structure.   This is a
      non-trivial task since some of the geometry structures
      are shared and thus pointed to in several places.

**************************************************************/
void Polygon_Free(poly,Free_Data)
  Polygon*		poly;
  Polygon_Data_Freer	Free_Data;
{
    Vector		*cur_vec;
    List		vectors;
    List		elt;
    List		elt_boundary;	/* traverse list of boundaries */
    List		elt_side;	/* traverse list of sides */

    vectors = Compile_Vector_List(poly);

    for (elt=vectors; (elt); elt=NXT(elt)) {
	FREE(GET(elt,Vector));
    }
    delete_list(vectors);

    for (elt_boundary = poly->boundaries; (elt_boundary);
	 elt_boundary = NXT(elt_boundary)) {
	
	for (elt_side = GET(elt_boundary,PolygonBoundary)->sides;
	     (elt_side); elt_side = NXT(elt_side)) {
	    FREE(GET(elt_side,PolygonSeg));
	}
	delete_list(GET(elt_boundary,PolygonBoundary)->sides);
	FREE(GET(elt_boundary,PolygonBoundary));
    }
    delete_list(poly->boundaries);
    if (Free_Data) (*Free_Data)(poly->data);
    FREE(poly);

}


/*************************************************************
                 stuff for printing/reading
*************************************************************/


static char	keyword_num_vectors[] = "%num_vectors:";
static char	keyword_polygon[] = "%polygon";
static char	keyword_num_bound[] = "num_boundaries:";
static char	keyword_boundary[] = "%polygon_boundary";
static char	keyword_num_sides[] = "num_sides:";
static char	keyword_side[] = "%polygon_side";
static char	keyword_interior[] = "interior:";
static char	keyword_begin[] = "begin:";
static char	keyword_end[] = "end:";
static char	keyword_edge[] = "edge:";
static char	keyword_null[] = "<null>";
static char	keyword_true[] = "true";
static char	keyword_false[] = "false";
static char	keyword_data[] = "%data";

static char	keyword_part[] = "part:";


int	Get_Index_In_List(the_list,the_item)
  List		the_list;
  void		*the_item;
{
    int		count;
    List	elt;

    for (elt=the_list, count=0;
	 (elt && (GET0(elt) != the_item));
	 elt = NXT(elt), count++);

    if (!elt)
	count = -1;

    return(count);
}



/*************************************************************
    Function:  Polygon_Print(fp,poly,Print_Data)
    Input:     fp -- an output stream
               poly -- a polygon
	       Print_Data -- pointer to function to print data
    Output:    <none>
    Effects:   Dumps polygon info to fp

    Description:  This prints the polygon to file in a form
                  that is readable by Polygon_Read()

**************************************************************/
void Polygon_Print(fp,poly,Print_Data)
  FILE*			fp;
  Polygon*		poly;
  Polygon_Data_Printer	Print_Data;
{
    List		elt_boundary;	/* traverse list of boundaries */
    List		elt_side;	/* traverse list of sides */
    PolygonBoundary*	cur_boundary;
    PolygonSeg*		cur_side;

    Vector		*cur_vec;
    List		vectors;
    List		elt;

    /*--- check for bad params ---*/
    if ((!fp) || (!poly)) return;

    /*---- print a dump of the crude polygon ----*/
    fprintf(fp,"%s\n",keyword_polygon);

    vectors = Compile_Vector_List(poly);
    fprintf(fp,"%s %d\n",keyword_num_vectors,length(vectors));
    
    for (elt=vectors; (elt); elt=NXT(elt)) {
	cur_vec = GET(elt,Vector);
	fprintf(fp,"  %lf %lf %lf\n",
		cur_vec->x, cur_vec->y, cur_vec->z);
    }

    fprintf(fp,"%s %d\n",keyword_num_bound,
	    length(poly->boundaries));

    for (elt_boundary = poly->boundaries; (elt_boundary);
	 elt_boundary = NXT(elt_boundary)) {
	cur_boundary = GET(elt_boundary,PolygonBoundary);

	fprintf(fp,"%s\n",keyword_boundary);
	fprintf(fp,"%s %s\n", keyword_interior,
		(cur_boundary->interior ? keyword_true : keyword_false));
	fprintf(fp,"%s %d\n",keyword_num_sides,
		length(cur_boundary->sides));

	for (elt_side = cur_boundary->sides;  (elt_side);
	     elt_side = NXT(elt_side)) {
	    cur_side = GET(elt_side,PolygonSeg);

	    fprintf(fp,"%s\n",keyword_side);
	    fprintf(fp,"%s %i\n", keyword_begin,
		    Get_Index_In_List(vectors,cur_side->begin));
	    fprintf(fp,"%s %i\n", keyword_edge,
		    Get_Index_In_List(vectors,cur_side->edge));
	    fprintf(fp,"%s %i\n", keyword_end,
		    Get_Index_In_List(vectors,cur_side->end));
	}
    }

    fprintf(fp,"%s\n",keyword_data);
    if (Print_Data)
	(*Print_Data)(fp,poly->data);
}




Boolean Check_Keyword(fp,keyword)
  FILE*		fp;
  char		*keyword;
{
    char	buffer[30];
    Boolean	error;


    error = ((fscanf(fp," %30s ",buffer)!=1) ||
	      (strcmp(buffer,keyword)!=0));
    
    if (error) {
	printf("Can't read keyword '%s'\n",keyword);
	printf("It read '%s'\n",buffer);
    }

    return(error);
}




/*************************************************************
    Function:  Polygon_Read(fp,Read_Data)
    Input:     fp -- an input stream
               Read_Data -- pointer to function to read data
    Output:    newly created polygon structure
    Effects:   Reads polygon from fp

    Description:  This reads polygons from a file created by
                  Polygon_Print()

**************************************************************/
Polygon* Polygon_Read(fp,Read_Data)
  FILE*			fp;
  Polygon_Data_Reader	Read_Data;
/*  void*		(*Read_Data)(FILE*);*/
{
    char		buffer[30];
    Boolean		error;

    Polygon*		newpoly;
    PolygonBoundary*	newbound;
    PolygonSeg*		newside;

    int			i,j,vec_ind;
    int			num_boundaries;
    int			num_sides;
    int			num_vecs;
    Vector		**vec_array;

    List		elt;


    error = FALSE;
    newpoly = NULL;
    vec_array = NULL;
    
    while (TRUE) {	/* loop to break out of on error */

	if (feof(fp)) {
	    error = TRUE;
	    break;
	}

	if (Check_Keyword(fp,keyword_polygon)) {
	    error = TRUE;
	    break;
	}

	if (Check_Keyword(fp,keyword_num_vectors)) {
	    error = TRUE;
	    break;
	}

	if ((fscanf(fp," %30s ",buffer)!=1) ||
	    ((sscanf(buffer," %i",&num_vecs)!=1))) {
	    printf("  Can't read number of vectors.\n");
	    printf("  It read '%s'\n",buffer);
	    error = TRUE;
	    break;
	} else {
	    vec_array = ALLOCATE_ARRAY(Vector*,num_vecs);
	    for (vec_ind=0; vec_ind<num_vecs; vec_ind++) {
		vec_array[vec_ind] = ALLOCATE(Vector);
	    }
	    for (vec_ind=0; vec_ind<num_vecs; vec_ind++) {
		if (fscanf(fp," %lf %lf %lf ",
			   &vec_array[vec_ind]->x,
			   &vec_array[vec_ind]->y,
			   &vec_array[vec_ind]->z)!=3) {
		    printf("  Error reading vector(%d)\n",vec_ind);
		    error = TRUE;
		    break;
		}
	    }
	    if (error) break;
	}


	if (Check_Keyword(fp,keyword_num_bound)) {
	    error = TRUE;
	    break;
	}

	if ((fscanf(fp," %30s ",buffer)!=1) ||
	    (sscanf(buffer," %i",&num_boundaries)!=1)) {
	    printf("  Can't read number of boundaries.\n");
	    printf("  It read '%s'\n",buffer);
	    error = TRUE;
	    break;
	}

	newpoly = ALLOCATE(Polygon);
	newpoly->boundaries = newpoly->data = NULL;

	for (i=0; i<num_boundaries; i++) {
	    /*---- Read a single boundary ----*/
	    newbound = ALLOCATE(PolygonBoundary);
	    newbound->sides = NULL;

	    if (Check_Keyword(fp,keyword_boundary)) {
		error = TRUE;
		break;
	    }

	    if (Check_Keyword(fp,keyword_interior)) {
		printf("  Boundary(%d)\n",i);
		error = TRUE;
		break;
	    }

	    if ((fscanf(fp," %30s ",buffer)!=1) ||
		((strcmp(buffer,keyword_true)!=0) &&
		 (strcmp(buffer,keyword_false)!=0))) {
		printf("  Can't read '%s' or '%s'\n",
		       keyword_true,keyword_false);
		printf("  It read '%s'\n",buffer);
		error = TRUE;
		break;
	    } else {
		newbound->interior =
		    (strcmp(buffer,keyword_true)==0);
	    }


	    if (Check_Keyword(fp,keyword_num_sides)) {
		error = TRUE;
		break;
	    }

	    if ((fscanf(fp," %30s ",buffer)!=1) ||
		(sscanf(buffer," %i",&num_sides)!=1)) {
		printf("  Can't read number of sides.\n");
		printf("  It read '%s'\n",buffer);
		error = TRUE;
		break;
	    }


	    for (j=0; j<num_sides; j++) {
		newside = ALLOCATE(PolygonSeg);

		/*---- Read a single side ----*/
		if (Check_Keyword(fp,keyword_side)) {
		    error = TRUE;
		    break;
		}

		if (Check_Keyword(fp,keyword_begin)) {
		    error = TRUE;
		    break;
		}
		if ((fscanf(fp," %30s ",buffer)!=1) ||
		    (sscanf(buffer," %i",&vec_ind)!=1)) {
		    printf("  Can't read begin vector index.\n");
		    printf("  It read '%s'\n",buffer);
		    error = TRUE;
		    break;
		} else {
		    newside->begin =
			((vec_ind<0) ? NULL : vec_array[vec_ind]);
		}


		if (Check_Keyword(fp,keyword_edge)) {
		    error = TRUE;
		    break;
		}
		if ((fscanf(fp," %30s ",buffer)!=1) ||
		    (sscanf(buffer," %i",&vec_ind)!=1)) {
		    printf("  Can't read edge vector index.\n");
		    printf("  It read '%s'\n",buffer);
		    error = TRUE;
		    break;
		} else {
		    newside->edge =
			((vec_ind<0) ? NULL : vec_array[vec_ind]);
		}


		if (Check_Keyword(fp,keyword_end)) {
		    error = TRUE;
		    break;
		}
		if ((fscanf(fp," %30s ",buffer)!=1) ||
		    (sscanf(buffer," %i",&vec_ind)!=1)) {
		    printf("  Can't read end vector index.\n");
		    printf("  It read '%s'\n",buffer);
		    error = TRUE;
		    break;
		} else {
		    newside->end =
			((vec_ind<0) ? NULL : vec_array[vec_ind]);
		}

		newbound->sides = add_to_end(newbound->sides, newside);
	    }
	    if (error) {
		FREE(newside);
		for (elt=newbound->sides; (elt); elt=NXT(elt)) {
		    FREE(GET(elt,PolygonSeg));
		}
		delete_list(newbound->sides);
		break;
	    }

	    newpoly->boundaries = add_to_end(newpoly->boundaries,newbound);
	}
	if (error) {
	    FREE(newbound);
	    for (elt=newpoly->boundaries; (elt); elt=NXT(elt)) {
		FREE(GET(elt,PolygonBoundary));
	    }
	    delete_list(newpoly->boundaries);
	    break;
	}

	if (Check_Keyword(fp,keyword_data)) {
	    error = TRUE;
	    break;
	}
    
	if (Read_Data)
	    newpoly->data = (*Read_Data)(fp);
    
	/*---- always force a break out of the while loop ----*/
	break;

    }	/*  end of while(TRUE) loop */

    if (error) {
	if (newpoly) FREE(newpoly);
	newpoly = NULL;

	if (vec_array)
	    for (vec_ind=0; vec_ind<num_vecs; vec_ind++) {
		FREE(vec_array[vec_ind]);
	    }
    }

    if (vec_array) FREE(vec_array);

    return(newpoly);
}






/*************************************************************
    Function:  TwoParts_NicePrint(fp,tp)
    Input:     fp -- an output stream
               tp -- a pointer to a TwoParts structure
    Output:    <none>
    Effects:   Dumps TwoParts structure to fp

    Description:  This prints the TwoParts structure to file
                  in a form pleasing to humans

**************************************************************/
void TwoParts_NicePrint(fp,tp)
  FILE		*fp;
  void		*tp;
{
    TwoParts	*tp_cast;

    tp_cast = (TwoParts*) tp;
    fprintf(fp,"(%d,%d)", tp_cast->part1->id, tp_cast->part2->id);
}



/*************************************************************
    Function:  TwoParts_Print(fp,tp)
    Input:     fp -- an output stream
               tp -- a pointer to a TwoParts structure
    Output:    <none>
    Effects:   Dumps TwoParts structure to fp

    Description:  This prints the TwoParts structure to file
                  in a form readable by TwoParts_Read()

**************************************************************/
void TwoParts_Print(fp,tp)
  FILE		*fp;
  void		*tp;
{
    TwoParts	*tp_cast;

    tp_cast = (TwoParts*) tp;
    fprintf(fp,"%s %d\n",keyword_part,tp_cast->part1->id);
    fprintf(fp,"%s %d\n",keyword_part,tp_cast->part2->id);
}


/*************************************************************
    Function:  TwoParts_Read(fp)
    Input:     fp -- an input stream
    Output:    newly created TwoParts structure
    Effects:   Reads TwoParts structure from fp

    Description:  This reads TwoParts structure from a file
                  created by TwoParts_Print()

**************************************************************/
void* TwoParts_Read(fp)
  FILE*		fp;
{
    TwoParts	*tp;
    int		id1,id2;
    Boolean	error;
    char	buffer[30];

    error = FALSE;

    while (TRUE) {
	if (Check_Keyword(fp,keyword_part)) {
	    error = TRUE;
	    break;
	}

	if ((fscanf(fp," %30s ",buffer)!=1) ||
	    (sscanf(buffer," %i",&id1)!=1)) {
	    printf("  Can't read id1.\n");
	    printf("  It read '%s'\n",buffer);
	    error = TRUE;
	    break;
	}

	if (Check_Keyword(fp,keyword_part)) {
	    error = TRUE;
	    break;
	}

	if ((fscanf(fp," %30s ",buffer)!=1) ||
	    (sscanf(buffer," %i",&id2)!=1)) {
	    printf("  Can't read id1.\n");
	    printf("  It read '%s'\n",buffer);
	    error = TRUE;
	    break;
	}

	/*---- always force a break out of the while loop ----*/
	break;
    }	/*  end of while(TRUE) loop */
	

    if (error) {
	tp = NULL;
    } else {
	tp = ALLOCATE(TwoParts);
	tp->part1 = ALLOCATE(Part);
	tp->part2 = ALLOCATE(Part);
	tp->part1->id = id1;
	tp->part2->id = id2;
    }


    return((void*) tp);
}


/*************************************************************
    Function:  TwoParts_Free(fp)
    Input:     tp -- a pointer to a TwoParts structure
    Output:    <none>
    Effects:   Frees the TwoParts structure

**************************************************************/
TwoParts_Free(tp)
  void		*tp;
{
    TwoParts	*tp_cast;

    tp_cast = (TwoParts*) tp;
    
    FREE(tp_cast->part1);
    FREE(tp_cast->part2);
    FREE(tp_cast);

}





#ifdef DEBUG

Vector Get_InputPoint2()
{
    Vector		point;
    double		temp;

    printf("Point: ");
    if ((scanf(" %lf %lf %lf",&point.x, &point.y, &point.z))!=3) {
	printf("Error reading input point...exiting...\n");
	exit(-1);
    }

    if ((temp=NORM2(point)) == 0.0) {
	printf ("input point was (0,0,0)...bad choice\n");
	exit(-1);
    } else {
	temp = 1/sqrt(temp);
	point.x *= temp;
	point.y *= temp;
	point.z *= temp;
    }

    return(point);
}


int Get_InputPolyID(num_polys)
  int	num_polys;
{
    int	i;

    printf("PolyID: ");
    scanf(" %d",&i);
    if ((i<0) || (i>=num_polys)) {
	printf("Bad polyID...using 0.\n");
	i=0;
    }

    return(i);
}


void Polygon_TestDriver(polylist)
  List	polylist;
{
    char	action;
    List	elt;
    int		i;
    int		num_polys;
    Polygon**	polys;
    Boolean	done;
    Vector	point;
    Boolean	result;

    printf("\nBegining Polygon_TestDriver...\n");
    num_polys = length(polylist);
    printf("Number of polygons is %d\n",num_polys);

    polys = ALLOCATE_ARRAY(Polygon*,num_polys);

    for (i=0, elt=polylist; (elt);
	 i++,elt = NXT(elt)) {
	polys[i] = GET(elt,Polygon);
    }

    printf("Current commands: [q,P,p,i]\n");

    done = FALSE;
    while (!done) {
	printf("\nCommand: ");
	if ((scanf(" %c",&action))!=1) {
	    action = '?';
	}

	switch(action) {
	  case '?':	/* special symbol for unrecognized command */
	    printf("Error reading command...\n");
	    break;

	  case 'q':	/* Quit */
	    printf("Quitting...\n");
	    done = TRUE;
	    break;

	  case 'P':	/* Print All */
	    for (i=0; i<num_polys; i++) {
		printf("Polygon(%d)\n", i);
		Polygon_Print(stdout,polys[i],NULL);
		printf("\n");
	    }
	    break;

	  case 'p':	/* Print one */
	    i = Get_InputPolyID(num_polys);
	    printf("Polygon(%d)\n", i);
	    Polygon_Print(stdout,polys[i],NULL);
	    printf("\n");
	    break;

	  case 'i':	/* interior test */
	    i = Get_InputPolyID(num_polys);
	    point = Get_InputPoint2();
	    result = Polygon_Interior(polys[i],point);
	    printf("Point is%s inside polygon(%d)\n",
		   (result ? "" : " not"), i);
	    break;

	  default:
	    printf("unrecognized command: '%c'\n",action);
	}
	
    }
}


#endif
