/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * polygon.h
 *
 * Type defintion for polygons
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/

/******************************************************************
Here is the typedef for describing a spherical polygon in full
generality. (i.e. convex, non-convex, simple, non-simple,
full-dimensional, degenerate, holes, separate components)
*******************************************************************/

#include <stdio.h>

#ifndef _POLYGON
#define  _POLYGON

#include "standards.h"

typedef struct PolygonSeg
{
    Vector	*begin;		/* first endpoint */
    Vector  	*edge;		/* "directed" edge */
    Vector  	*end;		/* second endpoint */
} PolygonSeg;

typedef struct PolygonBoundary
{
    List	sides;       	/* List of (Polygon Seg *)  */
    Boolean	interior;	/* is interior included? */
} PolygonBoundary;		/* "simple" polygon */

typedef struct Polygon
{
    List	boundaries;	/* List of (PolygonBoundary *) */
    void*	data;		/* this structure type is up to
				   the application */
} Polygon;


/*----------------------------------------------------------------*/
/*--- useful structure for polygon/boolean combination         ---*/
/*----------------------------------------------------------------*/
typedef struct PolySource {
    Polygon	*poly;		/* polygon source */
    Boolean	flipped;	/* was polygon flipped */
} PolySource;

/*----------------------------------------------------------------*/
/*--- this can replace a lists of PolySources, with the        ---*/
/*--- advantage that a list can be expanded in two different   ---*/
/*--- lists without having to use twice the space.             ---*/
/*----------------------------------------------------------------*/
typedef struct PolyTree {
    PolySource			*source;
    struct PolyTree		*next;
} PolyTree;



/******************************************************************
                      Some Useful Functions
*******************************************************************/

typedef void (*Polygon_Data_Printer)(FILE*,void*);
typedef void* (*Polygon_Data_Reader)(FILE*);
typedef int (*Polygon_Data_Freer)(void*);


void Polygon_Print(FILE*, Polygon*, Polygon_Data_Printer );
Polygon* Polygon_Read(FILE*, Polygon_Data_Reader );
void Polygon_Free(Polygon*,Polygon_Data_Freer);


/*----------------------------------------------------------*/
/*--- some useful routines of the above type.            ---*/
/*--- must cast this properly to send them as parameters ---*/
/*----------------------------------------------------------*/
void TwoParts_NicePrint(FILE*,void*);
void TwoParts_Print(FILE*,void*);
void* TwoParts_Read(FILE*);
int TwoParts_Free(void*);


/*************************************************************
    Function:  Polygon_Interior(poly,point)
    Input:     poly -- a spherical polygon
               point -- a point on the sphere
    Output:    TRUE -- point is in the interior of poly
               FALSE -- point is not in the interior of poly
    Effects:   <none>

    Description:  This routine is meant to test whether a point
      is in the *interior* of a polygon.  It is somewhat unpredictable
      for points lying exactly on the boundary.  This is due
      somewhat in part to the inherent floating point error.

      This routine does not even bother to check any polygon
      boundary with interior=FALSE, rather it simply ignores it.

**************************************************************/
Boolean Polygon_Interior(Polygon*, struct Vector);


#ifdef DEBUG
void Polygon_TestDriver(List);
#endif   /* DEBUG */


#endif   /* _POLYGON_TYPES */



/******************************************************************

        Interpretation
        --------------

To handle full generality, a "polygon" may require several components
to describe.  Standard convex polygons such as the ones we encounter
in the local motion will ALWAYS have a single component. (Exception:
the case of the "line" which has two antipodal points and no edges)

However to describe a square with a square hole, it requires giving
two oriented boundaries.  The external square oriented
counter-clockwise, says that locally, the left side of that curve is
the interior and the right side is the exterior, and a smaller square
oriented clockwise which says, again that the left side of that curve
is locally the interior, and the right side is the exterior.

At this point in time, all boundaries given should be disjoint,
however maybe this assumption is not necessary for the implementation.

The exact rule for determining if a point is inside or outside a
Polygon is by looking at the first edge crossed when walking in an
arbitrary direction.


  "data"
  This is an opaque structure that can be used as needed by the
  application using polygons.  However for polygons arising from
  assembly planning, this will be a TwoParts structure, having
  two fields, 'part1' and 'part2'.

  The interpretation in this case is that (part1) is blocked by
  (part2) over the entire sphere of directions.  However, the
  polygonal region described is considered the EXCEPTION to this
  blocking relation.

  The polygon (polygonal chain) described is considered to be an area
  where the blocking relationship is RELAXED.  That is, the group of
  directions where (part1) is free to move without being blocked by
  (part2).

  Also, the underlying assumption is that any direction in which
  (part1) is relaxed w.r.t. (part2), then in the opposite direction,
  the converse is true.  Therefore, a single polygon represents the
  full interaction between the pair of parts.

"sides" --
  The side list is an *ordered* and *oriented* list of all segments of
  the polygonal chain.  Each item gives pointers to the beginning
  vertex, the ending vertex, and the edge that connects them.  It is
  ASSUMED that the edge given really passes through both endpoints.
  And is is required that the end vertex of each segment is identical
  to the begin vertex of the next Segment.  To maintain this, it is 
  suggested that all neighboring Segments that share a common vertex
  point to a single Vector, rather than two copies of the same data in
  different locations.

  Notice that this condition must also be true for closed polygons, in
  that the end vertex for the final segment should match the start
  vertex of the first segment.

  The order of segments implicitly gives a counterclockwise
  orientation to each PolygonBoundary.  That is, the "inside" face is
  considered to be locally the one to the left of the segments.

  Each segment has three fields, begin, end, and edge, which represent
  the two vertices and the edge.   The specifics for these fields are
  discussed next.

  There are two special cases that arises in the use of segment.
  First, when a full great circle or full hemisphere must be
  represented. We will use a single PolygonSeg with two NULL vertices,
  but a true (oriented) edge.

  Second, when a lone vertex must be represented, we will use a single
  PolygonSeg with a true `begin' vertex, but a NULL edge and `end'
  vertex.


"begin" -- 
  Each vertex is simply a directional vector.  Although usually
  the vector will be nomalized, it is not explicitly required.

"end" -- 
  Each vertex is simply a directional vector.  Although usually
  the vector will be nomalized, it is not explicitly required.

"edge" -- 
  An edge is represented by the normal vector to the great circle
  connecting its two endpoints.  However notice that for any great
  circle, there are two choices for the normal.  This corresponds to
  the fact that for two points and a great circle, there are two
  different arcs connecting the points.

  The convention here is a right-hand rule.   That is, using your
  right hand, if your thumb points in the direction of the normal
  given, then the direction of travel on that edge is in the direction
  that your fingers curl.

  By this convention, for two non-antipodal points, the *shorter* of
  the two arcs connecting the endpoints can be given as the result of
  (endpoint_1 X endpoint_2).  And the *longer* arc is just the
  opposite.

  In the special case of a full great circle, that is an edge with NO
  vertices, the convention is that the normal faces the interior face.

  It is REQUIRED that any edge given has a normal which is
  perpendicular to both its endpoints.

"interior"
  If interior = TRUE, then the relaxed blocking relationship is
  defined on both the polygonal edges, and on the face enclosed to the
  left side of the edges.  If interior = FALSE then the relaxed
  blocking relationship is defined ONLY on the polygonal edges, not on
  either face.

  SPECIAL CASE: As mentioned above, in the case of a hemisphere, which
  has no vertices, the interior face is defined to be the face in the
  direction of the normal specifying the edge.

*******************************************************************/


/******************************************************************

    	Relation to previous 'types' of polygons
    	----------------------------------------

Below, we explain thew many "cases" of polygons that will arise from
local freedom cones will be represented in this structure.

------------------------------------------------------------------
Case:  "normal" polygon

Data:
  Single PolygonBoundary:
    List of k segments, oriented counterclockwise
    interior == TRUE

Interpretation:
  The blocking relation implicitly defined is that (part1) is free to
translate with respect to (part2) for any directions on the boundary,
or inside the INTERIOR face, but that (part1) is blocked by (part2) for
all other directions.

Real World Example:
  A phone receiver resting on the phone.

------------------------------------------------------------------
Case:  FullLine

Data:
  Two PolygonBoundary's:  each of which -- 
    List of 1 segment, which has a begin vertex, but a NULL edge
    	    	    	and a NULL end vertex.
    interior == FALSE

Interpretation:
  The blocking relation implicitly defined is that (part1) is free to
translate with respect to (part2) for either of the two directions,
but that (part1) is blocked by (part2) for all other directions.

Real World Example:
  A peg in a hole that can slide out either end.
  
------------------------------------------------------------------
Case:  HalfLine

Data:
  Single PolygonBoundary:
    List of 1 segment, which has a begin vertex, but a NULL edge
    	    	    	and a NULL end vertex
    interior == FALSE

Interpretation:
  The blocking relationship implicitly defined is that (part1) is free
to translate with respect to (part2) in the single direction given,
but that (part1) is blocked by (part2) for all other directions.

Real World Example:
  A peg in a hole with a bottom.
 
------------------------------------------------------------------
Case: Hemisphere

Data:
  Single PolygonBoundary:
    List of 1 segment, which has two NULL vertices, but a true edge
    	(with the edge normal, oriented towards the interior)
    interior == TRUE

Interpretation:
  The blocking relationship implicitly defined is that (part1) is free
to translate with respect to (part2) while on the great circle or in
the hemisphere given, but that (part1) is blocked by (part2) for all
other directions.

Real World Example:
  A book sitting on your hand.  (where your hand is simply a plane)
 
------------------------------------------------------------------
Case:  Plane

Data:
  Single PolygonBoundary:
    List of 1 segment, which has two NULL vertices, but a true edge
    interior == FALSE

Interpretation:
  The blocking relationship implicitly defined is that (part1) is free
to translate with respect to (part2) while on the great circle, but
that (part1) is blocked by (part2) for all other directions in either
hemisphere.

Real World Example:
  A book held flat between two other books (which are considered to be
a single part)
 
------------------------------------------------------------------
Case:  HalfPlane

Data:
  Single PolygonBoundary:
    List of 1 segments,  (v1,e1,v2)   where (v2 == -v1)
    interior == FALSE

Interpretation:
  The blocking relationship implicitly defined is that (part1) is free
to translate with respect to (part2) while on the half-circle, but
that (part1) is blocked by (part2) for all other directions.

Real World Example:
  A book held flat between two other books, with all three standing on
a flat desk.  (where the two other books and the desk are again
treated as a single part)
 
------------------------------------------------------------------
Case:  PartialPlane

Data:
  Single PolygonBoundary:
    List of 1 segments, (v1,e1,v2)
    interior == FALSE

Interpretation:
  The blocking relation implicitly defined is that (part1) is free to
translate with respect to (part2) while on the smaller arc, but that
(part1) is blocked by (part2) for all other directions.

Real World Example:
  A book held between two books, all three standing on a desk AND
leaning against a wall.
 
------------------------------------------------------------------
Case:  Wedge

Data:
  Single PolygonBoundary:
    List of 2 segments, oriented counterclockwise
    	    	    	(v1,e1,-v1) & (-v1,e2,v1)
    Interior == TRUE

Interpretation:
  The blocking relation implicitly defined is that (part1) is free to
translate with respect to (part2) while on either edge or the interior
region, and that (part1) is blocked by (part2) for all other
directions.

Real World Example:
  Any pieces that share exactly two, non-parallel faces.  For instance
a single book standing on a desk and pushed against a wall.
 
------------------------------------------------------------------
Case:  FullSphere

Data:
  Single PolygonBoundary:
    List of 0 segments
    Interior == TRUE

Interpretation:
  The blocking relation implicitly defined is that (part1) is free to
move with respect to (part2) to in any direction.  That is the entire
sphere makes up an EXCEPTION where (part1) is not blocked by (part2).

Real World Example:
  Two pieces that are not in contact.
 
------------------------------------------------------------------
Case:  EmptySphere

Data:
  Single PolygonBoundary:
    List of 0 segments
    Interior == FALSE

Interpretation:
  The blocking relation implicitly defined is that (part1) is blocked
by (part2) in all directions.  There are no EXCEPTIONS to this
blocking relation.

Real World Example:
  Two pieces that cannot be separated by any translations.
 
------------------------------------------------------------------
*******************************************************************/
