/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * spheregeom.c
 *
 * Geometric primitives on the sphere
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/

#include <math.h>
#include "spheregeom.h"
#include "standards.h"


/*
  EPS_1 specifies the maximum angle, theta, between two
  vectors which are considered to be equal.
  The value used is (sin^2(theta)).
*/
/*#define EPS_1  .001*/
/* arbitrary change. Look at this value more carefully.  */
#define EPS_1  .0001



/*************************************************************
  Function: NormalizeVector(vec)
  Input:
      vec --- vector to be normalized
  Output:
      Vector --- normalized vector
**************************************************************/
Vector NormalizeVector(Vector vec)
{
    Vector result;
    double scale;

    result = vec;

    /** must renormalize **/
    scale = SIZE(result);
    if (scale == 0.0) {
	/*printf("NormalizeVector received zero vector.\n");*/
    } else {
	scale = 1/scale;
    }

    SCALE(scale,result);

    return(result);
}



/*************************************************************
    Function:  VectorTest(U,V,perp)
    Input:     vectors U,V  (assumed to be Unit length)
                  (may be point/point, edge/edge, point/edge)
               perp -- TRUE if checking for (U perpendicular V)
	               FALSE if checking for (U parallel V)

    Assumptions:  <none>
    Output:    TRUE if test passes within error bound
               FALSE if test does not pass within error bound

    Description:  we assume that two vectors are "parallel" only
            if they are pointing in the "same" direction (within
	    error bounds), but NOT if they are opposite.

**************************************************************/
Boolean VectorTest(Vector U,Vector V,Boolean perp)
{
    double	calc;
    Boolean	result;

    /* Cos^2(theta) */
    calc = DOT(U,V);
    if ((!perp) && (calc < 0.0)) {
	/*---- shortcut if vectors are opposite ----*/
	result = FALSE;
    } else {
	calc = calc*calc;

	/* 1-Cos^2(theta) == Sin^2(theta) */
	if (!perp) calc = 1.0 - calc;

	result = (calc <= EPS_1);
    }

    return(result);
}


/*************************************************************
  Function:  Closer(V1,E,V2,V3,unit)
  Input:   points V1,V2,V3, a directed edge E, a Boolean
  Assumptions:  All 3 points lie on edge E
  Output:  Starting from V1, travel in direction of E.
             TRUE  ==  V2 is closer (or equal) than V3
             FALSE ==  V2 is farther than V3
**************************************************************/
Boolean Closer(Vector V1,Vector E,Vector V2,
	       Vector V3,Boolean unit)
{
    Vector  temp;
    Boolean test2,test3;
    double  t;

    if (!unit) {
	t = sqrt(NORM2(V2) / NORM2(V3));	/* this is RIGHT! */
    }


    CROSS(temp,E,V1);
    test2 = DOT(V2,temp) >= 0.0;	/* V2 le 180deg from V1 */
    test3 = DOT(V3,temp) >= 0.0;	/* V3 le 180deg from V1 */


    return(
	   test2 ?
	   (test3 ?		/* (test2 && test3) */
	    (unit ?
	     (DOT(V1,V2) >= DOT(V1,V3)) :
	     (DOT(V1,V2) >= t*DOT(V1,V3))) :
	    TRUE) :		/* (test2 && !test3) */
	
	   (test3 ?
	    FALSE :		/* (!test2 && test3)  */
	    (unit ?		/* (!test2 && !test3) */
	     (DOT(V1,V2) <= DOT(V1,V3)) :
	     (DOT(V1,V2) <= t*DOT(V1,V3))))
	   );
}




/*************************************************************
  Function:  SmallerAngle(E1,,V,E2,flip2,E3,flip3,unit)
  Input:   directed edges E1,E2,E3, a point V, three Booleans
  Assumptions:  All 3 edges incedent to point V

  Assigns E2* = E2  if (flip2 == false)
              = NEG(E2)  if (flip2 == true)
  Simliarly for E3.

  Compares the angle between (E1,E2*)  and angle (E1,E3*)

  Output:  TRUE == Angle(E1,E2*) is smaller (or equal)
           FALSE == Angle(E1,E3*) is smaller
**************************************************************/
Boolean SmallerAngle(Vector E1, Vector V, Vector E2, Boolean flip2,
	       Vector E3, Boolean flip3, Boolean unit)
{
    double	t;
    Vector  	_E2,_E3;	/* local copy...safer than
				   negating parameter */

    /**
      note -- this is sort of a dual of the Closer routine.
      However, for now, this routine does NOT care about the
      clockwise/counterclockwise orientation, rather just the
      absolute angle.   Because of this difference, the
      parameter V is not used here, but is left as a parameter
      for possible future use.
    **/

    _E2 = E2;
    if (flip2) NEG(_E2);
    _E3 = E3;
    if (flip3) NEG(_E3);


    if (!unit) {
	t = sqrt(NORM2(_E2) / NORM2(_E3));	/* this is RIGHT! */
    }

    return (
	    (unit ?
	     (DOT(E1,_E2) >= DOT(E1,_E3)) :
	     (DOT(E1,_E2) >= t*DOT(E1,_E3)))
	    );
}


/*************************************************************
  Function:  Cross_Left(E1,E2,V,unit)
  Input:   directed edges E1, E2, point V, a Boolean
  Assumptions:  Both Edges incedent at point V
  Output:  Walking along E1, when we cross E2 at V,
           TRUE  ==  E2 goes left to right
           FALSE ==  E2 goes right to left
**************************************************************/
Boolean Cross_Left(Vector E1,Vector V,Vector E2, Boolean unit)
{
    Vector  temp;
    Boolean test;

    CROSS(temp,E1,V);

    return( DOT(E2,temp) >= 0.0);
}
