/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * spheregeom.h
 *
 * Geometric primitives on the sphere
 *
 * Author:  Michael Goldwasser
 * Modified:  Jan 1, 1995 --- Original release (Goldwasser)
 *******************************************************/

#ifndef _SPHEREGEOM
#define _SPHEREGEOM

#include "standards.h"
#include <math.h>


/*===============================================================*/
/*=								=*/
/*=                       M a c r o s				=*/
/*=								=*/
/*===============================================================*/

#define NORM2(A)   (DOT(A,A))
#define SIZE(A)    sqrt(NORM2(A))
#define NEG(A)     ((A).x = -(A).x, (A).y = -(A).y, (A).z = -(A).z)
#define DOT(A,B)  ((A).x*(B).x + (A).y*(B).y + (A).z*(B).z)
#define CROSS(C,A,B)  (C).x=(A).y*(B).z-(A).z*(B).y, \
                      (C).y=(A).z*(B).x-(A).x*(B).z, \
                      (C).z=(A).x*(B).y-(A).y*(B).x
#define SCALE(K,A)      (A).x*=(K), (A).y*=(K), (A).z*=(K)
#define ADD(C,A,B)      (C).x=(A).x+(B).x, \
                        (C).y=(A).y+(B).y, \
                        (C).z=(A).z+(B).z
#define SUB(C,A,B)      (C).x=(A).x-(B).x, \
                        (C).y=(A).y-(B).y, \
                        (C).z=(A).z-(B).z


#define ON_HORIZON_VEC(v) ((v).z == 0.0)
#define ON_POLE_VEC(v)    (((v).z == 0.0) && ((v).x == 0.0))

#define ON_HORIZON_NORM(e) (((e).x == 0.0) && ((e).y == 0.0))

#define ZERO_VEC(v) (((v).x==0.0) && ((v).y==0.0) && ((v).z==0.0))



/*************************************************************
  Function: NormalizeVector(vec)
  Input:
      vec --- vector to be normalized
  Output:
      Vector --- normalized vector
**************************************************************/
struct Vector NormalizeVector(struct Vector);



/*************************************************************
    Function:  VectorTest(U,V,perp)
    Input:     vectors U,V  (assumed to be Unit length)
                  (may be point/point, edge/edge, point/edge)
               perp -- TRUE if checking for (U perpendicular V)
	               FALSE if checking for (U parallel V)

    Assumptions:  <none>
    Output:    TRUE if test passes within error bound
               FALSE if test does not pass within error bound

    Description:  we assume that two vectors are "parallel" only
            if they are pointing in the "same" direction (within
	    error bounds), but NOT if they are opposite.

**************************************************************/
Boolean VectorTest(struct Vector, struct Vector, Boolean);
#define VectorSame(U,V) VectorTest(U,V,FALSE)
#define VectorPerp(U,V) VectorTest(U,V,TRUE)



/*************************************************************
  Function:  Closer(V1,E,V2,V3,unit)
  Input:   points V1,V2,V3, a directed edge E, a Boolean
  Assumptions:  All 3 points lie on edge E
  Output:  Starting from V1, travel in direction of E.
             TRUE  ==  V2 is closer (or equal) than V3
             FALSE ==  V2 is farther than V3
**************************************************************/
Boolean Closer(struct Vector,struct Vector,struct Vector,
	       struct Vector,Boolean);


/*************************************************************
  Function:  SmallerAngle(E1,,V,E2,flip2,E3,flip3,unit)
  Input:   directed edges E1,E2,E3, a point V, three Booleans
  Assumptions:  All 3 edges incedent to point V

  Assigns E2* = E2  if (flip2 == false)
              = NEG(E2)  if (flip2 == true)
  Simliarly for E3.

  Compares the angle between (E1,E2*)  and angle (E1,E3*)

  Output:  TRUE == Angle(E1,E2*) is smaller (or equal)
           FALSE == Angle(E1,E3*) is smaller
**************************************************************/
Boolean SmallerAngle(struct Vector, struct Vector, struct Vector,
		     Boolean, struct Vector, Boolean, Boolean);


/*************************************************************
  Function:  Cross_Left(E1,E2,V,unit)
  Input:   directed edges E1, E2, point V, a Boolean
  Assumptions:  Both Edges incedent at point V
  Output:  Walking along E1, when we cross E2 at V,
           TRUE  ==  E2 goes left to right
           FALSE ==  E2 goes right to left
**************************************************************/
Boolean Cross_Left(struct Vector,struct Vector,struct Vector,
		   Boolean);

#endif
