/*

Notice

This software is public domain. It may freely be copied and used in
non-commercial products, assuming proper credit is given to the author, but it
should not be resold.
If you want to use the software for commercial products, contact the author.

This software is provided ``as is'' without warranty of any kind, either
expressed or implied.
The entire risk as to the quality and performance of the software is with you.
Should the software prove defective, you assume the cost of all necessary
servicing, repair or correction.

Please report any bugs in this package.   Although there is no
guarantee, such bugs may be fixed in a future release of the software.

Please send any comments to wass@cs.stanford.edu (Michael Goldwasser)

*/


/********************************************************
 * traverse.c
 *
 * Sample application which traverses an arrangement
 *
 * Author:  Michael Goldwasser, Bruce Romney
 * Modified:  Jan 1, 1995 --- Original release (authors)
 *******************************************************/

#include "standards.h"
#include "arrange.h"
#include "arrange_geom.h"
#include "arrange_macros.h"
#include "arrange_type.h"
#include "lists.h"
#include "polygon.h"
#include "traverse.h"

/*--- To bypass supporting routines/typedefs, compile with    ---*/
/*---   STRIPPED_VERSION being defined by the Makefile.       ---*/

#ifndef STRIPPED_VERSION
#include "dbg.h"
#else
void describe_dbg() {};
  void dbg_add_constraint() {};
  void dbg_delete_constraint() {};
  void* dbg_init() {};
  delete_dbg() {};
#endif /* STRIPPED_VERSION */
  /*---------------------------------------------------------------*/
  
#ifdef DEBUG
  extern int tr_debug;	/* 0 = no debugging printf's or verifications */
  /* 1 = debugging verifications but nothing else */
  /* 2 = some debugging printf's as well */
  /* 3 = lots of debugging printf's */
  
#endif  /* DEBUG */
  
  /* Define the basic element of the traversal path, a "step" between */
  /* trapezoids.							    */
  typedef struct Traverse_Step
{
    Trap *from;
    Trap *toward;
    TrapEdge *edge;	/* edge between the traps, if applicable; else NIL */
    Trap	*face_rep;	/* representative trap for the FROM face */
} Traverse_Step;


/*=================================================================*/
/*=								  =*/
/*=	 Local structure for marking arrangment cells		  =*/
/*=								  =*/
/*=	 This will rely on the fact that verts/traps have a	  =*/
/*=	 unique ID from <1..num_verts>  (or <1..num_traps>)	  =*/
/*=								  =*/
/*=================================================================*/
#define MARKED(id,bool)    ((bool)[id] == TRUE)
#define MARK_IT(id,bool)   ((bool)[id] = TRUE)
#define UNMARK_IT(id,bool) ((bool)[id] = FALSE)

#define MARKED_V(v, marks)     MARKED(V_id(v)-1, (marks)->verts)
#define MARK_IT_V(v, marks)    MARK_IT(V_id(v)-1, (marks)->verts)
#define UNMARK_IT_V(v, marks)  UNMARK_IT(V_id(v)-1, (marks)->verts)

#define MARKED_T(t, marks)     MARKED(T_id(t)-1, (marks)->traps)
#define MARK_IT_T(t, marks)    MARK_IT(T_id(t)-1, (marks)->traps)
#define UNMARK_IT_T(t, marks)  UNMARK_IT(T_id(t)-1, (marks)->traps)


typedef struct Traverse_Marks
{
    Boolean	*verts;
    Boolean	*traps;
} Traverse_Marks;


#ifdef DEBUG
/*===============================================================*/
/*=								=*/
/*=	 	Printing functions				=*/
/*=		 (for debugging)				=*/
/*=								=*/
/*===============================================================*/

char *show_cell(ArrangeCell *cell)
  /* (can only call this function once per line) */
{
    static char string[30];
    if (cell->t==vert)
	sprintf(string,"vert(%d)",V_id(cell->d.vert));
    
    else if (cell->t==edge)
	sprintf(string,"edge(%d)",E_id(cell->d.edge));
    else
	sprintf(string,"trap(%d)",T_id(cell->d.trap));
    return string;
}

char *show_intermediate_cell(Traverse_Step *ts)
  /* (can only call this function once per line) */
{
    static char string[30];
    if (ts->edge==NIL)
	sprintf(string,"thread");
    else
	sprintf(string,"edge(%d)",E_id(ts->edge));
    return string;
}

char *show_Traverse_Step(Traverse_Step *ts, Boolean forward)
  /* (can only call this function once per line) */
{
    static char string[30];
    if (forward)
	sprintf(string,"step (%d->%d via %s)", T_id(ts->from), T_id(ts->toward),
		show_intermediate_cell(ts));
    else
	sprintf(string,"step (%d->%d via %s)", T_id(ts->toward), T_id(ts->from),
		show_intermediate_cell(ts));
    
    return string;
}

print_polygons(PolyTree *pt0)
{
    PolyTree *pt;
    PolySource *ps;
    TwoParts *parts;
    int n=0;
    for (pt = pt0; (pt); pt = pt->next)
	{
	    ps = pt->source;
	    printf(" %c", (ps->flipped ? '-' : '+'));
	    TwoParts_NicePrint(stdout, ps->poly->data);
	    ++n;
	    /*	if (n%5==0) printf("\n"); */
	}
    if (n==0) printf("<none>");
    printf("\n");
}

print_polygons_at_vert(TrapVert *vert)
{
    printf("  Polygons at vert(%d):  ", V_id(vert));
    print_polygons(V_polys(vert));
}

print_polygons_at_edge(TrapEdge *edge)
{
    printf("    Polygons on edge(%d):       ", E_id(edge));
    print_polygons(E_polys_on(edge));
    printf("             on/above edge(%d): ", E_id(edge));
    print_polygons(E_polys_above(edge));
    printf("             on/below edge(%d): ", E_id(edge));
    print_polygons(E_polys_below(edge));
}

#endif  /* DEBUG */

/*===============================================================*/
/*=								=*/
/*=	 		Utility functions			=*/
/*=								=*/
/*===============================================================*/

Traverse_Step *make_Traverse_Step(from, toward, edge, rep)
  Trap		*from;
  Trap		*toward;
  TrapEdge	*edge;
  Trap		*rep;
{
    Traverse_Step *step;
    step = ALLOCATE(Traverse_Step);
    step->from = from;
    step->toward = toward;
    step->edge = edge;
    step->face_rep = rep;
    return step;
}

void clear_stack(List stack)
  /* Delete a list of Traverse_Step's, including the Traverse_Step's */
  /* themselves. */
{
    List elt;
    Traverse_Step *step;
    for (elt = stack; elt!=NIL; elt=NXT(elt))
	{
	    step = GET(elt, Traverse_Step);
	    FREE(step);
	}
    delete_list(stack);
}	



/*=================================================================*/
/*=								  =*/
/*=	 Functions to update DBGs				  =*/
/*=								  =*/
/*=================================================================*/

/*************************************************************
  Functions:  exit_polygon(dbg, ps, mirror),
  enter_polygon(dbg, ps, mirror)
  Input:     dbg --- a dbg
  ps  --- a PolySource
  mirror -- a Boolean
  Output:    <none>
  Effects:   Modify the DBG to reflect the fact that we are
  leaving/entering this polygon, or, if mirror==TRUE,
  its mirror.
  **************************************************************/
void exit_polygon(struct DBG *d, PolySource *ps, Boolean mirror)
{
    dbg_add_constraint(d, ps->poly->data, (ps->flipped != mirror));
}   

void enter_polygon(struct DBG *d, PolySource *ps, Boolean mirror)
{
    dbg_delete_constraint(d, ps->poly->data, (ps->flipped != mirror));
}   


/*************************************************************
  Functions:  exit_polygons(the_dbg, ptree, at_mirror),
  enter_polygons(the_dbg, ptree, at_mirror)
  
  Input:     the_dbg --- a valid dbg
  ptree --- the polytree for polygons we enter/exit
  at_mirror --- are we in reverse orientation?
  
  Output:    <none>
  Effects:   Updates the dbg to reflect entering or exiting the
  polygons in the polytree
  
  Description:  These routines will delete or add all the constraints
  for the polygons in 'ptree'.   This reflects the fact that
  we are entering/exiting a polygon and thus constraints get relaxed
  or un-relaxed.
  **************************************************************/
#define exit_polygons(d,p,m)   update_dbg_polytree(d,p,m,exit_polygon)
#define enter_polygons(d,p,m)  update_dbg_polytree(d,p,m,enter_polygon)


/*************************************************************
  Function:  update_dbg_polytree(the_dbg,ptree,at_mirror,Add_Delete)
  
  Input:     the_dbg --- a valid dbg
  ptree --- the ptree to be added/deleted
  at_mirror --- are we in reverse orientation?
  Add_Delete --- pointer to proper routine
  
  Output:    <none>
  Effects:   Updates the dbg by adding/deleting the polytree
  
  Description:  This routine is the 'basis' for exit_polygons()
  and enter_polygons().  It traverses the polytree, adding
  or deleting the correct edges from the DBG to account for
  exiting/entering the given polytree.  Notice that it also takes a
  parameter 'at_mirror' which tells whether or not the orientation of
  the dbg is the opposite of the orientation of the polytree.
  **************************************************************/
update_dbg_polytree(the_dbg, ptree, at_mirror, Add_Delete)
  struct DBG	*the_dbg;
  PolyTree	*ptree;
  Boolean	at_mirror;
  void	(*Add_Delete)();
{
    PolyTree	*walk_tree;
    TwoParts	*tp;
    
#ifdef DEBUG
    if ((ptree!=NIL) && (tr_debug>2))
	{
	    if (Add_Delete == (void *) exit_polygon)        printf("    Exitting");
	    else if (Add_Delete== (void *) enter_polygon)   printf("    Entering");
	    
	    if (at_mirror) printf(" mirrors");
	    printf(": ");
	}
#endif
    
    for (walk_tree = ptree;  walk_tree; walk_tree=walk_tree->next)
	{
	    Add_Delete(the_dbg, walk_tree->source, at_mirror);
	    
#ifdef DEBUG
	    if (tr_debug>2)
		{
		    printf(" %c", (walk_tree->source->flipped ? '-' : '+'));
		    TwoParts_NicePrint(stdout, walk_tree->source->poly->data);
		}
#endif
	}
    
#ifdef DEBUG
    if ((ptree!=NIL) && (tr_debug>2)) printf("\n");
#endif
}



/*************************************************************
  Function:  update_dbg_onto_edge(the_dbg, the_edge, from_above, flipped, on)
  Input:     the_dbg --- a dbg valid for before walking onto edge
  the_edge --- the edge to walk onto
  from_above --- coming from above to below?
  flipped --- is orientation flipped?
  on --- if FALSE, then skip the polygons defined
  only on (not above/below) the edge
  Output:    <none>
  Effects:   Updates the dbg to account for walking onto the edge
  
  Description:  This routine will delete the necessary relaxed
  constraints from the DBG to account for walking onto the edge.
  
  If the flipped flag is TRUE, then this routine assumes that all
  the polygons stored with the edge should be viewed reversed.
  This aids in dealing with mirror edges.
  
  If on is FALSE, then presumably we are in the process of crossing
  the edge, so we can skip the polygons defined only *on* the edge.
  
  If the edge has a mirror, this routine will ALSO update the
  DBG to reflect walking onto the mirror edge.
  **************************************************************/
update_dbg_onto_edge(the_dbg, the_edge, from_above, flipped, on)
  struct DBG	*the_dbg;
  TrapEdge	*the_edge;
  Boolean	from_above;
  Boolean	flipped;
  Boolean	on;
{
    
#ifdef DEBUG
    if (tr_debug>2)
	{
	    printf("  Moving onto edge(%d)\n",E_id(the_edge));
	    printf("    Previous point was %s this edge.\n",
		   (from_above ? "ABOVE" : "BELOW"));
	    print_polygons_at_edge(the_edge);
	    if (E_mirror(the_edge)!=NIL)
		print_polygons_at_edge(E_mirror(the_edge));
	}
#endif
    
    /*--- delete the constraints for the relaxed polygons we enter ---*/
    if (on) enter_polygons(the_dbg, E_polys_on(the_edge), flipped);
    enter_polygons(the_dbg, (from_above ?
			     E_polys_below(the_edge) :
			     E_polys_above(the_edge)),
		   flipped);
    
    if (E_mirror(the_edge)) {
	/*--- delete the constraints for the relaxed polygons we enter ---*/
	if (on) enter_polygons(the_dbg, E_polys_on(E_mirror(the_edge)),!flipped);
	enter_polygons(the_dbg,(from_above ?
				E_polys_below(E_mirror(the_edge)) :
				E_polys_above(E_mirror(the_edge))),
		       !flipped);
	
	/* (Why '!flipped'?  Because in the normal case (flipped==FALSE),  */
	/*    you want to enter the reversed polygons of the mirror edge.) */
    }
}



/*************************************************************
  Function:  update_dbg_off_of_edge(the_dbg,the_edge,to_below,flipped,on)
  Input:     the_dbg --- a dbg valid for being on the edge
  the_edge --- the edge to walk onto
  to_below --- coming from above to below?
  flipped --- is orientation flipped?
  on --- if FALSE, then skip the polygons defined
  only on (not above/below) the edge
  Output:    <none>
  Effects:   Updates the dbg to account for walking off of the edge
  
  Description:  This routine will add the necessary constraints
  to the DBG to account for walking off an edge and leaving some
  relaxed polygons.
  
  If the flipped flag is TRUE, then this routine assumes that all
  the polygons stored with the edge should be viewed reversed.
  This aids in dealing with mirror edges.
  
  If on is FALSE, then presumably we are in the process of crossing
  the edge, so we can skip the polygons defined only *on* the edge.
  
  If the edge has a mirror, this routine will ALSO update the
  DBG to reflect walking off of the mirror edge.
  **************************************************************/
update_dbg_off_of_edge(the_dbg, the_edge, to_below, flipped, on)
  struct DBG	*the_dbg;
  TrapEdge	*the_edge;
  Boolean	to_below;
  Boolean	flipped;
  Boolean	on;
{
    
#ifdef DEBUG
    if (tr_debug>2)
	{
	    printf("  Moving off of edge(%d)\n",E_id(the_edge));
	    printf("    Specifically, to point %s.\n",
		   show_vector(dbg_inq_vector(the_dbg)));
	    printf("    New point is %s this edge.\n",
		   ((!to_below) ? "ABOVE" : "BELOW"));
	    print_polygons_at_edge(the_edge);
	    if (E_mirror(the_edge)!=NIL)
		print_polygons_at_edge(E_mirror(the_edge));
	}
#endif
    
    /*--- add the constraints for the relaxed polygons we exit ---*/
    if (on) exit_polygons(the_dbg, E_polys_on(the_edge), flipped);
    exit_polygons(the_dbg, (to_below ?
			    E_polys_above(the_edge) :
			    E_polys_below(the_edge)),
		  flipped);
    
    if (E_mirror(the_edge)) {
	/*--- add the constraints for the relaxed polygons we exit ---*/
	if (on) exit_polygons(the_dbg, E_polys_on(E_mirror(the_edge)), !flipped);
	exit_polygons(the_dbg,(to_below ?
			       E_polys_above(E_mirror(the_edge)) :
			       E_polys_below(E_mirror(the_edge))),
		      !flipped);
	
	/* (Why '!flipped'?  Because in the normal case (flipped==FALSE),  */
	/*    you want to enter the reversed polygons of the mirror edge.) */
    }
}



/*************************************************************
  Function:  update_dbg_traverse_step(the_dbg,the_A,step,forward)
  Input:     the_dbg --- a dbg valid for before the update
  the_A --- an arrangement
  step --- a Traverse_Step
  forward ---  taking step forward?  (else backtracking)
  Output:    <none>
  Effects:   Updates the dbg to account for the traversal step
  
  Description:  This routine executes either a forward step or
  a backtracking step across the traversal step given.  This
  routine calls the necessary routines to change the DBG for
  crossing the edge involved.
  **************************************************************/
update_dbg_traverse_step(the_dbg, the_A, step, forward)
  struct DBG		*the_dbg;
  Arrange		*the_A;
  Traverse_Step	*step;
  Boolean		forward;
{
    Boolean from_above;
    
#ifdef DEBUG
    if (tr_debug>1)
	printf("TRAVERSE %s %s\n",
	       (forward ? "crossing" : "backtracking through"),
	       show_Traverse_Step(step,forward));
#endif
    
    /*--- Do we cross a real edge?  If not (just a thread), then ---*/
    /*--- (step->edge) will be FALSE.                            ---*/
    if (step->edge) {
	from_above = (forward == (step->edge == T_be(step->from)));
	
	update_dbg_onto_edge(the_dbg, step->edge, from_above, FALSE, FALSE);
	update_dbg_off_of_edge(the_dbg, step->edge, from_above, FALSE, FALSE);
    }
    
}


/*************************************************************
  Function:  update_dbg_trap_to_vert(the_dbg,the_A,the_vert,the_trap)
  Input:     the_dbg --- a dbg valid for before the update
  the_A --- an arrangement
  the_vert --- the vert in question
  the_trap --- the specific trap incident to the vert
  Output:    <none>
  Effects:   Updates the dbg to account for walking from trap onto vert
  
  Description:  This routine will delete the necessary
  constraints from the DBG to account for walking onto the
  vert from the trap.
  
  This routine assumes that the initial DBG is valid for the trap,
  and it deletes all constraints to account for moving onto the
  vert.
  
  This routine will handle vertices lying on the horizon, so long
  as the trap is incident to the vert. (as opposed to the mirror)
  **************************************************************/
#define update_dbg_trap_to_vert(d,A,v,t) update_dbg_near_vert(d,A,v,t,TRUE)


/*************************************************************
  Function:  update_dbg_vert_to_trap(the_dbg,the_A,the_vert,the_trap)
  Input:     the_dbg --- a dbg valid for before the update
  the_A --- an arrangement
  the_vert --- the vert in question
  the_trap --- the specific trap incident to the vert
  Output:    <none>
  Effects:   Updates the dbg to account for walking off vert to trap
  
  Description:  This routine will add the necessary
  constraints to the DBG to account for walking off of the
  vert onto the trap.
  
  This routine assumes that the initial DBG is valid for the vert,
  and it adds all constraints to account for moving off the vert
  in the direction of the trap.
  
  This routine will handle vertices lying on the horizon, so long
  as the trap is incident to the vert. (as opposed to the mirror)
  **************************************************************/
#define update_dbg_vert_to_trap(d,A,v,t) update_dbg_near_vert(d,A,v,t,FALSE)


/*************************************************************
  Function:  update_dbg_near_vert(the_dbg,the_A,the_vert,ref_trap,
  move_onto)
  Input:     the_dbg --- a dbg valid for before the update
  the_A --- an arrangement
  the_vert --- the vert in question
  ref_trap --- the specific trap incident to the vert
  move_onto --- moving onto the vert?  (else moving off)
  Output:    <none>
  Effects:   Updates the dbg to account for walking on/off the vert
  
  Description:  This routine will add/delete the necessary
  constraints to/from the DBG to account for walking on/off the
  vert from/to the ref_trap.
  
  If move_onto is TRUE, then this routine assumes that the initial
  DBG is valid for the ref_trap, and it deletes all constraints to
  account for moving onto the vert.
  
  If move_onto is FALSE, then this routine assumes that the initial
  DBG is valid on the vert, and it adds all the constraints to
  account for moving off of the vert into the ref_trap.
  
  This routine will handle vertices lying on the horizon, so long
  as the ref_trap is incident to the vert. (as opposed to the mirror)
  **************************************************************/
update_dbg_near_vert(the_dbg, the_A, the_vert, ref_trap, move_onto)
  struct DBG	*the_dbg;
  Arrange	*the_A;
  TrapVert	*the_vert;
  Trap		*ref_trap;
  Boolean	move_onto;
{
    TrapVert	*temp_vert;
    Trap	*cur_trap;
    Trap	*next_trap;
    TrapEdge	*edge_to_cross;
    Boolean	at_mirror;
    
    
    /*--- first take care of relaxations ON the vert ----*/
    if (move_onto) {
	enter_polygons(the_dbg, V_polys(the_vert), FALSE);
	if (V_mirror(the_vert))
	    enter_polygons(the_dbg, V_polys(V_mirror(the_vert)), TRUE);
    } else {
	exit_polygons(the_dbg, V_polys(the_vert), FALSE);
	if (V_mirror(the_vert))
	    exit_polygons(the_dbg, V_polys(V_mirror(the_vert)), TRUE);
    }
    
    
    
    /*--- now take care of relaxations from incident edges ----*/
    temp_vert = the_vert;
    cur_trap = ref_trap;
    next_trap = NIL;
    at_mirror = FALSE;
    while ((next_trap != ref_trap) || (at_mirror)) {
	
	/*--- advance to next trap in a counterclockwise order ---*/
	Arrange_vert_nexttrap(the_A, temp_vert, cur_trap, at_mirror,
			      &next_trap, &edge_to_cross);
	
	if (edge_to_cross) {
	    /*--- crossing an edge ---*/
	    update_dbg_onto_edge(the_dbg, edge_to_cross,
				 (edge_to_cross == T_be(cur_trap)),
				 at_mirror, TRUE);
	    /*--- move to otherside of horizon if necessary... ---*/	  
	    if (!next_trap) {
		/*--- switch to opposite side ---*/
		at_mirror = (!at_mirror);
		temp_vert = V_mirror(temp_vert);
		edge_to_cross = E_mirror(edge_to_cross);
		next_trap =
		    ((temp_vert == E_rvert(edge_to_cross)) ?
		     ((E_trTrap(edge_to_cross)) ?
		      E_trTrap(edge_to_cross) : E_brTrap(edge_to_cross)) :
		     ((E_tlTrap(edge_to_cross)) ?
		      E_tlTrap(edge_to_cross) : E_blTrap(edge_to_cross)));
	    }
	    
	    /*--- now move all the way across the edge ---*/
	    if (!move_onto)
		update_dbg_off_of_edge(the_dbg, edge_to_cross,
				       (edge_to_cross != T_be(next_trap)),
				       at_mirror, TRUE);
	}
	
	/*--- setup cur_trap for next pass ---*/
	cur_trap = next_trap;
    }

#ifdef DEBUG
    if (tr_debug>2)
	{
	    if (move_onto) printf("dbg at vert...\n");
	    else printf("dbg at trap...\n");
	    describe_dbg(the_dbg);
	}
#endif /* DEBUG */

}


/*=================================================================*/
/*=								  =*/
/*=	 Functions to generate from scratch the DBG		  =*/
/*=	     corresponding to a given NDBG cell			  =*/
/*=								  =*/
/*=================================================================*/

/*************************************************************
  Function:  create_dbg_at_vert(assem, the_A, polylist, vert)
  Input:     assem --- a valid assembly
  the_A --- an arrangement
  polylist --- a list of polygons in the arrangment
  vert  ---  the TrapVert to be condidered
  Output:    pointer to a newly created dbg for that vert
  Effects:   <none>
  
  Description:  This routine creates a brand new DBG which has
  the blocking relationships present on a vertex.
  
  **************************************************************/
struct DBG* create_dbg_at_vert(assem, the_A, polylist, the_vert)
  Assembly	*assem;
  Arrange	*the_A;
  List		polylist;
  TrapVert	*the_vert;
{
    Vector	dir,rev_dir;
    struct DBG	*new_dbg;
    List	elt;
    Polygon*	poly;
    TwoParts	*tp;
    
    dir = InqVertVector(V_direction(the_vert));
    rev_dir.x = -(dir.x);
    rev_dir.y = -(dir.y);
    rev_dir.z = -(dir.z);
    
    new_dbg = dbg_init(assem, &dir);
    
    /*--- check all polygons for neighborhood around vert ---*/
    for (elt = polylist; elt; elt=NXT(elt)) {
 	poly = GET(elt,Polygon);
 	if (!Polygon_Interior(poly,dir)) {
	    /*--- if NOT in the polygon, then add the constraint ---*/
	    tp = (TwoParts*) poly->data;
	    dbg_add_constraint(new_dbg, tp, FALSE);
 	}
 	if (!Polygon_Interior(poly,rev_dir)) {
	    /*--- if NOT in the rev polygon, then add the constraint ---*/
	    tp = (TwoParts*) poly->data;
	    dbg_add_constraint(new_dbg, tp, TRUE);
 	}
    }
    
    /*--- complete vert's DBG by moving onto vert from neighborhood ---*/
    update_dbg_trap_to_vert(new_dbg, the_A, the_vert,
			    (V_tlTrap(the_vert) ? V_tlTrap(the_vert) :
			     V_brTrap(the_vert)));
    
    return(new_dbg);
}



/*************************************************************
  Function:  create_dbg_at_cell(assem, the_A, polylist, cell)
  Input:     assem --- a valid assembly
  the_A --- an arrangement
  polylist --- a list of polygons in the arrangment
  cell  ---  the ArrangeCell to be condidered
  Output:    pointer to a newly created dbg for that cell
  Effects:   <none>
  
  Description:  This routine creates a brand new DBG which has
  the blocking relationships present at a representative point
  in the cell of the arrangment given.
  
  Note that this routine pretends there is a point in the
  'interior' of any cell, whether the cell is degenerate or
  not.   That is, if given a trap, then this will create the
  blocking graph that is true for any point in the trap, and
  it will NOT include any blocking relationships that are
  only true on the boundary edges/verts.
  
  Similarly for an edge, this will only include blocking
  relationships that are true in the interior of the edge,
  and so it will NOT include any blocking relationsips that
  are only true on the endpoints.
  
  **************************************************************/
struct DBG* create_dbg_at_cell(assem, the_A, polylist, cell)
  Assembly	*assem;
  Arrange	*the_A;
  List		polylist;
  ArrangeCell	cell;
{
    TrapVert	*the_vert;
    struct DBG	*new_dbg;
    
    Boolean	on_vert;
    Boolean	right_of_vert;
    Trap	*start_trap;
    
#ifdef DEBUG
    if (tr_debug>1)
	printf("Building DBG from scratch for %s\n", show_cell(&cell));
#endif
    
    switch(cell.t) {
      case vert:
	the_vert = cell.d.vert;
	on_vert = TRUE;
	if (V_tlTrap(the_vert)) {
	    start_trap = V_tlTrap(the_vert);
	    right_of_vert = FALSE;
	} else {
	    start_trap = V_brTrap(the_vert);
	    right_of_vert = TRUE;
	}
	break;
	
      case edge:
	the_vert = E_lvert(cell.d.edge);
	start_trap =
	    (E_tlTrap(cell.d.edge) ?
	     E_tlTrap(cell.d.edge) : E_blTrap(cell.d.edge));
	on_vert = FALSE;
	right_of_vert = TRUE;
	break;
	
      case trap:
	the_vert = T_lvert(cell.d.trap);
	start_trap = cell.d.trap;
	on_vert = FALSE;
	right_of_vert = TRUE;
	break;
    }
    
    
    new_dbg = create_dbg_at_vert(assem, the_A, polylist, the_vert);
    
    /*--- now we have the correct DBG for the_vert ---*/
    
    
    /*--- if on edge/trap, then move to DBG for start_trap ---*/
    if (cell.t != vert) {
	update_dbg_vert_to_trap(new_dbg, the_A, the_vert, start_trap);
    }
    
    /*--- if on edge, then move to DBG for the edge ---*/
    if (cell.t == edge) {
	update_dbg_onto_edge(new_dbg, cell.d.edge,
			     (cell.d.edge == T_be(start_trap)), FALSE, TRUE);
    }
    
#ifdef DEBUG
    if (tr_debug > 2) printf("done building DBG from scratch\n");
#endif
    
    return(new_dbg);
}



/*************************************************************
  Function:  create_first_dbg(assem, the_A, polylist, the_trap)
  Input:     assem --- a valid assembly
  the_A --- an arrangement
  polylist --- a list of polygons in the arrangment
  the_trap --- pointer to pointer to trap to use as basis
  (if NULL, procedure sets this variable
  correctly on return)
  Output:    pointer to a newly created dbg for that cell
  Effects:   <none>
  
  Description:  This routine creates a brand new DBG which has
  the blocking relationships present at a representative point
  in the_trap.
  
  If (*the_trap==NULL) then this routine is free to pick the
  trapezoid to start at, however if it is set to a trap,
  then that MUST be the trap used to start the dbg.
  
  **************************************************************/
struct DBG*	create_first_dbg(assem, the_A, polylist, the_trap)
  Assembly	*assem;
  Arrange	*the_A;
  List		polylist;
  Trap		**the_trap;	/* this is set before returning */
{
    struct DBG	*new_dbg;
    ArrangeCell	cell;
    
    if ((*the_trap)==NIL)
	(*the_trap) = V_tlTrap(the_A->tpole);
    
    cell.t = trap;
    cell.d.trap = (*the_trap);
    
    new_dbg = create_dbg_at_cell(assem, the_A, polylist, cell);
    
#ifdef DEBUG
    if (tr_debug>0) verify_dbg_trap(assem, new_dbg, *the_trap, the_A);
#endif
    
    return(new_dbg);
}



/*=================================================================*/
/*=								  =*/
/*=	 Functions to find a path across the NDBG		  =*/
/*=								  =*/
/*=================================================================*/

/*************************************************************
  Function:  reset_marks(the_A, marks)
  Input:     the_A --- an arrangement
  marks --- a pointer Traverse_Marks structure
  Output:    <none>
  Effects:   Sets up 'marks' based on the number of verts/traps
  in the arrangement
  
  Description:  This routine allocates two Boolean arrays to
  be used to mark visisted verts/traps.
  **************************************************************/
void reset_marks(the_A,marks)
  Arrange		*the_A;
  Traverse_Marks	*marks;
{
    TrapVert	*v;
    Trap	*t;
    int		id;
    
    marks->verts = ALLOCATE_ARRAY(Boolean, the_A->num_verts);
    marks->traps = ALLOCATE_ARRAY(Boolean, the_A->num_traps);
    
    for (id = 0; id < the_A->num_verts; id++)
	UNMARK_IT(id,marks->verts);
    for (id = 0; id < the_A->num_traps; id++)
	UNMARK_IT(id,marks->traps);
    
}


/*************************************************************
  Function:  free_marks(marks)
  Input:     marks --- a pointer Traverse_Marks structure
  Output:    <none>
  Effects:   frees up the dynamic arrays inside the 'marks'
  structure.  (however does not free the 'marks'
  structure itself.
  
  **************************************************************/
void free_marks(marks)
  Traverse_Marks	*marks;
{
    FREE(marks->verts);
    FREE(marks->traps);
    
    marks->verts = marks->traps = NULL;
}



/*************************************************************
  Function:  trap_push_unmarked_neighbors_T(the_A,marks,t,rep,
  find_all,previous)
  Input:     the_A --- an arrangement
  marks --- a pointer to Traverse_Marks structure
  t --- the current trap
  rep --- representative trap for the current face
  find_all --- continue even after finding one?
  previous --- previous list of Traverse_Step objects
  Output:    expanded List of Traverse_Step objects
  Effects:   <none>
  
  Description:  Finds all unvisited neighbors of trap 't' which
  can be reached by passing through a thread, and pushes them
  onto the front of a running list.
  
  **************************************************************/
List trap_push_unmarked_neighbors_T(the_A, marks, t, rep, find_all, previous)
  Arrange		*the_A;
  Traverse_Marks	*marks;
  Trap			*t;
  Trap			*rep;
  Boolean		find_all;
  List			previous;
{
    List	answer;
    answer = previous;
    
    while (TRUE) {	/* break out when done --- not a real loop */
	/*--- try to walk through left thread ---*/
	if (T_tln(t)) {
	    if (!MARKED_T(T_tln(t),marks))
		{
		    /*--- found one ---*/
		    answer = push(answer, make_Traverse_Step(t,T_tln(t),NULL,rep));
		    if (!find_all) break;
		}
	    else if (!MARKED_T(T_bln(t),marks))
		{
		    /*--- found one ---*/
		    answer = push(answer, make_Traverse_Step(t,T_bln(t),NULL,rep));
		    if (!find_all) break;
		}
	}
	
	/*--- try to walk through right thread ---*/
	if (T_trn(t)) {
	    if (!MARKED_T(T_trn(t),marks))
		{
		    /*--- found one ---*/
		    answer = push(answer, make_Traverse_Step(t,T_trn(t),NULL,rep));
		    if (!find_all) break;
		}
	    else if (!MARKED_T(T_brn(t),marks))
		{
		    /*--- found one ---*/
		    answer = push(answer, make_Traverse_Step(t,T_brn(t),NULL,rep));
		    if (!find_all) break;
		}
	}
	
	/*--- else...no unmarked neighbor ---*/
	break;
    }
    
    return(answer);
}


/*************************************************************
  Function:  trap_push_unmarked_neighbors_E(the_A,marks,t,rep,
  find_all,previous)
  Input:     the_A --- an arrangement
  marks --- a pointer to Traverse_Marks structure
  t --- the current trap
  rep --- representative trap for the current face
  find_all --- continue even after finding one?
  previous --- previous list of Traverse_Step objects
  Output:    expanded List of Traverse_Step objects
  Effects:   <none>
  
  Description:  If the current trap has no left thread, then this
  will find all unvisited neighbors which can be found by crossing
  the left side of an edge.  Similarly, when no right thread exists,
  this will look for unvisited neigbors through the right side
  of the edges.
  All such neighbors are pushed onto the front of a running list.
  
  **************************************************************/
List trap_push_unmarked_neighbors_E(the_A, marks, t, rep, find_all, previous)
  Arrange		*the_A;
  Traverse_Marks	*marks;
  Trap		*t;
  Trap		*rep;
  Boolean		find_all;
  List		previous;
{
    Trap	*temp;
    List	answer;
    answer = previous;
    
    while (TRUE) {	/* break out when done --- not a real loop */
	/*--- no left thread...try to cross over an edge ---*/
	if (!T_tln(t)) {
	    if ((temp = E_tlTrap(T_te(t))) && (!MARKED_T(temp,marks)))
		{
		    answer = push(answer, make_Traverse_Step(t,temp,T_te(t),rep));
		    if (!find_all) break;
		}
	    if ((temp = E_blTrap(T_be(t))) && (!MARKED_T(temp,marks)))
		{
		    answer = push(answer, make_Traverse_Step(t,temp,t->be,rep));
		    if (!find_all) break;
		}
	}
	
	/*--- no right thread...try to cross over an edge ---*/
	if (!T_trn(t)) {
	    if ((temp = E_trTrap(T_te(t))) && (!MARKED_T(temp,marks)))
		{
		    answer = push(answer, make_Traverse_Step(t,temp,t->te,rep));
		    if (!find_all) break;
		}
	    if ((temp = E_brTrap(T_be(t))) && (!MARKED_T(temp,marks)))
		{
		    answer = push(answer, make_Traverse_Step(t,temp,t->be,rep));
		    if (!find_all) break;
		}
	}
	
	/*--- else...no unmarked neighbor ---*/
	break;
    }
    
    return(answer);
}



/*************************************************************
  Function:  trap_push_unmarked_neighbors(the_A, marks, t, rep,
  find_all, previous)
  Input:     the_A --- an arrangement
  marks --- a pointer to Traverse_Marks structure
  t --- the current trap
  rep --- representative trap for the current face
  find_all --- continue even after finding one?
  previous --- previous list of Traverse_Step objects
  Output:    expanded List of Traverse_Step objects
  Effects:   <none>
  
  Description:  This function finds the neighbors of a trap and
  pushes them onto a running list.  It will first attempt to
  cross threads to get to the neighbors; then it will try to 
  cross edges at the common endpoints.
  **************************************************************/
List trap_push_unmarked_neighbors(the_A, marks, t, rep, find_all, previous)
  Arrange		*the_A;
  Traverse_Marks	*marks;
  Trap			*t;
  Trap			*rep;
  Boolean		find_all;
  List			previous;
{
    List answer;
    
    answer = trap_push_unmarked_neighbors_T(the_A, marks, t, rep, find_all, NULL);
    
    if (find_all || (!answer))
	answer = concatenate(answer, 
			     trap_push_unmarked_neighbors_E(the_A, marks, t, rep,
							    find_all,NIL));
    
    answer = concatenate(answer, previous);
    
    return(answer);
}



/*************************************************************
  Function:  process_unseen_trap(assem, the_dbg, the_A, marks, cur_trap,
  T_Callback, V_Callback, local_data)
  Input:     assem -- the Assembly
  the_dbg --- a valid dbg for cur_trap
  the_A --- an arrangement
  marks --- a pointer to Traverse_Marks structure
  cur_trap --- the current trap
  T_Callback --- a Trap Callback routine
  V_Callback --- a TrapVert Callback routine
  local_data --- local_data for callback routines
  Output:    Boolean flag: should we interrupt the traversal?
  Effects:   calls the necessary callback routines
  
  **************************************************************/
Boolean process_unseen_trap(assem, the_dbg, the_A, marks, cur_trap,
			    T_Callback, V_Callback, local_data)
  Assembly		*assem;
  struct DBG		*the_dbg;
  Arrange		*the_A;
  Traverse_Marks	*marks;
  Trap			*cur_trap;
  Trap_Callback		T_Callback;
  Vert_Callback		V_Callback;
  void			*local_data;
{
    Boolean	interrupt_traversal;
    
#ifdef DEBUG
    if (tr_debug>2)
	printf("Process_Trap(%d)...\n",T_id(cur_trap));
#endif
    
    interrupt_traversal = FALSE;
    
    /*--- mark the trap we're about to process ---*/
    MARK_IT_T(cur_trap,marks);
    
    if (T_Callback)
	interrupt_traversal =
	    T_Callback(assem, the_dbg,
		       cur_trap, the_A,
		       local_data);
    
    if (V_Callback) {
	/*--- check if we've visited the verts yet ---*/
	if ((!MARKED_V(T_rvert(cur_trap),marks)) &&
	    (!interrupt_traversal)) {
	    /*--- a new vert ---*/
	    
	    MARK_IT_V(T_rvert(cur_trap),marks);
	    
	    update_dbg_trap_to_vert(the_dbg,the_A,
				    T_rvert(cur_trap),cur_trap);
	    
	    interrupt_traversal = 
		V_Callback(assem, the_dbg, T_rvert(cur_trap),
			   the_A, local_data);
	    
	    update_dbg_vert_to_trap(the_dbg,the_A,
				    T_rvert(cur_trap),cur_trap);
	}
	
	if ((!MARKED_V(T_lvert(cur_trap),marks)) &&
	    (!interrupt_traversal)) {
	    /*--- a new vert ---*/
	    
	    MARK_IT_V(T_lvert(cur_trap),marks);
	    
	    update_dbg_trap_to_vert(the_dbg,the_A,
				    T_lvert(cur_trap),cur_trap);
	    
	    interrupt_traversal = 
		V_Callback(assem, the_dbg, T_lvert(cur_trap),
			   the_A, local_data);
	    
	    update_dbg_vert_to_trap(the_dbg,the_A,
				    T_lvert(cur_trap),cur_trap);
	}
    }
    
    return(interrupt_traversal);
}



/*************************************************************
  Function:  process_unseen_face(assem, the_dbg, the_A, marks,
  cur_trap, stack, 
  T_Callback, V_Callback, local_data)
  Input:     assem -- the Assembly
  the_dbg --- a valid dbg for cur_trap
  the_A --- an arrangement
  marks --- a pointer to Traverse_Marks structure
  cur_trap --- the current trap
  stack  ---  stack to push new steps to different faces
  T_Callback --- a Trap Callback routine
  V_Callback --- a TrapVert Callback routine
  local_data --- local_data for callback routines
  Output:    Boolean flag: should we interrupt the traversal?
  Effects:   calls the necessary callback routines
  
  **************************************************************/
Boolean process_unseen_face(assem, the_dbg, the_A, marks, cur_rep,
			    stack, T_Callback, V_Callback, local_data)
  Assembly		*assem;
  struct DBG		*the_dbg;
  Arrange		*the_A;
  Traverse_Marks	*marks;
  Trap			*cur_rep;
  List			*stack;
  Trap_Callback		T_Callback;
  Vert_Callback		V_Callback;
  void			*local_data;
{
    Boolean		interrupt_traversal;
    Boolean		backup;
    List		face_stack;
    Trap		*cur_trap;
    Traverse_Step 	*step;
    
    interrupt_traversal = FALSE;
    face_stack = NULL;
    cur_trap = cur_rep;
    
#ifdef DEBUG
    if (tr_debug>2) {
	printf("About to process face(%d) ...\n", T_id(cur_rep));
    }
#endif
    
    /*--- do the Trap Callback only once per face ---*/
    if (T_Callback)
	interrupt_traversal =
	    T_Callback(assem, the_dbg, cur_rep, the_A, local_data);
    
    
    /*-------At this point, we have the DBG for our starting trap-------*/
    /*-------  and 'cur_trap' is set to be that starting trap    -------*/
    face_stack = NIL;
    
    if (!interrupt_traversal) {
	do
	    {
		/*--- Note: calling without the Trap Callback ---*/
		interrupt_traversal = 
		    process_unseen_trap(assem, the_dbg, the_A,
					marks, cur_trap,
					NULL, V_Callback, local_data);
		if (interrupt_traversal) break;
		
		/*--- add trap's unvisited neighbors to face_stack ---*/
		face_stack =
		    trap_push_unmarked_neighbors_T(the_A, marks, cur_trap,
						   cur_rep, TRUE, face_stack);
		
		/*--- add trap's unvisited neighbors to stack ---*/
		(*stack) =
		    trap_push_unmarked_neighbors_E(the_A, marks, cur_trap,
						   cur_rep, TRUE, (*stack));
		
		if (face_stack==NIL) break;
		
		
		/* Find the first step from the running list which	*/
		/* leads to an unvisited trap on this face.		*/
		do
		    {
			step = GET(face_stack, Traverse_Step);
			if (step->from!=cur_trap)
			    {
				/* backtrack */
				cur_trap = step->from;
			    }
			backup = (MARKED_T(step->toward,marks));
			if (backup) POP(face_stack, Traverse_Step);
		    }
		while (backup && (face_stack!=NIL));
		
		if (face_stack!=NIL)
		    {
			/*--- found unvisited neighbor ---*/
			cur_trap = step->toward;
		    }
	    }
	while (face_stack!=NIL);
    }
    
    /*--- if finished early, clear the stack ---*/
    if (face_stack) clear_stack(face_stack);
    
    
    return(interrupt_traversal);
}



/*-------------------------------------*/
/*    Face-by-face traversal routine   */
/*-------------------------------------*/


void traverse_ndbg_fbf(assem, the_A, polylist,
		  T_Callback, V_Callback,
		  local_data)
  Assembly		*assem;
  Arrange		*the_A;
  List			polylist;
  Trap_Callback		T_Callback;
  Vert_Callback		V_Callback;
  void			*local_data;
{
    List		stack;
    Trap 		*cur_rep;
    Boolean 		done;
    Boolean 		backup;
    struct DBG 		*the_dbg;
    Traverse_Step 	*step;
    Traverse_Marks	marks;
    
    reset_marks(the_A,&marks);
    
    cur_rep = NIL;	/* leave choice to create_first_dbg() */
    the_dbg = create_first_dbg(assem, the_A, polylist, &cur_rep);
    
    
    /*-------At this point, we have the DBG for our starting trap-------*/
    /*-------  and 'cur_rep' is set to be that starting trap    -------*/
    stack = NIL;
    
    do
	{
	    /*--- process face, and add unvisited neighbors to stack ---*/
	    done = process_unseen_face(assem, the_dbg, the_A, &marks,
				       cur_rep, &stack,
				       T_Callback, V_Callback, local_data);
	    if (done) break;
	    
	    if (stack==NIL) break;
	    
	    /* Find the first step from the running list which leads to */
	    /* an unvisited trap.					    */
	    do
		{
		    step = GET(stack, Traverse_Step);
		    if (step->face_rep!=cur_rep)
			{
			    /* backtrack && undo any crossing rules */
			    update_dbg_traverse_step(the_dbg, the_A, step, FALSE);
			    cur_rep = step->face_rep;
			}
		    backup = (MARKED_T(step->toward,&marks));
		    if (backup) POP(stack, Traverse_Step);
		}
	    while (backup && (stack!=NIL));
	    
	    if (stack!=NIL)
		{
		    /*--- found unvisited neighbor ---*/
		    update_dbg_traverse_step(the_dbg, the_A, step, TRUE);
		    cur_rep = step->toward;
		}
	}
    while (stack!=NIL);
    
    /*--- if finished early, clear the stack ---*/
    if (stack) clear_stack(stack);
    
    free_marks(&marks);
    
    delete_dbg(the_dbg);
}



/*-------------------------------*/
/*    Trap by Trap Traversal     */
/*-------------------------------*/

void traverse_ndbg(assem, the_A, polylist,
	      T_Callback, V_Callback, local_data)
  Assembly		*assem;
  Arrange		*the_A;
  List		polylist;
  Trap_Callback	T_Callback;
  Vert_Callback	V_Callback;
  void		*local_data;
{
    List		stack;
    Trap 		*cur_trap;
    Boolean 		done;
    Boolean 		backup;
    struct DBG 		*the_dbg;
    Traverse_Step 	*stepp;
    Traverse_Marks	marks;
    
    reset_marks(the_A,&marks);
    
    cur_trap = NIL;	/* leave choice to create_first_dbg() */
    the_dbg = create_first_dbg(assem, the_A, polylist, &cur_trap);
    
#ifdef DEBUG
    if (tr_debug>2) {
	printf("cur_trap(%d)\n",T_id(cur_trap));
	describe_dbg(the_dbg);
    }
#endif
    
    /*-------At this point, we have the DBG for our starting trap-------*/
    /*-------  and 'cur_trap' is set to be that starting trap    -------*/
    stack = NIL;
    
    do
	{
	    done = process_unseen_trap(assem, the_dbg, the_A, &marks, cur_trap,
				       T_Callback, V_Callback, local_data);
	    if (done) break;
	    
	    /*--- add trap's unvisited neighbors to stack ---*/
	    stack=trap_push_unmarked_neighbors(the_A, &marks, cur_trap,
					       NULL, TRUE, stack);
	    if (stack==NIL) break;
	    
	    /* Find the first step from the running list which leads to */
	    /* an unvisited trap.					    */
	    do
		{
		    stepp = GET(stack, Traverse_Step);
		    if (stepp->from!=cur_trap)
			{
			    /* backtrack && undo any crossing rules */
			    update_dbg_traverse_step(the_dbg, the_A, stepp, FALSE);
			    cur_trap = stepp->from;
#ifdef DEBUG
			    if (tr_debug>0)
				verify_dbg_trap(assem, the_dbg, stepp->toward, the_A);
#endif
			}
		    backup = (MARKED_T(stepp->toward,&marks));
		    if (backup) POP(stack, Traverse_Step);
		}
	    while (backup && (stack!=NIL));
	    
	    if (stack!=NIL)
		{
		    /*--- found unvisited neighbor ---*/
		    update_dbg_traverse_step(the_dbg, the_A, stepp, TRUE);
		    cur_trap = stepp->toward;
#ifdef DEBUG
		    if (tr_debug>0)
			verify_dbg_trap(assem, the_dbg, stepp->toward, the_A);
#endif
		}
	}
    while (stack!=NIL);
    
    /*--- if finished early, clear the stack ---*/
    if (stack) clear_stack(stack);
    
    free_marks(&marks);
    
    delete_dbg(the_dbg);
}


